/*
 * Copyright (C) 2011, 2014 Filip Pizlo. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY FILIP PIZLO ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL FILIP PIZLO OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

/*
 * FIXME: this should have a corresponding .cpp file, to help reduce users'
 * compile times and binary sizes.
 */

#ifndef TSF_CPP_HELPERS_HPP
#define TSF_CPP_HELPERS_HPP

#include "tsf.h"

#include <exception>
#include <string>
#include <string.h>
#include <stdlib.h>

namespace tsf {

// exception baseclass; never instantiated directly.
class exception: public std::exception {
protected:
    std::string msg;
    char *suffix;

public:
    exception()
        : msg(std::string("TSF Error: ")+tsf_get_error())
    {
        const char *cur_suffix=tsf_get_error_suffix();
        if (cur_suffix==NULL) {
            this->suffix=NULL;
        } else {
            this->suffix=strdup(cur_suffix);
        }
    }
    
    virtual ~exception() throw() {
        if (suffix!=NULL) {
            free(suffix);
        }
    }
    
    virtual tsf_error_t code() const throw() =0;
    
    virtual const char *what() const throw() {
        return msg.c_str();
    }
    
    virtual void set_tsf_error() const throw() =0;
};

// baseclass of exceptions that don't have a TSF infix; never
// instantiated directly.
class simple_exception: public exception {
public:
    simple_exception() {}
    
    virtual void set_tsf_error() const throw() {
        if (suffix!=NULL) {
            tsf_set_error(code(),"%s",suffix);
        } else {
            tsf_set_error(code(),NULL);
        }
    }
};

#define TSF_EXCEPTION_CLASS_DECL(errorcode,name)                \
    class name##_exception: public simple_exception {           \
    public:                                                     \
        name##_exception() {}                                   \
                                                                \
        virtual tsf_error_t code() const throw() {              \
            return errorcode;                                   \
        }                                                       \
    }

TSF_EXCEPTION_CLASS_DECL(TSF_E_INVALID_ARG,invalid_arg);
TSF_EXCEPTION_CLASS_DECL(TSF_E_BAD_STATE,bad_state);
TSF_EXCEPTION_CLASS_DECL(TSF_E_STR_OVERFLOW,str_overflow);
TSF_EXCEPTION_CLASS_DECL(TSF_E_PARSE_ERROR,parse);
TSF_EXCEPTION_CLASS_DECL(TSF_E_UNEXPECTED_TYPE,unexpected_type);
TSF_EXCEPTION_CLASS_DECL(TSF_E_ELEMENT_NULL,element_null);
TSF_EXCEPTION_CLASS_DECL(TSF_E_NAME_COLLISION,name_collision);
TSF_EXCEPTION_CLASS_DECL(TSF_E_EOF,eof);
TSF_EXCEPTION_CLASS_DECL(TSF_E_UNEXPECTED_EOF,unexpected_eof);
TSF_EXCEPTION_CLASS_DECL(TSF_E_ERRONEOUS_EOF,erroneous_eof);
TSF_EXCEPTION_CLASS_DECL(TSF_E_NO_EOF,no_eof);
TSF_EXCEPTION_CLASS_DECL(TSF_E_BAD_TYPE,bad_type);
TSF_EXCEPTION_CLASS_DECL(TSF_E_BAD_TYPE_KIND,bad_type_kind);
TSF_EXCEPTION_CLASS_DECL(TSF_E_ELE_NOT_FOUND,ele_not_found);
TSF_EXCEPTION_CLASS_DECL(TSF_E_NO_STRUCT_MAP,no_struct_map);
TSF_EXCEPTION_CLASS_DECL(TSF_E_BAD_STRUCT_MAP,bad_struct_map);
TSF_EXCEPTION_CLASS_DECL(TSF_E_TYPE_MISMATCH,type_mismatch);
TSF_EXCEPTION_CLASS_DECL(TSF_E_BAD_CHOICE_VALUE,bad_choice_value);
TSF_EXCEPTION_CLASS_DECL(TSF_E_COMPARE_FAIL,compare_fail);
TSF_EXCEPTION_CLASS_DECL(TSF_E_COMPILE_ERROR,compile);
TSF_EXCEPTION_CLASS_DECL(TSF_E_DLOADING_ERROR,dloading);
TSF_EXCEPTION_CLASS_DECL(TSF_E_BAD_JUMP,bad_jump);
TSF_EXCEPTION_CLASS_DECL(TSF_E_SEMANTIC,semantic);
TSF_EXCEPTION_CLASS_DECL(TSF_E_CAS_FAIL,cas_fail);
TSF_EXCEPTION_CLASS_DECL(TSF_E_HOSTNAME,hostname);
TSF_EXCEPTION_CLASS_DECL(TSF_E_NOT_ALLOWED,not_allowed);
TSF_EXCEPTION_CLASS_DECL(TSF_E_NOT_IMPLEMENTED,not_implemented);
TSF_EXCEPTION_CLASS_DECL(TSF_E_NOT_SUPPORTED,not_supported);
TSF_EXCEPTION_CLASS_DECL(TSF_E_INTERNAL,internal);
TSF_EXCEPTION_CLASS_DECL(TSF_E_EXTERNAL,external);

class sys_exception: public exception {
private:
    int the_errno;
    
public:
    sys_exception()
        : the_errno(tsf_get_system_errno())
    {
    }
    
    virtual ~sys_exception() throw() {}
    
    virtual tsf_error_t code() const throw() {
        return TSF_E_ERRNO;
    }
    
    int sys_errno() const throw() {
        return the_errno;
    }
    
    virtual void set_tsf_error() const throw() {
        if (suffix!=NULL) {
            tsf_set_specific_errno(the_errno,"%s",suffix);
        } else {
            tsf_set_specific_errno(the_errno,NULL);
        }
    }
};

class zlib_exception: public exception {
private:
    int error_code_;
    std::string error_msg_;

public:
    zlib_exception()
        : error_code_(tsf_get_zlib_error_code())
        , error_msg_(tsf_get_zlib_error_msg()) {
    }
    
    virtual ~zlib_exception() throw() {}

    virtual tsf_error_t code() const throw() {
        return TSF_E_ZLIB_ERROR;
    }
    
    int zlib_error_code() const throw() {
        return error_code_;
    }
    
    const char *zlib_error_msg() const throw() {
        return error_msg_.c_str();
    }
    
    virtual void set_tsf_error() const throw() {
        if (suffix!=NULL) {
            tsf_set_zlib_error(error_code_,
                               error_msg_.c_str(),
                               "%s",suffix);
        } else {
            tsf_set_zlib_error(error_code_,
                               error_msg_.c_str(),
                               NULL);
        }
    }
};

class libbzip2_exception: public exception {
private:
    int error_code;

public:
    libbzip2_exception()
        : error_code(tsf_get_libbzip2_error_code())
    {
    }
    
    virtual ~libbzip2_exception() throw() {}

    virtual tsf_error_t code() const throw() {
        return TSF_E_LIBBZIP2_ERROR;
    }
    
    int libbzip2_error_code() const throw() {
        return error_code;
    }
    
    virtual void set_tsf_error() const throw() {
        if (suffix!=NULL) {
            tsf_set_libbzip2_error(error_code,"%s",suffix);
        } else {
            tsf_set_libbzip2_error(error_code,NULL);
        }
    }
};

class death_by_signal_exception: public exception {
private:
    int deadly_signal_;

public:
    death_by_signal_exception()
        : deadly_signal_(tsf_get_deadly_signal())
    {
    }
    
    virtual ~death_by_signal_exception() throw() {}

    virtual tsf_error_t code() const throw() {
        return TSF_E_DEATH_BY_SIGNAL;
    }
    
    int deadly_signal() const throw() {
        return deadly_signal_;
    }
    
    virtual void set_tsf_error() const throw() {
        if (suffix!=NULL) {
            tsf_set_deadly_signal(deadly_signal_,"%s",suffix);
        } else {
            tsf_set_deadly_signal(deadly_signal_,NULL);
        }
    }
};

static void throw_exception_from_error() {
    switch (tsf_get_error_code()) {
    case TSF_E_ERRNO: throw sys_exception();
    case TSF_E_ZLIB_ERROR: throw zlib_exception();
    case TSF_E_LIBBZIP2_ERROR: throw libbzip2_exception();
    case TSF_E_INVALID_ARG: throw invalid_arg_exception();
    case TSF_E_BAD_STATE: throw bad_state_exception();
    case TSF_E_STR_OVERFLOW: throw str_overflow_exception();
    case TSF_E_PARSE_ERROR: throw parse_exception();
    case TSF_E_UNEXPECTED_TYPE: throw unexpected_type_exception();
    case TSF_E_ELEMENT_NULL: throw element_null_exception();
    case TSF_E_NAME_COLLISION: throw name_collision_exception();
    case TSF_E_EOF: throw eof_exception();
    case TSF_E_UNEXPECTED_EOF: throw unexpected_eof_exception();
    case TSF_E_ERRONEOUS_EOF: throw erroneous_eof_exception();
    case TSF_E_NO_EOF: throw no_eof_exception();
    case TSF_E_BAD_TYPE: throw bad_type_exception();
    case TSF_E_BAD_TYPE_KIND: throw bad_type_kind_exception();
    case TSF_E_ELE_NOT_FOUND: throw ele_not_found_exception();
    case TSF_E_NO_STRUCT_MAP: throw no_struct_map_exception();
    case TSF_E_BAD_STRUCT_MAP: throw bad_struct_map_exception();
    case TSF_E_TYPE_MISMATCH: throw type_mismatch_exception();
    case TSF_E_BAD_CHOICE_VALUE: throw bad_choice_value_exception();
    case TSF_E_COMPARE_FAIL: throw compare_fail_exception();
    case TSF_E_COMPILE_ERROR: throw compile_exception();
    case TSF_E_DEATH_BY_SIGNAL: throw death_by_signal_exception();
    case TSF_E_DLOADING_ERROR: throw dloading_exception();
    case TSF_E_BAD_JUMP: throw bad_jump_exception();
    case TSF_E_SEMANTIC: throw semantic_exception();
    case TSF_E_CAS_FAIL: throw cas_fail_exception();
    case TSF_E_HOSTNAME: throw hostname_exception();
    case TSF_E_NOT_ALLOWED: throw not_allowed_exception();
    case TSF_E_NOT_IMPLEMENTED: throw not_implemented_exception();
    case TSF_E_NOT_SUPPORTED: throw not_supported_exception();
    case TSF_E_EXTERNAL: throw external_exception();
    case TSF_E_INTERNAL: throw internal_exception();
    default: abort();
    }
}

template< typename T >
inline T check(T value) {
    if (!value) {
        throw_exception_from_error();
    }
    return value;
}

inline int32_t check(int32_t value) {
    if (value<0) {
        throw_exception_from_error();
    }
    return value;
}

static void set_error_from_exception() {
    try {
        throw;
    } catch (const exception& e) {
        e.set_tsf_error();
    } catch (const std::exception &e) {
        tsf_set_error(TSF_E_EXTERNAL,"C++ exception: %s",e.what());
    } catch (...) {
        tsf_set_error(TSF_E_EXTERNAL,"Unknown C++ exception");
    }
}

} // namespace tsf

#endif
