/*
 * Copyright (C) 2003, 2004, 2005 Filip Pizlo. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY FILIP PIZLO ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL FILIP PIZLO OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "tsf.h"
#include "complex_any_test_types.h"
#include "tsf_test.h"

#include <stdarg.h>


/* the foundation of the application.  these functions will be called in
 * strange and contrived combinations in order to stress the any type and
 * its associated facilities in weird and unnatural ways. */

tsf_buffer_t *cat_chat_msg_create(void *region,
                                  const char *target,
                                  const char *source,
                                  const char *message,
                                  size_t message_len) {
    cat_chat_msg_t data;
    data.target=(char*)target;
    data.source=(char*)source;
    data.message.data=(uint8_t*)message;
    data.message.len=message_len;
    
    return cat_chat_msg__generate_in(&data,region);
}

uint64_t cat_file_frag_hash_code(const void *file_frag,
                                 size_t file_frag_len) {
    uint8_t *cur;
    uint64_t hash=0;
    for (cur=(uint8_t*)file_frag;
         cur<((uint8_t*)file_frag)+file_frag_len;
         ++cur) {
        hash+=*cur;
    }
    return hash;
}

tsf_buffer_t *cat_file_frag_create(void *region,
                                   const void *file_frag,
                                   size_t file_frag_len) {
    cat_file_frag_t data;
    data.hash=cat_file_frag_hash_code(file_frag,file_frag_len);
    data.data.data=(uint8_t*)file_frag;
    data.data.len=file_frag_len;
    
    return cat_file_frag__generate_in(&data,region);
}

tsf_buffer_t *cat_topo_msg_create(void *region,
                                  const char *source,
                                  const char **adjacency,
                                  size_t num_adjacents) {
    cat_topo_msg_t data;
    data.source=(char*)source;
    data.adjacency.data=(char**)adjacency;
    data.adjacency.len=num_adjacents;
    
    return cat_topo_msg__generate_in(&data,region);
}

tsf_buffer_t *cat_topo_msg_create_easy(void *region,
                                       const char *source,
                                       ...) {
    const char **adjacency=NULL;
    size_t num_adjacents=0;
    va_list lst;
    tsf_buffer_t *ret;
    
    va_start(lst,source);
    
    for (;;) {
        const char *cur=va_arg(lst,const char*);
        if (cur==NULL) {
            break;
        }
        
        if (adjacency==NULL) {
            adjacency=malloc(sizeof(const char*));
            if (adjacency==NULL) {
                tsf_set_errno("Could not malloc const char *");
                va_end(lst);
                return NULL;
            }
        } else {
            const char **new_array=
                realloc(adjacency,
                        sizeof(const char*)*(num_adjacents+1));
            if (new_array==NULL) {
                tsf_set_errno("Could not realloc array of const char*");
                va_end(lst);
                free(adjacency);
                return NULL;
            }
            adjacency=new_array;
        }
        
        adjacency[num_adjacents++]=cur;
    }
    
    ret=cat_topo_msg_create(region,source,adjacency,num_adjacents);
    
    free(adjacency);
    
    return ret;
}

cat_carrier_t *cat_carrier_create(void *region) {
    cat_carrier_t *ret=
        tsf_region_alloc(region,sizeof(cat_carrier_t));
    
    if (ret==NULL) {
        tsf_set_errno("Could not tsf_region_alloc cat_carrier_t");
        return NULL;
    }
    
    ret->data=NULL;
    ret->len=0;
    
    return ret;
}

tsf_bool_t cat_carrier_append(void *region,
                              cat_carrier_t *carrier,
                              tsf_buffer_t *buf) {
    tsf_buffer_t **new_array;
    
    if (buf==NULL) {
        return tsf_false;
    }
    
    new_array=
        tsf_region_alloc(region,
            sizeof(tsf_buffer_t*)*(1+carrier->len));
        
    if (new_array==NULL) {
        tsf_set_errno("Could not tsf_region_alloc array of "
                      "tsf_buffer_t*");
        return tsf_false;
    }
    
    memcpy(new_array,
           carrier->data,
           sizeof(tsf_buffer_t*)*carrier->len);
    
    carrier->data=new_array;
    
    carrier->data[carrier->len++]=buf;
    
    return tsf_true;
}

/* and here come the real tests.... */

void test_empty_carrier() {
    void *region;
    cat_carrier_t *carrier;
    tsf_buffer_t *buffer;
    
    TEST_ASSERT_ERRNO(region=tsf_region_create(0));
    
    TEST_ASSERT_TSFERROR(carrier=cat_carrier_create(region));
    TEST_ASSERT_TSFERROR(buffer=cat_carrier__generate_in(carrier,region));
    
    TEST_ASSERT_TSFERROR(carrier=cat_carrier__parse_in(buffer,region));
    
    TEST_ASSERT(carrier->len==0);
    
    tsf_region_free(region);
}

void test_n_chat_message(unsigned num) {
    void *region;
    cat_carrier_t *carrier;
    tsf_buffer_t *buffer;
    unsigned n;
    
    TEST_ASSERT_ERRNO(region=tsf_region_create(0));
    
    TEST_ASSERT_TSFERROR(carrier=cat_carrier_create(region));
    
    for (n=num;n-->0;) {
        TEST_ASSERT_TSFERROR(cat_carrier_append(region,carrier,
            cat_chat_msg_create(region,
                                "target",
                                "source",
                                "Hello, World!",
                                sizeof("Hello, world!"))));
    }
    
    TEST_ASSERT(carrier->len==num);
    
    TEST_ASSERT_TSFERROR(buffer=cat_carrier__generate_in(carrier,region));
    
    TEST_ASSERT_TSFERROR(carrier=cat_carrier__parse_in(buffer,region));
    
    TEST_ASSERT(carrier->len==num);
    
    for (n=num;n-->0;) {
        cat_chat_msg_t *chat_msg;
        
        TEST_ASSERT_TSFERROR(chat_msg=
            cat_chat_msg__parse_in(carrier->data[n],region));
        
        TEST_ASSERT(!strcmp(chat_msg->target,"target"));
        TEST_ASSERT(!strcmp(chat_msg->source,"source"));
        TEST_ASSERT(chat_msg->message.len==sizeof("Hello, World!"));
        TEST_ASSERT(!strcmp((const char*)chat_msg->message.data,
                            "Hello, World!"));
    }
    
    tsf_region_free(region);
}

void test_n_file_fragment(unsigned num) {
    void *region;
    cat_carrier_t *carrier;
    tsf_buffer_t *buffer;
    unsigned n;
    
    TEST_ASSERT_ERRNO(region=tsf_region_create(0));
    
    TEST_ASSERT_TSFERROR(carrier=cat_carrier_create(region));
    
    for (n=num;n-->0;) {
        TEST_ASSERT_TSFERROR(cat_carrier_append(region,carrier,
            cat_file_frag_create(region,
                                 "I'm a file fragment",
                                 sizeof("I'm a file fragment"))));
    }
    
    TEST_ASSERT(carrier->len==num);
    
    TEST_ASSERT_TSFERROR(buffer=cat_carrier__generate_in(carrier,region));
    
    TEST_ASSERT_TSFERROR(carrier=cat_carrier__parse_in(buffer,region));
    
    TEST_ASSERT(carrier->len==num);
    
    for (n=num;n-->0;) {
        cat_file_frag_t *file_frag;
        
        TEST_ASSERT_TSFERROR(file_frag=
            cat_file_frag__parse_in(carrier->data[n],region));
        
        TEST_ASSERT(file_frag->data.len
                    ==sizeof("I'm a file fragment"));
        
        TEST_ASSERT(!memcmp(file_frag->data.data,
                            "I'm a file fragment",
                            sizeof("I'm a file fragment")));
        
        TEST_ASSERT(file_frag->hash
                    == cat_file_frag_hash_code(file_frag->data.data,
                                               file_frag->data.len));
    }
    
    tsf_region_free(region);
}

void test_n_topo_message(unsigned num) {
    void *region;
    cat_carrier_t *carrier;
    tsf_buffer_t *buffer;
    unsigned n;
    
    TEST_ASSERT_ERRNO(region=tsf_region_create(0));
    
    TEST_ASSERT_TSFERROR(carrier=cat_carrier_create(region));
    
    for (n=num;n-->0;) {
        TEST_ASSERT_TSFERROR(cat_carrier_append(region,carrier,
            cat_topo_msg_create_easy(region,
                                     "bocks",
                                     "sphinx",
                                     "kaczor",
                                     "max",
                                     NULL)));
    }
    
    TEST_ASSERT(carrier->len==num);
    
    TEST_ASSERT_TSFERROR(buffer=cat_carrier__generate_in(carrier,region));
    
    TEST_ASSERT_TSFERROR(carrier=cat_carrier__parse_in(buffer,region));
    
    TEST_ASSERT(carrier->len==num);
    
    for (n=num;n-->0;) {
        cat_topo_msg_t *topo_msg;
        
        TEST_ASSERT_TSFERROR(topo_msg=
            cat_topo_msg__parse_in(carrier->data[n],region));
        
        TEST_ASSERT(!strcmp(topo_msg->source,"bocks"));
        
        TEST_ASSERT(topo_msg->adjacency.len==3);
        
        TEST_ASSERT(!strcmp(topo_msg->adjacency.data[0],"sphinx"));
        TEST_ASSERT(!strcmp(topo_msg->adjacency.data[1],"kaczor"));
        TEST_ASSERT(!strcmp(topo_msg->adjacency.data[2],"max"));
    }
    
    tsf_region_free(region);
}

void test_one_of_each() {
    void *region;
    cat_carrier_t *carrier;
    tsf_buffer_t *buffer;
    
    cat_chat_msg_t *chat_msg;
    cat_file_frag_t *file_frag;
    cat_topo_msg_t *topo_msg;
    
    TEST_ASSERT_ERRNO(region=tsf_region_create(0));
    
    TEST_ASSERT_TSFERROR(carrier=cat_carrier_create(region));
    
    TEST_ASSERT_TSFERROR(cat_carrier_append(region,carrier,
        cat_chat_msg_create(region,
                            "bob",
                            "fred",
                            "Impeach Bush",
                            sizeof("Impeach Bush"))));
    
    TEST_ASSERT_TSFERROR(cat_carrier_append(region,carrier,
        cat_file_frag_create(region,
                             "Britney Spears MP3",
                             sizeof("Britney Spears MP3"))));
    
    TEST_ASSERT_TSFERROR(cat_carrier_append(region,carrier,
        cat_topo_msg_create_easy(region,
                                 "gaea",
                                 "saji",
                                 "knyr",
                                 NULL)));
    
    TEST_ASSERT(carrier->len == 3);
    
    TEST_ASSERT_TSFERROR(buffer=cat_carrier__generate_in(carrier,region));
    
    TEST_ASSERT_TSFERROR(carrier = cat_carrier__parse_in(buffer, region));
    
    TEST_ASSERT(carrier->len == 3);
    
    TEST_ASSERT_TSFERROR(chat_msg=
        cat_chat_msg__parse_in(carrier->data[0],region));
    
    TEST_ASSERT(!strcmp(chat_msg->target,"bob"));
    TEST_ASSERT(!strcmp(chat_msg->source,"fred"));
    TEST_ASSERT(chat_msg->message.len==sizeof("Impeach Bush"));
    TEST_ASSERT(!strcmp((const char*)chat_msg->message.data,
                        "Impeach Bush"));
    
    TEST_ASSERT_TSFERROR(file_frag=
        cat_file_frag__parse_in(carrier->data[1],region));
    
    TEST_ASSERT(file_frag->data.len
                ==sizeof("Britney Spears MP3"));
    TEST_ASSERT(!strcmp((const char*)file_frag->data.data,
                        "Britney Spears MP3"));
    TEST_ASSERT(file_frag->hash
                == cat_file_frag_hash_code("Britney Spears MP3",
                                           sizeof("Britney Spears MP3")));
    
    TEST_ASSERT_TSFERROR(topo_msg=
        cat_topo_msg__parse_in(carrier->data[2],region));
    
    TEST_ASSERT(!strcmp(topo_msg->source,"gaea"));
    TEST_ASSERT(topo_msg->adjacency.len==2);
    TEST_ASSERT(!strcmp(topo_msg->adjacency.data[0],"saji"));
    TEST_ASSERT(!strcmp(topo_msg->adjacency.data[1],"knyr"));
    
    tsf_region_free(region);
}

int main(int c,char **v) {
#ifdef TSF_FAULT_INJECT
    fprintf(stderr,"%s: this test is known to fail fault injection miserably.\n",
                   v[0]);
    return 0;
#endif
    
    TEST_INIT(c,v);
    
    DO_TEST(test_empty_carrier());
    
    DO_TEST(test_n_chat_message(0));
    DO_TEST(test_n_chat_message(1));
    DO_TEST(test_n_chat_message(2));
    DO_TEST(test_n_chat_message(10));
    
    DO_TEST(test_n_file_fragment(0));
    DO_TEST(test_n_file_fragment(1));
    DO_TEST(test_n_file_fragment(2));
    DO_TEST(test_n_file_fragment(10));
    
    DO_TEST(test_n_topo_message(0));
    DO_TEST(test_n_topo_message(1));
    DO_TEST(test_n_topo_message(2));
    DO_TEST(test_n_topo_message(10));
    
    DO_TEST(test_one_of_each());
    
    TEST_DONE();
    
    return 0;
}





