/*
 * Copyright (C) 2004, 2005, 2014 Filip Pizlo. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY FILIP PIZLO ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL FILIP PIZLO OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "tsf_internal.h"
#include "tsf_test.h"

#include <errno.h>
#include <signal.h>

#ifdef TSF_HAVE_ZLIB
#include <zlib.h>
#endif

#ifdef TSF_HAVE_BZIP2
#include <bzlib.h>
#endif

static void test_set_error(tsf_error_t code,
			   const char *msg) {
    tsf_set_error(code, msg);
    TA(tsf_get_error_code() == code);
    TASE(msg,tsf_get_error_suffix());
}

static void test_error_string(tsf_error_t code, const char *expected) {
    TASE(expected, tsf_get_message_for_error_code(code));
    tsf_set_error(code, NULL);
    TA(strstr(tsf_get_error(), expected));
}

static void test_set_errno(int my_errno,
			   const char *msg) {
    errno=my_errno;
    tsf_set_errno(msg);
    TA(tsf_get_error_code() == TSF_E_ERRNO);
    TA(tsf_get_system_errno() == my_errno);
    TA(strstr(tsf_get_error(), "System error"));
    TASE(strerror(my_errno),tsf_get_error_infix());
    TASE(msg,tsf_get_error_suffix());
}

static void test_set_specific_errno(int my_errno,
				    const char *msg) {
    tsf_set_specific_errno(my_errno,msg);
    TA(tsf_get_error_code() == TSF_E_ERRNO);
    TA(tsf_get_system_errno() == my_errno);
    TASE(strerror(my_errno),tsf_get_error_infix());
    TASE(msg,tsf_get_error_suffix());
}

static void test_set_deadly_signal(int signal,
				   const char *msg) {
    tsf_set_deadly_signal(signal,msg);
    TA(tsf_get_error_code() == TSF_E_DEATH_BY_SIGNAL);
    TA(tsf_get_deadly_signal() == signal);
    TA(strstr(tsf_get_error(), "Child process died because it received a signal"));
    TASE(tsf_asprintf("Received signal #%d",signal),
	 tsf_get_error_infix());
    TASE(msg,tsf_get_error_suffix());
}

#ifdef TSF_HAVE_ZLIB

static void test_set_zlib_error(int zlib_error,
                                const char *zlib_msg,
				const char *msg,
                                int resulting_code,
                                int resulting_errno) {
    tsf_set_zlib_error(zlib_error,zlib_msg,msg);
    TA(tsf_get_error_code()==resulting_code);
    if (resulting_code==TSF_E_ZLIB_ERROR) {
        TA(tsf_get_zlib_error_code()==zlib_error);
        if (zlib_msg!=NULL) {
            TA(tsf_get_zlib_error_msg()==zlib_msg);
	    TASE(zlib_msg,tsf_get_error_infix());
        } else {
	    TA(tsf_get_error_infix()!=NULL);
	}
    }
    if (resulting_errno>=0) {
        TA(tsf_get_system_errno()==resulting_errno);
	TASE(strerror(resulting_errno),
	     tsf_get_error_infix());
    }
    TASE(msg,tsf_get_error_suffix());
}

static void test_set_zlib_error_with_errno(int zlib_error,
                                           const char *zlib_msg,
                                           const char *msg,
                                           int resulting_code,
                                           int my_errno) {
    errno = my_errno;
    test_set_zlib_error(zlib_error, zlib_msg, msg, resulting_code, my_errno);
}
#endif

#ifdef TSF_HAVE_BZIP2

static void test_set_libbzip2_error(int bzip2_error,
				    const char *msg,
                                    int resulting_code,
                                    int resulting_errno) {
    tsf_set_libbzip2_error(bzip2_error,msg);
    TA(tsf_get_error_code()==resulting_code);
    if (resulting_code==TSF_E_LIBBZIP2_ERROR) {
        TA(tsf_get_libbzip2_error_code()==bzip2_error);
	TA(tsf_get_error_infix()!=NULL);
    }
    if (resulting_errno>=0) {
        TA(tsf_get_system_errno()==resulting_errno);
	TASE(strerror(resulting_errno),
	     tsf_get_error_infix());
    }
    TASE(msg,tsf_get_error_suffix());
}

#endif

int main(int c,char **v) {
    tsf_error_t i;

    TEST_INIT(c,v);
    
    for (i=0;
         i<TSF_E_LAST;
         ++i) {
        TEST_PUSH("i=%d",i);
        switch (i) {
            case TSF_E_ERRNO:
            case TSF_E_DEATH_BY_SIGNAL:
                break;
            default:
                DO_TEST(test_set_error(i,NULL));
                DO_TEST(test_set_error(i,"hello, world"));
                break;
        }
        TEST_POP();
    }
    
    /* make sure that error messages make sense */
    DO_TEST(test_error_string(TSF_E_ERRNO, "System error"));
    DO_TEST(test_error_string(TSF_E_ZLIB_ERROR, "ZLib error"));
    DO_TEST(test_error_string(TSF_E_LIBBZIP2_ERROR, "libbzip2 error"));
    DO_TEST(test_error_string(TSF_E_INVALID_ARG, "Invalid argument"));
    DO_TEST(test_error_string(TSF_E_BAD_STATE, "Bad state"));
    DO_TEST(test_error_string(TSF_E_STR_OVERFLOW, "String too long"));
    DO_TEST(test_error_string(TSF_E_PARSE_ERROR, "Parse error"));
    DO_TEST(test_error_string(TSF_E_UNEXPECTED_TYPE, "Parsed an unexpected type"));
    DO_TEST(test_error_string(TSF_E_ELEMENT_NULL, "Element in container is NULL"));
    DO_TEST(test_error_string(TSF_E_NAME_COLLISION, "Name already used"));
    DO_TEST(test_error_string(TSF_E_EOF, "End of file"));
    DO_TEST(test_error_string(TSF_E_UNEXPECTED_EOF, "Unexpected end of file"));
    DO_TEST(test_error_string(TSF_E_ERRONEOUS_EOF, "Erroneous end of file"));
    DO_TEST(test_error_string(TSF_E_NO_EOF, "End of file not reached"));
    DO_TEST(test_error_string(TSF_E_BAD_TYPE, "Parameter has bad type"));
    DO_TEST(test_error_string(TSF_E_BAD_TYPE_KIND, "Parameter has bad type kind"));
    DO_TEST(test_error_string(TSF_E_ELE_NOT_FOUND, "Element not found"));
    DO_TEST(test_error_string(TSF_E_NO_STRUCT_MAP, "Type lacks struct mapping"));
    DO_TEST(test_error_string(TSF_E_BAD_STRUCT_MAP, "The provided structure mapping is invalid"));
    DO_TEST(test_error_string(TSF_E_TYPE_MISMATCH, "Type mismatch"));
    DO_TEST(test_error_string(TSF_E_BAD_CHOICE_VALUE, "Bad choice value"));
    DO_TEST(test_error_string(TSF_E_COMPARE_FAIL, "Comparison failed"));
    DO_TEST(test_error_string(TSF_E_COMPILE_ERROR, "C compile error"));
    DO_TEST(test_error_string(TSF_E_DEATH_BY_SIGNAL, "Child process died because it received a signal"));
    DO_TEST(test_error_string(TSF_E_DLOADING_ERROR, "Dynamic loading error"));
    DO_TEST(test_error_string(TSF_E_BAD_JUMP, "Jump to bad label in GPC prototype"));
    DO_TEST(test_error_string(TSF_E_SEMANTIC, "Violation of GPC semantics"));
    DO_TEST(test_error_string(TSF_E_CAS_FAIL, "Compare-and-swap failed"));
    DO_TEST(test_error_string(TSF_E_HOSTNAME, "Hostname lookup failure"));
    DO_TEST(test_error_string(TSF_E_NOT_ALLOWED, "Operation not allowed"));
    DO_TEST(test_error_string(TSF_E_NOT_IMPLEMENTED, "Operation not implemented"));
    DO_TEST(test_error_string(TSF_E_NOT_SUPPORTED, "Operation not supported"));
    DO_TEST(test_error_string(TSF_E_EXTERNAL, "External error"));
    DO_TEST(test_error_string(TSF_E_INTERNAL, "Internal error"));
    
    /* just try a couple of system errno's */
    DO_TEST(test_set_errno(EFAULT,NULL));
    DO_TEST(test_set_errno(EPERM,NULL));
    DO_TEST(test_set_errno(ENOSYS,NULL));
    DO_TEST(test_set_errno(EINTR,NULL));
    
    DO_TEST(test_set_errno(EFAULT,"hello, world"));
    DO_TEST(test_set_errno(EPERM,"hello, world"));
    DO_TEST(test_set_errno(ENOSYS,"hello, world"));
    DO_TEST(test_set_errno(EINTR,"hello, world"));
    
    /* do it again... */
    DO_TEST(test_set_specific_errno(EFAULT,NULL));
    DO_TEST(test_set_specific_errno(EPERM,NULL));
    DO_TEST(test_set_specific_errno(ENOSYS,NULL));
    DO_TEST(test_set_specific_errno(EINTR,NULL));
    
    DO_TEST(test_set_specific_errno(EFAULT,"hello, world"));
    DO_TEST(test_set_specific_errno(EPERM,"hello, world"));
    DO_TEST(test_set_specific_errno(ENOSYS,"hello, world"));
    DO_TEST(test_set_specific_errno(EINTR,"hello world"));
    
    DO_TEST(test_set_deadly_signal(SIGINT,NULL));
    DO_TEST(test_set_deadly_signal(SIGKILL,NULL));   /* heh */
    DO_TEST(test_set_deadly_signal(SIGQUIT,NULL));
        
    DO_TEST(test_set_deadly_signal(SIGINT,"foo"));
    DO_TEST(test_set_deadly_signal(SIGKILL,"bar"));   /* heh */
    DO_TEST(test_set_deadly_signal(SIGQUIT,"snafu"));
        
#ifdef TSF_HAVE_ZLIB
    DO_TEST(test_set_zlib_error_with_errno(Z_ERRNO,NULL,NULL,TSF_E_ERRNO,EINVAL));
    
    DO_TEST(test_set_zlib_error(Z_MEM_ERROR,NULL,NULL,TSF_E_ERRNO,ENOMEM));
    DO_TEST(test_set_zlib_error(Z_DATA_ERROR,NULL,NULL,TSF_E_ZLIB_ERROR,-1));
    DO_TEST(test_set_zlib_error(Z_DATA_ERROR,"data error",NULL,TSF_E_ZLIB_ERROR,-1));

    DO_TEST(test_set_zlib_error_with_errno(Z_ERRNO,NULL,"z",TSF_E_ERRNO,EINVAL));
    
    DO_TEST(test_set_zlib_error(Z_MEM_ERROR,NULL,"l",TSF_E_ERRNO,ENOMEM));
    DO_TEST(test_set_zlib_error(Z_DATA_ERROR,NULL,"i",TSF_E_ZLIB_ERROR,-1));
    DO_TEST(test_set_zlib_error(Z_DATA_ERROR,"data error","b",TSF_E_ZLIB_ERROR,-1));
#endif
        
#ifdef TSF_HAVE_BZIP2
    errno=EINVAL;
    DO_TEST(test_set_libbzip2_error(BZ_IO_ERROR,NULL,TSF_E_ERRNO,EINVAL));
    
    DO_TEST(test_set_libbzip2_error(BZ_MEM_ERROR,NULL,TSF_E_ERRNO,ENOMEM));
    DO_TEST(test_set_libbzip2_error(BZ_UNEXPECTED_EOF,NULL,TSF_E_ERRONEOUS_EOF,-1));
    DO_TEST(test_set_libbzip2_error(BZ_DATA_ERROR,NULL,TSF_E_LIBBZIP2_ERROR,-1));

    errno=EINVAL;
    DO_TEST(test_set_libbzip2_error(BZ_IO_ERROR,"b",TSF_E_ERRNO,EINVAL));
    
    DO_TEST(test_set_libbzip2_error(BZ_MEM_ERROR,"z",TSF_E_ERRNO,ENOMEM));
    DO_TEST(test_set_libbzip2_error(BZ_UNEXPECTED_EOF,"i",TSF_E_ERRONEOUS_EOF,-1));
    DO_TEST(test_set_libbzip2_error(BZ_DATA_ERROR,"p",TSF_E_LIBBZIP2_ERROR,-1));
#endif

    TEST_DONE();
    
    return 0;
}

