/*
 * Copyright (C) 2011 Filip Pizlo. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY FILIP PIZLO ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL FILIP PIZLO OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "tsf.h"
#include "tsf_internal.h"
#include "tsf_test.h"
#include "interesting_data.h"

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>

static tsf_bool_t remote;
static char *hostname;
static int port;
static tsf_fsdb_t *file;

static void t_reset() {
    if (!remote) {
        if (1) system("rm -rf fsdb_test.dat");
    }
}

static void t_create(tsf_bool_t truncate) {
    if (remote) {
        TAT(file=tsf_fsdb_open_remote(hostname,port,NULL,NULL,NULL));
    } else {
        TAT(file=tsf_fsdb_open_local("fsdb_test.dat",NULL,NULL,NULL,tsf_true,truncate));
    }
}

static void t_open(tsf_bool_t update) {
    if (remote) {
        TAT(file=tsf_fsdb_open_remote(hostname,port,NULL,NULL,NULL));
    } else {
        TAT(file=tsf_fsdb_open_local("fsdb_test.dat",NULL,NULL,NULL,update,tsf_false));
    }
}

static void t_close() {
    tsf_fsdb_close(file);
}

static void t_destroy() {
    t_close();
    t_reset();
}

static void test_empty() {
    t_reset();
    t_create(tsf_true);
    t_destroy();
}

static void test_empty_reopen() {
    t_reset();
    t_create(tsf_true);
    t_close();
    t_open(tsf_true);
    t_destroy();
}

static void test_empty_reopen_ro() {
    if (!remote) {
        t_reset();
        t_create(tsf_true);
        t_close();
        t_open(tsf_false);
        t_destroy();
    }
}

static void test_no_void(tsf_bool_t reopen,
                         tsf_bool_t update) {
    if (!remote) {
        tsf_buffer_t *key;
        
        t_reset();
        t_create(tsf_true);
        
        if (reopen) {
            t_close();
            t_open(update);
        }
        
        TAT(key=tsf_reflect_make_buffer(tsf_reflect_create_void()));
        
        TET(tsf_fsdb_get(file,key), TSF_E_ELE_NOT_FOUND);
        
        tsf_buffer_destroy(key);
        
        t_destroy();
    }
}

static void test_void_to_void(tsf_bool_t reopen,
                              tsf_bool_t update) {
    tsf_buffer_t *key,*val;
    tsf_fsdb_out_t *out;
    tsf_fsdb_in_t *in;
    
    t_reset();
    t_create(tsf_true);
    
    TEST_ASSERT_TSFERROR(
        key=tsf_reflect_make_buffer(tsf_reflect_create_void()));
    TEST_ASSERT_TSFERROR(
        val=tsf_reflect_make_buffer(tsf_reflect_create_void()));
    
    TEST_ASSERT_TSFERROR(out=tsf_fsdb_put(file,key));
    
    TAT(tsf_fsdb_out_write(out,val));
    TAT(tsf_fsdb_out_commit(out));
    
    tsf_buffer_destroy(key);
    tsf_buffer_destroy(val);
    
    if (reopen) {
        t_close();
        t_open(update);
    }
    
    TEST_ASSERT_TSFERROR(
        key=tsf_reflect_make_buffer(tsf_reflect_create_void()));

    TEST_ASSERT_TSFERROR(in=tsf_fsdb_get(file,key));
    
    TAT(val=tsf_fsdb_in_read(in));
    tsf_fsdb_in_close(in);
    
    TEST_ASSERT(tsf_type_get_kind_code(tsf_buffer_get_type(val))
                ==TSF_TK_VOID);
    
    tsf_buffer_destroy(key);
    tsf_buffer_destroy(val);
    
    t_destroy();
}

static void test_void_to_int8(tsf_bool_t reopen,
                              tsf_bool_t update) {
    tsf_buffer_t *key,*val;
    tsf_reflect_t *r;
    tsf_fsdb_out_t *out;
    tsf_fsdb_in_t *in;
    
    t_reset();
    t_create(tsf_true);
    
    TEST_ASSERT_TSFERROR(r=tsf_reflect_create_int8_t(100));
    
    TEST_ASSERT_TSFERROR(
        key=tsf_reflect_make_buffer(tsf_reflect_create_void()));
    TEST_ASSERT_TSFERROR(val=tsf_reflect_make_buffer(r));
    
    TEST_ASSERT_TSFERROR(out=tsf_fsdb_put(file,key));
    
    TAT(tsf_fsdb_out_write(out,val));
    TAT(tsf_fsdb_out_commit(out));
    
    tsf_buffer_destroy(key);
    tsf_buffer_destroy(val);
    tsf_reflect_destroy(r);
    
    if (reopen) {
        t_close();
        t_open(update);
    }
    
    TEST_ASSERT_TSFERROR(
        key=tsf_reflect_make_buffer(tsf_reflect_create_void()));

    TEST_ASSERT_TSFERROR(in=tsf_fsdb_get(file,key));
    
    TAT(val=tsf_fsdb_in_read(in));
    tsf_fsdb_in_close(in);
    
    TEST_ASSERT(tsf_type_get_kind_code(tsf_buffer_get_type(val))
                ==TSF_TK_INT8);
    
    TEST_ASSERT_TSFERROR(r=tsf_reflect_from_buffer(val));
    
    TEST_ASSERT(tsf_type_get_kind_code(tsf_reflect_get_type(val))
                ==TSF_TK_INT8);
    
    TEST_ASSERT(tsf_primitive_reflect_get(r,int8_t)==100);
    
    tsf_buffer_destroy(key);
    tsf_buffer_destroy(val);
    tsf_reflect_destroy(r);
    
    t_destroy();
}

static void test_int8_to_void(int8_t begin,
                              int8_t end,
                              tsf_bool_t reopen,
                              tsf_bool_t update) {
    tsf_buffer_t *key,*val;
    tsf_reflect_t *r;
    int8_t cur;
    tsf_fsdb_out_t *out;
    tsf_fsdb_in_t *in;
    
    t_reset();
    t_create(tsf_true);
    
    for (cur=begin;
         cur<end;
         ++cur) {
        TEST_ASSERT_TSFERROR(r=tsf_reflect_create_int8_t(cur));
        
        TEST_ASSERT_TSFERROR(key=tsf_reflect_make_buffer(r));
        
        TEST_ASSERT_TSFERROR(
            val=tsf_reflect_make_buffer(tsf_reflect_create_void()));
        
        TEST_ASSERT_TSFERROR(out=tsf_fsdb_put(file,key));
        
        TAT(tsf_fsdb_out_write(out,val));
        TAT(tsf_fsdb_out_commit(out));
        
        tsf_buffer_destroy(key);
        tsf_buffer_destroy(val);
        tsf_reflect_destroy(r);
    }
    
    if (reopen) {
        t_close();
        t_open(update);
    }
    
    for (cur=begin;
         cur<end;
         ++cur) {
        TEST_ASSERT_TSFERROR(r=tsf_reflect_create_int8_t(cur));
        
        TEST_ASSERT_TSFERROR(key=tsf_reflect_make_buffer(r));
        
        TEST_ASSERT_TSFERROR(in=tsf_fsdb_get(file,key));
        
        TAT(val=tsf_fsdb_in_read(in));
        tsf_fsdb_in_close(in);
        
        TEST_ASSERT(tsf_type_get_kind_code(tsf_buffer_get_type(val))
                    ==TSF_TK_VOID);
        
        tsf_buffer_destroy(key);
        tsf_buffer_destroy(val);
        tsf_reflect_destroy(r);
    }
    
    t_destroy();
}

static void test_put_n_remove_n(uint32_t n,
                                tsf_bool_t reopen_after_put,
                                tsf_bool_t reopen_after_remove,
                                tsf_bool_t update_after_remove) {
    tsf_buffer_t *key,*val;
    tsf_reflect_t *r;
    int8_t cur;
    tsf_fsdb_out_t *out;
    
    t_reset();
    t_create(tsf_true);
    
    for (cur=n;cur-->0;) {
        TEST_ASSERT_TSFERROR(r=tsf_reflect_create_uint32_t(cur));
        
        TEST_ASSERT_TSFERROR(key=tsf_reflect_make_buffer(r));
        
        TEST_ASSERT_TSFERROR(
            val=tsf_reflect_make_buffer(tsf_reflect_create_void()));
        
        TEST_ASSERT_TSFERROR(out=tsf_fsdb_put(file,key));
        
        TAT(tsf_fsdb_out_write(out,val));
        TAT(tsf_fsdb_out_commit(out));
        
        tsf_buffer_destroy(key);
        tsf_buffer_destroy(val);
        tsf_reflect_destroy(r);
    }
    
    if (reopen_after_put) {
        t_close();
        t_open(tsf_true);
    }
    
    for (cur=n;cur-->0;) {
        TEST_ASSERT_TSFERROR(r=tsf_reflect_create_uint32_t(cur));
        
        TEST_ASSERT_TSFERROR(key=tsf_reflect_make_buffer(r));
        
        TEST_ASSERT_TSFERROR(tsf_fsdb_rm(file,key));

        tsf_buffer_destroy(key);
        tsf_reflect_destroy(r);
    }
    
    if (reopen_after_remove) {
        t_close();
        t_open(update_after_remove);
    }
    
    for (cur=n;cur-->0;) {
        TEST_ASSERT_TSFERROR(r=tsf_reflect_create_uint32_t(cur));
        
        TEST_ASSERT_TSFERROR(key=tsf_reflect_make_buffer(r));
        
        TET(tsf_fsdb_get(file,key), TSF_E_ELE_NOT_FOUND);

        tsf_buffer_destroy(key);
        tsf_reflect_destroy(r);
    }
    
    t_destroy();
}

static void test_put_two_types_remove_first(tsf_bool_t reopen_after_put,
                                            tsf_bool_t reopen_after_remove,
                                            tsf_bool_t update_after_remove) {
    tsf_buffer_t *key,*val;
    tsf_reflect_t *r;
    tsf_fsdb_out_t *out;
    tsf_fsdb_in_t *in;
    
    t_reset();
    t_create(tsf_true);
    
    /* put the first one */
    TEST_ASSERT_TSFERROR(r=tsf_reflect_create_string("first"));
    TEST_ASSERT_TSFERROR(key=tsf_reflect_make_buffer(r));
    tsf_reflect_destroy(r);
    
    TEST_ASSERT_TSFERROR(r=tsf_reflect_create_int8_t(1));
    TEST_ASSERT_TSFERROR(val=tsf_reflect_make_buffer(r));
    tsf_reflect_destroy(r);
    
    TEST_ASSERT_TSFERROR(out=tsf_fsdb_put(file,key));
    TAT(tsf_fsdb_out_write(out,val));
    TAT(tsf_fsdb_out_commit(out));

    tsf_buffer_destroy(key);
    tsf_buffer_destroy(val);
    
    /* put the second one */
    TEST_ASSERT_TSFERROR(r=tsf_reflect_create_string("second"));
    TEST_ASSERT_TSFERROR(key=tsf_reflect_make_buffer(r));
    tsf_reflect_destroy(r);
    
    TEST_ASSERT_TSFERROR(r=tsf_reflect_create_uint8_t(42));
    TEST_ASSERT_TSFERROR(val=tsf_reflect_make_buffer(r));
    tsf_reflect_destroy(r);
    
    TEST_ASSERT_TSFERROR(out=tsf_fsdb_put(file,key));
    TAT(tsf_fsdb_out_write(out,val));
    TAT(tsf_fsdb_out_commit(out));
    tsf_buffer_destroy(key);
    tsf_buffer_destroy(val);
    
    if (reopen_after_put) {
        t_close();
        t_open(tsf_true);
    }
    
    /* remove the first one */
    TEST_ASSERT_TSFERROR(r=tsf_reflect_create_string("first"));
    TEST_ASSERT_TSFERROR(key=tsf_reflect_make_buffer(r));
    tsf_reflect_destroy(r);
    
    TEST_ASSERT_TSFERROR(tsf_fsdb_rm(file,key));
    tsf_buffer_destroy(key);
    
    if (reopen_after_remove) {
        t_close();
        t_open(update_after_remove);
    }
    
    TAT(r=tsf_reflect_create_string("second"));
    TAT(key=tsf_reflect_make_buffer(r));
    tsf_reflect_destroy(r);

    TAT(in=tsf_fsdb_get(file,key));
    TAT(val=tsf_fsdb_in_read(in));
    tsf_fsdb_in_close(in);
    TAT(r=tsf_reflect_from_buffer(val));
    TA(tsf_reflect_get_kind_code(r)==TSF_TK_UINT8);
    TA(tsf_primitive_reflect_get(r,uint8_t)==42);
    
    tsf_reflect_destroy(r);
    tsf_buffer_destroy(val);
    
    t_destroy();
}

static void test_interesting_data(uint32_t num_keys,
                                  uint32_t num_strings,
                                  uint32_t num_arrays) {
    tsf_buffer_t *buf;
    tsf_fsdb_in_t *in;
    tsf_fsdb_out_t *out;
    funKey_t key;
    something_t some;
    someStrings_t strs;
    someArrays_t arrs;
    something_t *somep;
    someStrings_t *strsp;
    someArrays_t *arrsp;
    
    uint32_t i,j,k;
    
    t_reset();
    t_create(tsf_true);
    
    for (i=0;i<num_keys;++i) {
        key.x=i;
        key.y=0;
        key.z=42;
        key.text=tsf_asprintf("foo%x",i);
        
        TAT(buf=funKey__generate(&key));
        TAT(out=tsf_fsdb_put(file,buf));
        tsf_buffer_destroy(buf);
        free(key.text);
        
        some.command.value=something__command__hello;
        TAT(buf=something__generate(&some));
        TAT(tsf_fsdb_out_write(out,buf));
        tsf_buffer_destroy(buf);
        
        for (j=0;j<num_strings;++j) {
            strs.one=tsf_asprintf("%d bar %d",j,i);
            strs.two=tsf_asprintf("baz %d %d",i*2,j*3);
            
            TAT(buf=someStrings__generate(&strs));
            TAT(tsf_fsdb_out_write(out,buf));
            tsf_buffer_destroy(buf);
            
            free(strs.one);
            free(strs.two);
        }
        
        for (j=0;j<num_arrays;++j) {
            int64_t i64a[30];
            char *sa[40];
            int8_t i8a[50];
            
            arrs.one.len=(i+j)%31;
            arrs.one.data=i64a;
            for (k=0;k<arrs.one.len;++k) {
                arrs.one.data[k]=((int64_t)i)*10000+((int64_t)(j+k))*9999999;
            }
            
            arrs.two.len=(i+j)%41;
            arrs.two.data=sa;
            for (k=0;k<arrs.two.len;++k) {
                arrs.two.data[k]=tsf_asprintf("array %d %d %d",
                                              i,j,k);
            }
            
            arrs.thr.len=(i+j)%51;
            arrs.thr.data=i8a;
            for (k=0;k<arrs.thr.len;++k) {
                arrs.thr.data[k]=(int8_t)(i*3+j+k*2);
            }
            
            TAT(buf=someArrays__generate(&arrs));
            TAT(tsf_fsdb_out_write(out,buf));
            tsf_buffer_destroy(buf);
            
            for (k=0;k<arrs.two.len;++k) {
                free(arrs.two.data[k]);
            }
        }
        
        some.command.value=something__command__bye;
        TAT(buf=something__generate(&some));
        TAT(tsf_fsdb_out_write(out,buf));
        tsf_buffer_destroy(buf);
        
        TAT(tsf_fsdb_out_commit(out));
    }
    
    for (i=0;i<num_keys;++i) {
        key.x=i;
        key.y=0;
        key.z=42;
        key.text=tsf_asprintf("foo%x",i);
        
        TAT(buf=funKey__generate(&key));
        TAT(in=tsf_fsdb_get(file,buf));
        tsf_buffer_destroy(buf);
        free(key.text);
        
        TAT(buf=tsf_fsdb_in_read(in));
        TAT(somep=something__parse(buf));
        tsf_buffer_destroy(buf);
        
        TA(somep->command.value==something__command__hello);
        tsf_region_free(somep);
        
        for (j=0;j<num_strings;++j) {
            char *tmpstr;
            
            TAT(buf=tsf_fsdb_in_read(in));
            TAT(strsp=someStrings__parse(buf));
            tsf_buffer_destroy(buf);
            
            tmpstr=tsf_asprintf("%d bar %d",j,i);
            TA(!strcmp(strsp->one,tmpstr));
            free(tmpstr);
            
            tmpstr=tsf_asprintf("baz %d %d",i*2,j*3);
            TA(!strcmp(strsp->two,tmpstr));
            free(tmpstr);
            
            tsf_region_free(strsp);
        }
        
        for (j=0;j<num_arrays;++j) {
            TAT(buf=tsf_fsdb_in_read(in));
            TAT(arrsp=someArrays__parse(buf));
            tsf_buffer_destroy(buf);
            
            TA(arrsp->one.len==(i+j)%31);
            TA(arrsp->two.len==(i+j)%41);
            TA(arrsp->thr.len==(i+j)%51);
            
            for (k=0;k<arrsp->one.len;++k) {
                TA(arrsp->one.data[k]==((int64_t)i)*10000+((int64_t)(j+k))*9999999);
            }
            
            for (k=0;k<arrsp->two.len;++k) {
                char *tmpstr;
                
                tmpstr=tsf_asprintf("array %d %d %d",
                                    i,j,k);
                TA(!strcmp(arrsp->two.data[k],tmpstr));
                free(tmpstr);
            }
            
            for (k=0;k<arrsp->thr.len;++k) {
                TA(arrsp->thr.data[k]==(int8_t)(i*3+j+k*2));
            }
                    
            tsf_region_free(arrsp);
        }
        
        TAT(buf=tsf_fsdb_in_read(in));
        TAT(somep=something__parse(buf));
        tsf_buffer_destroy(buf);
        
        TA(somep->command.value==something__command__bye);
        tsf_region_free(somep);
        
        TET(tsf_fsdb_in_read(in),TSF_E_EOF);
        
        tsf_fsdb_in_close(in);
    }
    
    t_destroy();
}

int main(int c,char **v) {
    if (c==1) {
        /* running in local mode! */
        remote=tsf_false;
    } else if (c==3) {
        /* running in remote mode! */
        hostname=strdup(v[1]);
        if (sscanf(v[2],"%d",&port)!=1) {
            fprintf(stderr,"%s is not a valid port number\n",v[2]);
            return 1;
        }
        remote=tsf_true;
        fprintf(stderr,"Testing remote database %s:%d\n",hostname,port);
    } else {
        fprintf(stderr,"Usage: ./fsdb_test\n");
        fprintf(stderr,"   or: ./fsdb_test <host> <port>\n");
        return 1;
    }
    
    TEST_INIT(c,v);
    
    DO_TEST(test_empty());
    DO_TEST(test_empty_reopen());
    DO_TEST(test_empty_reopen_ro());
    
    DO_TEST(test_no_void(tsf_false,tsf_false));
    DO_TEST(test_no_void(tsf_true,tsf_false));
    DO_TEST(test_no_void(tsf_true,tsf_true));
    
    DO_TEST(test_void_to_void(tsf_false,tsf_false));
    DO_TEST(test_void_to_void(tsf_true,tsf_false));
    DO_TEST(test_void_to_void(tsf_true,tsf_true));
    
    DO_TEST(test_void_to_int8(tsf_false,tsf_false));
    DO_TEST(test_void_to_int8(tsf_true,tsf_false));
    DO_TEST(test_void_to_int8(tsf_true,tsf_true));
    
    DO_TEST(test_int8_to_void(0,1,tsf_false,tsf_false));
    DO_TEST(test_int8_to_void(0,1,tsf_true,tsf_false));
    DO_TEST(test_int8_to_void(0,1,tsf_true,tsf_true));
    
    DO_TEST(test_int8_to_void(0,2,tsf_false,tsf_false));
    DO_TEST(test_int8_to_void(0,2,tsf_true,tsf_false));
    DO_TEST(test_int8_to_void(0,2,tsf_true,tsf_true));
    
    DO_TEST(test_int8_to_void(0,3,tsf_false,tsf_false));
    DO_TEST(test_int8_to_void(0,3,tsf_true,tsf_false));
    DO_TEST(test_int8_to_void(0,3,tsf_true,tsf_true));
    
    DO_TEST(test_put_n_remove_n(1,
                                tsf_false,
                                tsf_false,tsf_false));
    DO_TEST(test_put_n_remove_n(1,
                                tsf_true,
                                tsf_false,tsf_false));
    DO_TEST(test_put_n_remove_n(1,
                                tsf_true,
                                tsf_true,tsf_false));
    DO_TEST(test_put_n_remove_n(1,
                                tsf_true,
                                tsf_true,tsf_true));
    
    DO_TEST(test_put_n_remove_n(2,
                                tsf_false,
                                tsf_false,tsf_false));
    DO_TEST(test_put_n_remove_n(2,
                                tsf_true,
                                tsf_false,tsf_false));
    DO_TEST(test_put_n_remove_n(2,
                                tsf_true,
                                tsf_true,tsf_false));
    DO_TEST(test_put_n_remove_n(2,
                                tsf_true,
                                tsf_true,tsf_true));
    
    DO_TEST(test_put_n_remove_n(10,
                                tsf_false,
                                tsf_false,tsf_false));
    DO_TEST(test_put_n_remove_n(10,
                                tsf_true,
                                tsf_false,tsf_false));
    DO_TEST(test_put_n_remove_n(10,
                                tsf_true,
                                tsf_true,tsf_false));
    DO_TEST(test_put_n_remove_n(10,
                                tsf_true,
                                tsf_true,tsf_true));
    
    DO_TEST(test_put_two_types_remove_first(tsf_false,
                                            tsf_false,tsf_false));
    DO_TEST(test_put_two_types_remove_first(tsf_true,
                                            tsf_false,tsf_false));
    DO_TEST(test_put_two_types_remove_first(tsf_true,
                                            tsf_true,tsf_false));
    DO_TEST(test_put_two_types_remove_first(tsf_true,
                                            tsf_true,tsf_true));
    
    DO_TEST(test_interesting_data(1,0,0));
    DO_TEST(test_interesting_data(1,5,0));
    DO_TEST(test_interesting_data(1,0,5));
    DO_TEST(test_interesting_data(1,5,5));
    DO_TEST(test_interesting_data(5,0,0));
    DO_TEST(test_interesting_data(5,5,0));
    DO_TEST(test_interesting_data(5,0,5));
    DO_TEST(test_interesting_data(5,5,5));
    DO_TEST(test_interesting_data(50,0,0));
    DO_TEST(test_interesting_data(50,50,0));
    DO_TEST(test_interesting_data(50,0,50));
    DO_TEST(test_interesting_data(50,50,50));
    
    TEST_DONE();
    
    return 0;
}


