/*
 * Copyright (C) 2003, 2004, 2005, 2014 Filip Pizlo. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY FILIP PIZLO ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL FILIP PIZLO OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "tsf.h"
#include "gpc.h"
#include "tsf_format.h"
#include "tsf_test.h"

#include <unistd.h>
#include <fcntl.h>
#include <stdarg.h>

static tsf_bool_t use_conversion = tsf_false;

static void *flip(tsf_type_t *dest,
                  tsf_type_t *src,
                  void *data,
                  tsf_bool_t fail) {
    int fd;
    tsf_serial_in_man_t *in_man;
    tsf_serial_out_man_t *out_man;
    char template_name[64];
    tsf_buffer_t *buf;
    tsf_genrtr_t *gen;
    tsf_parser_t *parser;
    tsf_copier_t *copier;
    
    if (use_conversion) {
        copier=tsf_copier_create(dest,src);
        if (fail) {
            TET(copier, TSF_E_TYPE_MISMATCH);
            return NULL;
        } else {
            TEST_ASSERT_TSFERROR(copier);
        }
        
        TEST_ASSERT_TSFERROR(data=tsf_copier_clone(copier,data));
        
        tsf_copier_destroy(copier);
        
        return data;
    }
    
    TAT(gen=tsf_generator_create(src));
    TAT(parser=tsf_parser_create(dest));
    
    strcpy(template_name,TSF_TMP_DIRECTORY "/conv_testXXXXXXXX");
    
    TEST_ASSERT_ERRNO(fd=mkstemp(template_name));
    
    TEST_ASSERT_TSFERROR(
        in_man=tsf_serial_in_man_create(tsf_fd_reader,&fd,NULL));
    TEST_ASSERT_TSFERROR(
        out_man=tsf_serial_out_man_create(tsf_fd_writer,&fd));
    
    TEST_ASSERT_TSFERROR(buf=tsf_generator_generate(gen,data));
    
    TEST_ASSERT_TSFERROR(tsf_serial_out_man_write(out_man,buf));
    tsf_buffer_destroy(buf);
    
    TEST_ASSERT_ERRNO(lseek(fd,SEEK_SET,0)==0);
    
    TEST_ASSERT_TSFERROR(buf=tsf_serial_in_man_read(in_man));
    
    data=tsf_parser_parse(parser,buf);
    if (fail) {
        TET(data, TSF_E_TYPE_MISMATCH);
    } else {
        TEST_ASSERT_TSFERROR(data);
    }
    
    tsf_parser_destroy(parser);
    tsf_buffer_destroy(buf);
    tsf_generator_destroy(gen);
    tsf_serial_in_man_destroy(in_man);
    tsf_serial_out_man_destroy(out_man);
    
    close(fd);
    unlink(template_name);
    
    return data;
}

#define DEFINE_PRIM_TEST_IMPL(type_lc1, type_uc1, type_lc2, type_uc2, input_value, output_value, negated_output_value) \
static void test_##type_lc1##_to_##type_lc2() { \
    tsf_type_t *gen_type;   \
    tsf_type_t *parse_type; \
    type_lc1 data1; \
    type_lc2 *data2;    \
    \
    TEST_ASSERT_TSFERROR(gen_type=tsf_type_create(TSF_TK_##type_uc1));  \
    TEST_ASSERT_TSFERROR(parse_type=tsf_type_create(TSF_TK_##type_uc2));    \
    \
    data1=input_value;    \
    \
    data2=flip(parse_type,gen_type,&data1,tsf_false);    \
    TEST_ASSERT(*data2==output_value); \
    tsf_region_free(data2); \
    \
    if (tsf_type_kind_is_signed(TSF_TK_##type_uc1)) {   \
        data1=-input_value;   \
        \
        data2=flip(parse_type,gen_type,&data1,tsf_false);    \
        TEST_ASSERT(*data2==negated_output_value);    \
        tsf_region_free(data2); \
    }   \
    \
    tsf_type_destroy(gen_type); \
    tsf_type_destroy(parse_type);   \
}

#define DEFINE_PRIM_ARRAY_TEST_IMPL(type_lc1,type_uc1,type_lc2,type_uc2)    \
static void test_array_of_##type_lc1##_to_##type_lc2(uint32_t len) {    \
    tsf_type_t *gen_type;   \
    tsf_type_t *parse_type; \
    tsf_native_array_t data1;   \
    tsf_native_array_t *data2;  \
    uint32_t i; \
    \
    TEST_ASSERT_TSFERROR(gen_type=tsf_type_create_array(    \
        tsf_type_create(TSF_TK_##type_uc1)));   \
    TEST_ASSERT_TSFERROR(parse_type=tsf_type_create_array(  \
        tsf_type_create(TSF_TK_##type_uc2)));   \
    \
    data1.len=len;  \
    TEST_ASSERT_ERRNO(data1.data=malloc(sizeof(type_lc1)*len)); \
    \
    for (i=0;i<len;++i) {   \
        ((type_lc1*)data1.data)[i]=(type_lc1)i; \
    }   \
    \
    data2=flip(parse_type,gen_type,&data1,tsf_false);    \
    \
    free(data1.data);   \
    \
    for (i=0;i<len;++i) {   \
        TEST_ASSERT(((type_lc2*)data2->data)[i]   \
                    ==cast_##type_lc1##_to_##type_lc2(i)); \
    }   \
    \
    tsf_region_free(data2); \
    \
    tsf_type_destroy(gen_type); \
    tsf_type_destroy(parse_type);   \
}

#define DEFINE_PRIM_ARRAY_TO_BITVECTOR_TEST_IMPL(type_lc1,type_uc1)    \
static void test_array_of_##type_lc1##_to_tsf_bool_t(uint32_t len) {    \
    tsf_type_t *gen_type;   \
    tsf_type_t *parse_type; \
    tsf_native_array_t data1;   \
    tsf_native_bitvector_t *data2;  \
    uint32_t i; \
    \
    TEST_ASSERT_TSFERROR(gen_type=tsf_type_create_array(    \
        tsf_type_create(TSF_TK_##type_uc1)));   \
    TEST_ASSERT_TSFERROR(parse_type=tsf_type_create_array(  \
        tsf_type_create(TSF_TK_BIT)));   \
    \
    data1.len=len;  \
    TEST_ASSERT_ERRNO(data1.data=malloc(sizeof(type_lc1)*len)); \
    \
    for (i=0;i<len;++i) {   \
        ((type_lc1*)data1.data)[i]=(type_lc1)(i%3); \
    }   \
    \
    data2=flip(parse_type,gen_type,&data1,tsf_false);    \
    \
    free(data1.data);   \
    \
    for (i=0;i<len;++i) {   \
        TEST_ASSERT((((data2->bits)[i>>3]>>(i&7))&1)   \
                    ==cast_##type_lc1##_to_tsf_bool_t(i%3)); \
    }   \
    \
    tsf_region_free(data2); \
    \
    tsf_type_destroy(gen_type); \
    tsf_type_destroy(parse_type);   \
}

#define DEFINE_BITVECTOR_TO_PRIM_ARRAY_TEST_IMPL(type_lc2,type_uc2)    \
static void test_array_of_tsf_bool_t_to_##type_lc2(uint32_t len) {    \
    tsf_type_t *gen_type;   \
    tsf_type_t *parse_type; \
    tsf_native_bitvector_t data1;   \
    tsf_native_array_t *data2;  \
    uint32_t i; \
    \
    TEST_ASSERT_TSFERROR(gen_type=tsf_type_create_array(    \
        tsf_type_create(TSF_TK_BIT)));   \
    TEST_ASSERT_TSFERROR(parse_type=tsf_type_create_array(  \
        tsf_type_create(TSF_TK_##type_uc2)));   \
    \
    data1.num_bits=len;  \
    TEST_ASSERT_ERRNO(data1.bits=malloc((len+7)>>3)); \
    \
    for (i=0;i<len;++i) {   \
        if ((i%3)) { \
            data1.bits[i>>3]|=(1<<(i&7)); \
        } else { \
            data1.bits[i>>3]&=~(1<<(i&7)); \
        } \
    }   \
    \
    data2=flip(parse_type,gen_type,&data1,tsf_false);    \
    \
    free(data1.bits);   \
    \
    for (i=0;i<len;++i) {   \
        TEST_ASSERT(((type_lc2*)data2->data)[i]   \
                    ==cast_tsf_bool_t_to_##type_lc2((i%3)?tsf_true:tsf_false)); \
    }   \
    \
    tsf_region_free(data2); \
    \
    tsf_type_destroy(gen_type); \
    tsf_type_destroy(parse_type);   \
}

#define DEFINE_PRIM_TEST(type_lc1, type_uc1, type_lc2, type_uc2, input_value, output_value, negated_output_value) \
static type_lc2 cast_##type_lc1##_to_##type_lc2(type_lc1 x) { \
    return (type_lc2)x; \
} \
DEFINE_PRIM_TEST_IMPL(type_lc1, type_uc1, type_lc2, type_uc2, input_value, output_value, negated_output_value) \
DEFINE_PRIM_ARRAY_TEST_IMPL(type_lc1,type_uc1,type_lc2,type_uc2)

DEFINE_PRIM_TEST(int8_t,INT8,int8_t,INT8,5,5,-5)
DEFINE_PRIM_TEST(int8_t,INT8,uint8_t,UINT8,5,5,251)
DEFINE_PRIM_TEST(int8_t,INT8,int16_t,INT16,5,5,-5)
DEFINE_PRIM_TEST(int8_t,INT8,uint16_t,UINT16,5,5,65531)
DEFINE_PRIM_TEST(int8_t,INT8,int32_t,INT32,5,5,-5)
DEFINE_PRIM_TEST(int8_t,INT8,uint32_t,UINT32,5,5,UINT32_C(4294967291))
DEFINE_PRIM_TEST(int8_t,INT8,int64_t,INT64,5,5,-5)
DEFINE_PRIM_TEST(int8_t,INT8,uint64_t,UINT64,5,5,UINT64_C(18446744073709551611))
DEFINE_PRIM_TEST(int8_t,INT8,tsf_integer_t,INTEGER,5,5,-5)
DEFINE_PRIM_TEST(int8_t,INT8,tsf_long_t,LONG,5,5,-5)
DEFINE_PRIM_TEST(int8_t,INT8,float,FLOAT,5,5,-5)
DEFINE_PRIM_TEST(int8_t,INT8,double,DOUBLE,5,5,-5)

DEFINE_PRIM_TEST(uint8_t,UINT8,int8_t,INT8,5,5,0)
DEFINE_PRIM_TEST(uint8_t,UINT8,uint8_t,UINT8,5,5,0)
DEFINE_PRIM_TEST(uint8_t,UINT8,int16_t,INT16,5,5,0)
DEFINE_PRIM_TEST(uint8_t,UINT8,uint16_t,UINT16,5,5,0)
DEFINE_PRIM_TEST(uint8_t,UINT8,int32_t,INT32,5,5,0)
DEFINE_PRIM_TEST(uint8_t,UINT8,uint32_t,UINT32,5,5,0)
DEFINE_PRIM_TEST(uint8_t,UINT8,int64_t,INT64,5,5,0)
DEFINE_PRIM_TEST(uint8_t,UINT8,uint64_t,UINT64,5,5,0)
DEFINE_PRIM_TEST(uint8_t,UINT8,tsf_integer_t,INTEGER,5,5,0)
DEFINE_PRIM_TEST(uint8_t,UINT8,tsf_long_t,LONG,5,5,0)
DEFINE_PRIM_TEST(uint8_t,UINT8,float,FLOAT,5,5,0)
DEFINE_PRIM_TEST(uint8_t,UINT8,double,DOUBLE,5,5,0)

DEFINE_PRIM_TEST(int16_t,INT16,int8_t,INT8,5000,-120,120)
DEFINE_PRIM_TEST(int16_t,INT16,uint8_t,UINT8,5000,136,120)
DEFINE_PRIM_TEST(int16_t,INT16,int16_t,INT16,5000,5000,-5000)
DEFINE_PRIM_TEST(int16_t,INT16,uint16_t,UINT16,5000,5000,60536)
DEFINE_PRIM_TEST(int16_t,INT16,int32_t,INT32,5000,5000,-5000)
DEFINE_PRIM_TEST(int16_t,INT16,uint32_t,UINT32,5000,5000,UINT32_C(4294962296))
DEFINE_PRIM_TEST(int16_t,INT16,int64_t,INT64,5000,5000,-5000)
DEFINE_PRIM_TEST(int16_t,INT16,uint64_t,UINT64,5000,5000,UINT64_C(18446744073709546616))
DEFINE_PRIM_TEST(int16_t,INT16,tsf_integer_t,INTEGER,5000,5000,-5000)
DEFINE_PRIM_TEST(int16_t,INT16,tsf_long_t,LONG,5000,5000,-5000)
DEFINE_PRIM_TEST(int16_t,INT16,float,FLOAT,5000,5000,-5000)
DEFINE_PRIM_TEST(int16_t,INT16,double,DOUBLE,5000,5000,-5000)

DEFINE_PRIM_TEST(uint16_t,UINT16,int8_t,INT8,5000,-120,0)
DEFINE_PRIM_TEST(uint16_t,UINT16,uint8_t,UINT8,5000,136,0)
DEFINE_PRIM_TEST(uint16_t,UINT16,int16_t,INT16,5000,5000,0)
DEFINE_PRIM_TEST(uint16_t,UINT16,uint16_t,UINT16,5000,5000,0)
DEFINE_PRIM_TEST(uint16_t,UINT16,int32_t,INT32,5000,5000,0)
DEFINE_PRIM_TEST(uint16_t,UINT16,uint32_t,UINT32,5000,5000,0)
DEFINE_PRIM_TEST(uint16_t,UINT16,int64_t,INT64,5000,5000,0)
DEFINE_PRIM_TEST(uint16_t,UINT16,uint64_t,UINT64,5000,5000,0)
DEFINE_PRIM_TEST(uint16_t,UINT16,tsf_integer_t,INTEGER,5000,5000,0)
DEFINE_PRIM_TEST(uint16_t,UINT16,tsf_long_t,LONG,5000,5000,0)
DEFINE_PRIM_TEST(uint16_t,UINT16,float,FLOAT,5000,5000,0)
DEFINE_PRIM_TEST(uint16_t,UINT16,double,DOUBLE,5000,5000,0)

DEFINE_PRIM_TEST(int32_t,INT32,int8_t,INT8,5000000,64,-64)
DEFINE_PRIM_TEST(int32_t,INT32,uint8_t,UINT8,5000000,64,192)
DEFINE_PRIM_TEST(int32_t,INT32,int16_t,INT16,5000000,19264,-19264)
DEFINE_PRIM_TEST(int32_t,INT32,uint16_t,UINT16,5000000,19264,46272)
DEFINE_PRIM_TEST(int32_t,INT32,int32_t,INT32,5000000,5000000,-5000000)
DEFINE_PRIM_TEST(int32_t,INT32,uint32_t,UINT32,5000000,5000000,UINT32_C(4289967296))
DEFINE_PRIM_TEST(int32_t,INT32,int64_t,INT64,5000000,5000000,-5000000)
DEFINE_PRIM_TEST(int32_t,INT32,uint64_t,UINT64,5000000,5000000,UINT64_C(18446744073704551616))
DEFINE_PRIM_TEST(int32_t,INT32,tsf_integer_t,INTEGER,5000000,5000000,-5000000)
DEFINE_PRIM_TEST(int32_t,INT32,tsf_long_t,LONG,5000000,5000000,-5000000)
DEFINE_PRIM_TEST(int32_t,INT32,float,FLOAT,5000000,5000000,-5000000)
DEFINE_PRIM_TEST(int32_t,INT32,double,DOUBLE,5000000,5000000,-5000000)

DEFINE_PRIM_TEST(uint32_t,UINT32,int8_t,INT8,5000000,64,0)
DEFINE_PRIM_TEST(uint32_t,UINT32,uint8_t,UINT8,5000000,64,0)
DEFINE_PRIM_TEST(uint32_t,UINT32,int16_t,INT16,5000000,19264,0)
DEFINE_PRIM_TEST(uint32_t,UINT32,uint16_t,UINT16,5000000,19264,0)
DEFINE_PRIM_TEST(uint32_t,UINT32,int32_t,INT32,5000000,5000000,0)
DEFINE_PRIM_TEST(uint32_t,UINT32,uint32_t,UINT32,5000000,5000000,0)
DEFINE_PRIM_TEST(uint32_t,UINT32,int64_t,INT64,5000000,5000000,0)
DEFINE_PRIM_TEST(uint32_t,UINT32,uint64_t,UINT64,5000000,5000000,0)
DEFINE_PRIM_TEST(uint32_t,UINT32,tsf_integer_t,INTEGER,5000000,5000000,0)
DEFINE_PRIM_TEST(uint32_t,UINT32,tsf_long_t,LONG,5000000,5000000,0)
DEFINE_PRIM_TEST(uint32_t,UINT32,float,FLOAT,5000000,5000000,0)
DEFINE_PRIM_TEST(uint32_t,UINT32,double,DOUBLE,5000000,5000000,0)

DEFINE_PRIM_TEST(int64_t,INT64,int8_t,INT8,INT64_C(5000000001),1,-1)
DEFINE_PRIM_TEST(int64_t,INT64,uint8_t,UINT8,INT64_C(5000000001),1,255)
DEFINE_PRIM_TEST(int64_t,INT64,int16_t,INT16,INT64_C(5000000001),-3583,3583)
DEFINE_PRIM_TEST(int64_t,INT64,uint16_t,UINT16,INT64_C(5000000001),61953,3583)
DEFINE_PRIM_TEST(int64_t,INT64,int32_t,INT32,INT64_C(5000000001),705032705,-705032705)
DEFINE_PRIM_TEST(int64_t,INT64,uint32_t,UINT32,INT64_C(5000000001),705032705,UINT32_C(3589934591))
DEFINE_PRIM_TEST(int64_t,INT64,int64_t,INT64,INT64_C(5000000001),INT64_C(5000000001),INT64_C(-5000000001))
DEFINE_PRIM_TEST(int64_t,INT64,uint64_t,UINT64,INT64_C(5000000001),UINT64_C(5000000001),UINT64_C(18446744068709551615))
DEFINE_PRIM_TEST(int64_t,INT64,tsf_integer_t,INTEGER,INT64_C(5000000001),705032705,-705032705)
DEFINE_PRIM_TEST(int64_t,INT64,tsf_long_t,LONG,INT64_C(5000000001),INT64_C(5000000001),INT64_C(-5000000001))
DEFINE_PRIM_TEST(int64_t,INT64,float,FLOAT,INT64_C(5000000001),5000000000.f,-5000000000.f)
DEFINE_PRIM_TEST(int64_t,INT64,double,DOUBLE,INT64_C(5000000001),5000000001.,-5000000001.)

DEFINE_PRIM_TEST(uint64_t,UINT64,int8_t,INT8,UINT64_C(5000000003),3,0)
DEFINE_PRIM_TEST(uint64_t,UINT64,uint8_t,UINT8,UINT64_C(5000000003),3,0)
DEFINE_PRIM_TEST(uint64_t,UINT64,int16_t,INT16,UINT64_C(5000000003),-3581,0)
DEFINE_PRIM_TEST(uint64_t,UINT64,uint16_t,UINT16,UINT64_C(5000000003),61955,0)
DEFINE_PRIM_TEST(uint64_t,UINT64,int32_t,INT32,UINT64_C(5000000003),705032707,0)
DEFINE_PRIM_TEST(uint64_t,UINT64,uint32_t,UINT32,UINT64_C(5000000003),705032707,0)
DEFINE_PRIM_TEST(uint64_t,UINT64,int64_t,INT64,UINT64_C(5000000003),5000000003,0)
DEFINE_PRIM_TEST(uint64_t,UINT64,uint64_t,UINT64,UINT64_C(5000000003),5000000003,0)
DEFINE_PRIM_TEST(uint64_t,UINT64,tsf_integer_t,INTEGER,UINT64_C(5000000003),705032707,0)
DEFINE_PRIM_TEST(uint64_t,UINT64,tsf_long_t,LONG,UINT64_C(5000000003),5000000003,0)
DEFINE_PRIM_TEST(uint64_t,UINT64,float,FLOAT,UINT64_C(5000000003),5000000000.f,0)
DEFINE_PRIM_TEST(uint64_t,UINT64,double,DOUBLE,UINT64_C(5000000003),5000000003,0)

DEFINE_PRIM_TEST(tsf_integer_t,INTEGER,int8_t,INT8,5000000,64,-64)
DEFINE_PRIM_TEST(tsf_integer_t,INTEGER,uint8_t,UINT8,5000000,64,192)
DEFINE_PRIM_TEST(tsf_integer_t,INTEGER,int16_t,INT16,5000000,19264,-19264)
DEFINE_PRIM_TEST(tsf_integer_t,INTEGER,uint16_t,UINT16,5000000,19264,46272)
DEFINE_PRIM_TEST(tsf_integer_t,INTEGER,int32_t,INT32,5000000,5000000,-5000000)
DEFINE_PRIM_TEST(tsf_integer_t,INTEGER,uint32_t,UINT32,5000000,5000000,UINT32_C(4289967296))
DEFINE_PRIM_TEST(tsf_integer_t,INTEGER,int64_t,INT64,5000000,5000000,-5000000)
DEFINE_PRIM_TEST(tsf_integer_t,INTEGER,uint64_t,UINT64,5000000,5000000,UINT64_C(18446744073704551616))
DEFINE_PRIM_TEST(tsf_integer_t,INTEGER,tsf_integer_t,INTEGER,5000000,5000000,-5000000)
DEFINE_PRIM_TEST(tsf_integer_t,INTEGER,tsf_long_t,LONG,5000000,5000000,-5000000)
DEFINE_PRIM_TEST(tsf_integer_t,INTEGER,float,FLOAT,5000000,5000000,-5000000)
DEFINE_PRIM_TEST(tsf_integer_t,INTEGER,double,DOUBLE,5000000,5000000,-5000000)

DEFINE_PRIM_TEST(tsf_long_t,LONG,int8_t,INT8,INT64_C(5000000001),1,-1)
DEFINE_PRIM_TEST(tsf_long_t,LONG,uint8_t,UINT8,INT64_C(5000000001),1,255)
DEFINE_PRIM_TEST(tsf_long_t,LONG,int16_t,INT16,INT64_C(5000000001),-3583,3583)
DEFINE_PRIM_TEST(tsf_long_t,LONG,uint16_t,UINT16,INT64_C(5000000001),61953,3583)
DEFINE_PRIM_TEST(tsf_long_t,LONG,int32_t,INT32,INT64_C(5000000001),705032705,-705032705)
DEFINE_PRIM_TEST(tsf_long_t,LONG,uint32_t,UINT32,INT64_C(5000000001),705032705,UINT32_C(3589934591))
DEFINE_PRIM_TEST(tsf_long_t,LONG,int64_t,INT64,INT64_C(5000000001),INT64_C(5000000001),INT64_C(-5000000001))
DEFINE_PRIM_TEST(tsf_long_t,LONG,uint64_t,UINT64,INT64_C(5000000001),UINT64_C(5000000001),UINT64_C(18446744068709551615))
DEFINE_PRIM_TEST(tsf_long_t,LONG,tsf_integer_t,INTEGER,INT64_C(5000000001),705032705,-705032705)
DEFINE_PRIM_TEST(tsf_long_t,LONG,tsf_long_t,LONG,INT64_C(5000000001),INT64_C(5000000001),INT64_C(-5000000001))
DEFINE_PRIM_TEST(tsf_long_t,LONG,float,FLOAT,INT64_C(5000000001),5000000000.f,-5000000000.f)
DEFINE_PRIM_TEST(tsf_long_t,LONG,double,DOUBLE,INT64_C(5000000001),5000000001.,-5000000001.)

DEFINE_PRIM_TEST(float,FLOAT,int8_t,INT8,1234.5f,-46,46)
DEFINE_PRIM_TEST(float,FLOAT,uint8_t,UINT8,1234.5f,210,46)
DEFINE_PRIM_TEST(float,FLOAT,int16_t,INT16,1234.5f,1234,-1234)
DEFINE_PRIM_TEST(float,FLOAT,uint16_t,UINT16,1234.5f,1234,64302)
DEFINE_PRIM_TEST(float,FLOAT,int32_t,INT32,1234.5f,1234,-1234)
DEFINE_PRIM_TEST(float,FLOAT,uint32_t,UINT32,1234.5f,1234,UINT32_C(4294966062))
DEFINE_PRIM_TEST(float,FLOAT,int64_t,INT64,1234.5f,1234,-1234)
DEFINE_PRIM_TEST(float,FLOAT,uint64_t,UINT64,1234.5f,1234,UINT64_C(18446744073709550382))
DEFINE_PRIM_TEST(float,FLOAT,tsf_integer_t,INTEGER,1234.5f,1234,-1234)
DEFINE_PRIM_TEST(float,FLOAT,tsf_long_t,LONG,1234.5f,1234,-1234)
DEFINE_PRIM_TEST(float,FLOAT,float,FLOAT,1234.5f,1234.5f,-1234.5f)
DEFINE_PRIM_TEST(float,FLOAT,double,DOUBLE,1234.5f,1234.5,-1234.5)

DEFINE_PRIM_TEST(double,DOUBLE,int8_t,INT8,5000000042.9,42,-42)
DEFINE_PRIM_TEST(double,DOUBLE,uint8_t,UINT8,5000000042.9,42,214)
DEFINE_PRIM_TEST(double,DOUBLE,int16_t,INT16,5000000042.9,-3542,3542)
DEFINE_PRIM_TEST(double,DOUBLE,uint16_t,UINT16,5000000042.9,61994,3542)
DEFINE_PRIM_TEST(double,DOUBLE,int32_t,INT32,5000000042.9,705032746,-705032746)
DEFINE_PRIM_TEST(double,DOUBLE,uint32_t,UINT32,5000000042.9,705032746,UINT32_C(3589934550))
DEFINE_PRIM_TEST(double,DOUBLE,int64_t,INT64,5000000042.9,INT64_C(5000000042),INT64_C(-5000000042))
DEFINE_PRIM_TEST(double,DOUBLE,uint64_t,UINT64,5000000042.9,UINT64_C(5000000042),UINT64_C(18446744068709551574))
DEFINE_PRIM_TEST(double,DOUBLE,tsf_integer_t,INTEGER,5000000042.9,705032746,-705032746)
DEFINE_PRIM_TEST(double,DOUBLE,tsf_long_t,LONG,5000000042.9,INT64_C(5000000042),INT64_C(-5000000042))
DEFINE_PRIM_TEST(double,DOUBLE,float,FLOAT,5000000042.9,5000000000.f,-5000000000.f)
DEFINE_PRIM_TEST(double,DOUBLE,double,DOUBLE,5000000042.9,5000000042.9,-5000000042.9)

#define DEFINE_PRIM_TO_BOOL_TEST(type_lc1,type_uc1,value) \
static tsf_bool_t cast_##type_lc1##_to_tsf_bool_t(type_lc1 x) { \
    return x?tsf_true:tsf_false;\
} \
DEFINE_PRIM_TEST_IMPL(type_lc1,type_uc1,tsf_bool_t,BIT,value,cast_##type_lc1##_to_tsf_bool_t(value),cast_##type_lc1##_to_tsf_bool_t(-value)) \
DEFINE_PRIM_ARRAY_TO_BITVECTOR_TEST_IMPL(type_lc1,type_uc1)

#if 0
     ; /* stupid emacs */
#endif

DEFINE_PRIM_TO_BOOL_TEST(int8_t,INT8,1)
DEFINE_PRIM_TO_BOOL_TEST(uint8_t,UINT8,1)
DEFINE_PRIM_TO_BOOL_TEST(int16_t,INT16,1)
DEFINE_PRIM_TO_BOOL_TEST(uint16_t,UINT16,1)
DEFINE_PRIM_TO_BOOL_TEST(int32_t,INT32,1)
DEFINE_PRIM_TO_BOOL_TEST(uint32_t,UINT32,1)
DEFINE_PRIM_TO_BOOL_TEST(int64_t,INT64,1)
DEFINE_PRIM_TO_BOOL_TEST(uint64_t,UINT64,1)
DEFINE_PRIM_TO_BOOL_TEST(tsf_integer_t,INTEGER,1)
DEFINE_PRIM_TO_BOOL_TEST(tsf_long_t,LONG,1)
DEFINE_PRIM_TO_BOOL_TEST(float,FLOAT,1)
DEFINE_PRIM_TO_BOOL_TEST(double,DOUBLE,1)
			      
#define DEFINE_BOOL_TO_PRIM_TEST(type_lc2,type_uc2,value) \
static tsf_bool_t cast_tsf_bool_t_to_##type_lc2(tsf_bool_t x) { \
    return x?(type_lc2)1:(type_lc2)0;\
} \
DEFINE_PRIM_TEST_IMPL(tsf_bool_t, BIT, type_lc2, type_uc2, value, cast_tsf_bool_t_to_##type_lc2(value), 0) \
DEFINE_BITVECTOR_TO_PRIM_ARRAY_TEST_IMPL(type_lc2,type_uc2)

DEFINE_BOOL_TO_PRIM_TEST(int8_t,INT8,1)
DEFINE_BOOL_TO_PRIM_TEST(uint8_t,UINT8,1)
DEFINE_BOOL_TO_PRIM_TEST(int16_t,INT16,1)
DEFINE_BOOL_TO_PRIM_TEST(uint16_t,UINT16,1)
DEFINE_BOOL_TO_PRIM_TEST(int32_t,INT32,1)
DEFINE_BOOL_TO_PRIM_TEST(uint32_t,UINT32,1)
DEFINE_BOOL_TO_PRIM_TEST(int64_t,INT64,1)
DEFINE_BOOL_TO_PRIM_TEST(uint64_t,UINT64,1)
DEFINE_BOOL_TO_PRIM_TEST(tsf_integer_t,INTEGER,1)
DEFINE_BOOL_TO_PRIM_TEST(tsf_long_t,LONG,1)
DEFINE_BOOL_TO_PRIM_TEST(float,FLOAT,1)
DEFINE_BOOL_TO_PRIM_TEST(double,DOUBLE,1)

#if 0
     } /* stupid emacs */
#endif
     
static void test_tsf_bool_t_to_tsf_bool_t() {}
static void test_array_of_tsf_bool_t_to_tsf_bool_t() {}
			      
typedef struct {
    float w;
    float x;
    float y;
    float z;
} my_primitive_struct_t;

static void test_struct_succeed() {
    tsf_type_t *stt1,*stt2;
    my_primitive_struct_t data1,*data2;
    
    TEST_ASSERT_TSFERROR(stt1=tsf_type_create(TSF_TK_STRUCT));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_set_size(&stt1,sizeof(my_primitive_struct_t)));

    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(&stt1,"w",
            tsf_type_create(TSF_TK_FLOAT)));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(&stt1,"w",
            tsf_offsetof(my_primitive_struct_t,w)));

    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(&stt1,"x",
            tsf_type_create(TSF_TK_FLOAT)));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(&stt1,"x",
            tsf_offsetof(my_primitive_struct_t,x)));

    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(&stt1,"y",
            tsf_type_create(TSF_TK_FLOAT)));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(&stt1,"y",
            tsf_offsetof(my_primitive_struct_t,y)));

    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(&stt1,"z",
            tsf_type_create(TSF_TK_FLOAT)));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(&stt1,"z",
            tsf_offsetof(my_primitive_struct_t,z)));

    TEST_ASSERT_TSFERROR(stt2=tsf_type_create(TSF_TK_STRUCT));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_set_size(&stt2,sizeof(my_primitive_struct_t)));

    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(&stt2,"x",
            tsf_type_create(TSF_TK_FLOAT)));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(&stt2,"x",
            tsf_offsetof(my_primitive_struct_t,x)));

    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(&stt2,"y",
            tsf_type_create(TSF_TK_FLOAT)));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(&stt2,"y",
            tsf_offsetof(my_primitive_struct_t,y)));

    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(&stt2,"z",
            tsf_type_create(TSF_TK_FLOAT)));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(&stt2,"z",
            tsf_offsetof(my_primitive_struct_t,z)));

    data1.w=1;
    data1.x=2;
    data1.y=3;
    data1.z=4;
    
    data2=flip(stt2,stt1,&data1,tsf_false);

    TEST_ASSERT(data2->x==2);
    TEST_ASSERT(data2->y==3);
    TEST_ASSERT(data2->z==4);
    
    tsf_region_free(data2);
    tsf_type_destroy(stt1);
    tsf_type_destroy(stt2);
}

static void test_struct_fail() {
    tsf_type_t *stt1,*stt2;
    my_primitive_struct_t data1;
    
    TEST_ASSERT_TSFERROR(stt1=tsf_type_create(TSF_TK_STRUCT));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_set_size(&stt1,sizeof(my_primitive_struct_t)));

    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(&stt1,"x",
            tsf_type_create(TSF_TK_FLOAT)));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(&stt1,"x",
            tsf_offsetof(my_primitive_struct_t,x)));

    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(&stt1,"y",
            tsf_type_create(TSF_TK_FLOAT)));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(&stt1,"y",
            tsf_offsetof(my_primitive_struct_t,y)));

    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(&stt1,"z",
            tsf_type_create(TSF_TK_FLOAT)));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(&stt1,"z",
            tsf_offsetof(my_primitive_struct_t,z)));

    TEST_ASSERT_TSFERROR(stt2=tsf_type_create(TSF_TK_STRUCT));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_set_size(&stt2,sizeof(my_primitive_struct_t)));

    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(&stt2,"w",
            tsf_type_create(TSF_TK_FLOAT)));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(&stt2,"w",
            tsf_offsetof(my_primitive_struct_t,w)));

    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(&stt2,"x",
            tsf_type_create(TSF_TK_FLOAT)));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(&stt2,"x",
            tsf_offsetof(my_primitive_struct_t,x)));

    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(&stt2,"y",
            tsf_type_create(TSF_TK_FLOAT)));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(&stt2,"y",
            tsf_offsetof(my_primitive_struct_t,y)));

    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(&stt2,"z",
            tsf_type_create(TSF_TK_FLOAT)));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(&stt2,"z",
            tsf_offsetof(my_primitive_struct_t,z)));

    data1.x=2;
    data1.y=3;
    data1.z=4;

    TEST_ASSERT(flip(stt2,stt1,&data1,tsf_true)==NULL);

    tsf_type_destroy(stt1);
    tsf_type_destroy(stt2);
}

typedef struct {
    float z;
    float y;
    float x;
    float w;
} my_primitive_reverse_struct_t;

static void test_struct_reverse() {
    tsf_type_t *stt1,*stt2;
    my_primitive_struct_t data1;
    my_primitive_reverse_struct_t *data2;
    
    TEST_ASSERT_TSFERROR(stt1=tsf_type_create(TSF_TK_STRUCT));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_set_size(&stt1,sizeof(my_primitive_struct_t)));

    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(&stt1,"w",
            tsf_type_create(TSF_TK_FLOAT)));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(&stt1,"w",
            tsf_offsetof(my_primitive_struct_t,w)));

    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(&stt1,"x",
            tsf_type_create(TSF_TK_FLOAT)));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(&stt1,"x",
            tsf_offsetof(my_primitive_struct_t,x)));

    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(&stt1,"y",
            tsf_type_create(TSF_TK_FLOAT)));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(&stt1,"y",
            tsf_offsetof(my_primitive_struct_t,y)));

    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(&stt1,"z",
            tsf_type_create(TSF_TK_FLOAT)));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(&stt1,"z",
            tsf_offsetof(my_primitive_struct_t,z)));

    TEST_ASSERT_TSFERROR(stt2=tsf_type_create(TSF_TK_STRUCT));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_set_size(&stt2,
            sizeof(my_primitive_reverse_struct_t)));

    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(&stt2,"z",
            tsf_type_create(TSF_TK_FLOAT)));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(&stt2,"z",
            tsf_offsetof(my_primitive_reverse_struct_t,z)));

    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(&stt2,"y",
            tsf_type_create(TSF_TK_FLOAT)));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(&stt2,"y",
            tsf_offsetof(my_primitive_reverse_struct_t,y)));

    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(&stt2,"x",
            tsf_type_create(TSF_TK_FLOAT)));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(&stt2,"x",
            tsf_offsetof(my_primitive_reverse_struct_t,x)));

    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(&stt2,"w",
            tsf_type_create(TSF_TK_FLOAT)));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(&stt2,"w",
            tsf_offsetof(my_primitive_reverse_struct_t,w)));

    data1.w=1;
    data1.x=2;
    data1.y=3;
    data1.z=4;
    
    data2=flip(stt2,stt1,&data1,tsf_false);
    
    TEST_ASSERT(data2->w==1);
    TEST_ASSERT(data2->x==2);
    TEST_ASSERT(data2->y==3);
    TEST_ASSERT(data2->z==4);
    
    tsf_region_free(data2);
    tsf_type_destroy(stt1);
    tsf_type_destroy(stt2);
}

/* 2) need more struct tests.  for example: bit skipping, array
 *    skipping (with multiple levels of nesting), static-size
 *    struct (with no arrays) skipping, dynamic-size struct (with
 *    arrays) skipping, and skipping of every type of primitive.
 */

typedef struct {
    float a;
    float b;
    
    /* wanna make this structure native aligned. */
#if (((NATIVE_ALIGNMENT- \
       (SIZEOF_FLOAT*2))& \
      (NATIVE_ALIGNMENT-1)) != 0)
    char padding[(TSF_NATIVE_ALIGNMENT-
                  (sizeof(float)*2))&
                 (TSF_NATIVE_ALIGNMENT-1)];
#endif
} my_skipping_struct_t;

/* returns the struct size of the type stuff added */
typedef size_t (*type_create_t)(tsf_type_t **stt,
				uint32_t offset,
				void *arg);

typedef void (*verify_data_zero_t)(void *data,
				   void *arg);

static void test_option_field(type_create_t creator,
			      verify_data_zero_t verifier,
			      void *arg) {
    tsf_type_t *stt1,*stt2;
    my_skipping_struct_t *data1,*data2;
    uint32_t gen_size=sizeof(my_skipping_struct_t);
    
    TEST_ASSERT_TSFERROR(stt1=tsf_type_create(TSF_TK_STRUCT));

    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(&stt1,"a",
            tsf_type_create(TSF_TK_FLOAT)));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(&stt1,"a",
            tsf_offsetof(my_skipping_struct_t,a)));

    gen_size+=creator(&stt1,sizeof(my_skipping_struct_t),arg);

    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(&stt1,"b",
            tsf_type_create(TSF_TK_FLOAT)));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(&stt1,"b",
            tsf_offsetof(my_skipping_struct_t,b)));

    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_set_size(&stt1,gen_size));
    
    TEST_ASSERT_TSFERROR(stt2=tsf_type_create(TSF_TK_STRUCT));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_set_size(&stt2,sizeof(my_skipping_struct_t)));

    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(&stt2,"a",
            tsf_type_create(TSF_TK_FLOAT)));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(&stt2,"a",
            tsf_offsetof(my_skipping_struct_t,a)));

    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(&stt2,"b",
            tsf_type_create(TSF_TK_FLOAT)));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(&stt2,"b",
            tsf_offsetof(my_skipping_struct_t,b)));
    
    TEST_ASSERT_ERRNO(data1=malloc(sizeof(my_skipping_struct_t)));
    
    data1->a=2;
    data1->b=4;
    
    data2=flip(stt1,stt2,data1,tsf_false);

    verifier(data2+1,arg);

    free(data1);
    
    TEST_ASSERT(data2->a==2);
    TEST_ASSERT(data2->b==4);
    
    tsf_region_free(data2);
    tsf_type_destroy(stt1);
    tsf_type_destroy(stt2);
}

typedef void (*skipping_data_write_t)(void *data,
                                      void *arg);

typedef void (*skipping_data_destroy_t)(void *data,
                                        void *arg);

static void test_skipping_struct(type_create_t creator,
                                 skipping_data_write_t writer,
                                 skipping_data_destroy_t destroyer,
                                 void *arg) {
    tsf_type_t *stt1,*stt2;
    my_skipping_struct_t *data1,*data2;
    uint32_t gen_size=sizeof(my_skipping_struct_t);
    
    TEST_ASSERT_TSFERROR(stt1=tsf_type_create(TSF_TK_STRUCT));

    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(&stt1,"a",
            tsf_type_create(TSF_TK_FLOAT)));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(&stt1,"a",
            tsf_offsetof(my_skipping_struct_t,a)));

    gen_size+=creator(&stt1,sizeof(my_skipping_struct_t),arg);

    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(&stt1,"b",
            tsf_type_create(TSF_TK_FLOAT)));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(&stt1,"b",
            tsf_offsetof(my_skipping_struct_t,b)));

    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_set_size(&stt1,gen_size));
    
    TEST_ASSERT_TSFERROR(stt2=tsf_type_create(TSF_TK_STRUCT));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_set_size(&stt2,sizeof(my_skipping_struct_t)));

    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(&stt2,"a",
            tsf_type_create(TSF_TK_FLOAT)));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(&stt2,"a",
            tsf_offsetof(my_skipping_struct_t,a)));

    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(&stt2,"b",
            tsf_type_create(TSF_TK_FLOAT)));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(&stt2,"b",
            tsf_offsetof(my_skipping_struct_t,b)));
    
    TEST_ASSERT_ERRNO(data1=malloc(gen_size));
    
    data1->a=2;
    data1->b=4;
    
    writer(data1+1,arg);
    
    data2=flip(stt2,stt1,data1,tsf_false);

    destroyer(data1+1,arg);
    free(data1);
    
    TEST_ASSERT(data2->a==2);
    TEST_ASSERT(data2->b==4);
    
    tsf_region_free(data2);
    tsf_type_destroy(stt1);
    tsf_type_destroy(stt2);
}

/* bit skipping */

static size_t bit_creator(tsf_type_t **stt,
                          uint32_t offset,
                          void *arg) {
    char name[64];
    uint32_t i;
    uint32_t num_bits=*((uint32_t*)arg);
    
    for (i=0;i<num_bits;++i) {
        sprintf(name,fui32,i);
        TAT(tsf_struct_type_append(stt,name,
                tsf_type_create(TSF_TK_BIT)));
	TAT(tsf_struct_type_set_optional(stt,name,tsf_true));
        TAT(tsf_native_struct_type_map(stt,name,
                offset+i*sizeof(tsf_bool_t)));
    }
    
    return sizeof(tsf_bool_t)*num_bits;
}

static void bit_writer(void *data,
                       void *arg) {
}

static void bit_destroyer(void *data,
                          void *arg) {
}

static void bit_verifier(void *data,
			 void *arg) {
    uint32_t num_bits=*((uint32_t*)arg);
    tsf_bool_t *cur=(tsf_bool_t*)data;
    while (num_bits-->0) {
	TA(*cur++==tsf_false);
    }
}

static void test_skip_bits(uint32_t num_bits) {
    test_skipping_struct(bit_creator,
                         bit_writer,
                         bit_destroyer,
                         &num_bits);
}

static void test_default_bits(uint32_t num_bits) {
    test_option_field(bit_creator,
		      bit_verifier,
		      &num_bits);
}

/* array skipping (with multiple levels of nesting) */

static size_t array_creator(tsf_type_t **stt,
                            uint32_t offset,
                            void *arg) {
    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(stt,"array",
            tsf_type_create_array(tsf_type_create(TSF_TK_INT32))));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(stt,"array",offset));
    
    return sizeof(tsf_native_array_t);
}

static void array_writer(void *data,
                         void *arg) {
    uint32_t *num_elements=arg;
    tsf_native_array_t *array=data;
    
    array->len=*num_elements;
    
    if (array->len>0) {
        TEST_ASSERT_ERRNO(array->data=malloc(sizeof(int32_t)*array->len));
    } else {
        array->data=NULL;
    }
}

static void array_destroyer(void *data,
                            void *arg) {
    tsf_native_array_t *array=data;
    if (array->data!=NULL) {
        free(array->data);
    }
}

static void test_skip_array(uint32_t num_elements) {
    test_skipping_struct(array_creator,
                         array_writer,
                         array_destroyer,
                         &num_elements);
}

typedef struct {
    uint32_t outer_num_elements;
    uint32_t inner_num_elements;
} array_of_array_params_t;

static size_t array_of_array_creator(tsf_type_t **stt,
                                     uint32_t offset,
                                     void *arg) {
    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(stt,"array",
            tsf_type_create_array(
                tsf_type_create_array(
                    tsf_type_create(TSF_TK_INT32)))));
    
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(stt,"array",offset));
    
    return sizeof(tsf_native_array_t);
}

static void array_of_array_writer(void *data,
                                  void *arg) {
    array_of_array_params_t *params=arg;
    tsf_native_array_t *outer_array=data,*cur;
    
    outer_array->len=params->outer_num_elements;
    
    TEST_ASSERT_ERRNO(
        outer_array->data=malloc(sizeof(tsf_native_array_t)*
                                 outer_array->len));
    
    cur=outer_array->data;
    cur+=outer_array->len;
    while (cur-->(tsf_native_array_t*)outer_array->data) {
        cur->len=params->inner_num_elements;
        TEST_ASSERT_ERRNO(cur->data=malloc(sizeof(int32_t)*cur->len));
    }
}

static void array_of_array_destroyer(void *data,
                                     void *arg) {
    tsf_native_array_t *outer_array=data;
    if (outer_array->len>0) {
        tsf_native_array_t *cur=outer_array->data;
        cur+=outer_array->len;
        while (cur-->(tsf_native_array_t*)outer_array->data) {
            if (cur->data!=NULL) {
                free(cur->data);
            }
        }
        free(outer_array->data);
    }
}

static void test_skip_array_of_array(uint32_t outer_num_elements,
                                     uint32_t inner_num_elements) {
    array_of_array_params_t params;
    params.outer_num_elements=outer_num_elements;
    params.inner_num_elements=inner_num_elements;
    test_skipping_struct(array_of_array_creator,
                         array_of_array_writer,
                         array_of_array_destroyer,
                         &params);
}

/* static-size struct (with no arrays) skipping, dynamic-size struct (with
 * arrays) skipping
 */

static size_t struct_creator(tsf_type_t **stt,
                             uint32_t offset,
                             void *arg) {
    tsf_type_t *sub_stt=tsf_type_create(TSF_TK_STRUCT);
    
    TAT(sub_stt);
    
    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(&sub_stt,"x",
            tsf_type_create(TSF_TK_INT8)));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(&sub_stt,"x",0));
    
    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(&sub_stt,"y",
            tsf_type_create(TSF_TK_INT8)));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(&sub_stt,"y",1));
    
    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(&sub_stt,"z",
            tsf_type_create(TSF_TK_INT8)));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(&sub_stt,"z",2));
    
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_set_size(&sub_stt,3));
    
    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(stt,"struct",sub_stt));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(stt,"struct",offset));
    
    return 3;
}

static void struct_writer(void *data,
                          void *arg) {
}

static void struct_destroyer(void *data,
                             void *arg) {
}

static void test_skip_struct() {
    test_skipping_struct(struct_creator,
                         struct_writer,
                         struct_destroyer,
                         NULL);
}

static size_t dyn_size_struct_creator(tsf_type_t **stt,
                                      uint32_t offset,
                                      void *arg) {
    tsf_type_t *sub_stt;
    
    TAT(sub_stt=tsf_type_create(TSF_TK_STRUCT));
    
    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(&sub_stt,"x",
            tsf_type_create(TSF_TK_INT8)));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(&sub_stt,"x",0));
    
    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(&sub_stt,"y",
            tsf_type_create(TSF_TK_INT8)));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(&sub_stt,"y",1));
    
    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(&sub_stt,"array",
            tsf_type_create_array(
                tsf_type_create(TSF_TK_INT8))));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(&sub_stt,"array",TSF_NATIVE_ALIGNMENT));
    
    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(&sub_stt,"z",
            tsf_type_create(TSF_TK_INT8)));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(&sub_stt,"z",TSF_NATIVE_ALIGNMENT+
                                                sizeof(tsf_native_array_t)));
    
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_set_size(&sub_stt,
                                        TSF_NATIVE_ALIGNMENT+
                                        sizeof(tsf_native_array_t)+1));
    
    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(stt,"struct",sub_stt));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(stt,"struct",offset));
    
    return TSF_NATIVE_ALIGNMENT+sizeof(tsf_native_array_t)+1;
}

static void dyn_size_struct_writer(void *data,
                                   void *arg) {
    tsf_native_array_t *array=(tsf_native_array_t*)
        (((uint8_t*)data)+TSF_NATIVE_ALIGNMENT);
    uint32_t *num_elements=arg;
    
    array->len=*num_elements;
    if (array->len>0) {
        array->data=malloc(sizeof(int8_t)*array->len);
    } else {
        array->data=NULL;
    }
}

static void dyn_size_struct_destroyer(void *data,
                                      void *arg) {
    tsf_native_array_t *array=(tsf_native_array_t*)
        (((uint8_t*)data)+TSF_NATIVE_ALIGNMENT);
    if (array->data!=NULL) {
        free(array->data);
    }
}

static void test_skip_dyn_size_struct(uint32_t num_elements) {
    test_skipping_struct(dyn_size_struct_creator,
                         dyn_size_struct_writer,
                         dyn_size_struct_destroyer,
                         &num_elements);
}

/* skipping of every type of primitive. */

static size_t primitive_creator(tsf_type_t **stt,
                                uint32_t offset,
                                void *arg) {
    tsf_type_kind_t *type_kind=arg;

    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(stt,"primitive",
            tsf_type_create(*type_kind)));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(stt,"primitive",offset));
    
    return tsf_primitive_type_kind_size_of(*type_kind);
}

static void primitive_writer(void *data,
                             void *arg) {
}

static void primitive_destroyer(void *data,
                                void *arg) {
}

static void test_skip_primitive(tsf_type_kind_t type_kind) {
    test_skipping_struct(primitive_creator,
                         primitive_writer,
                         primitive_destroyer,
                         &type_kind);
}

/* skip void! */

static size_t void_creator(tsf_type_t **stt,
                           uint32_t offset,
                           void *arg) {
    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(stt,"void",
            tsf_type_create(TSF_TK_VOID)));
    
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(stt,"void",0));
    
    return 0;
}

static void void_writer(void *data,
                        void *arg) {
}

static void void_destroyer(void *data,
                           void *arg) {
}

static void test_skip_void() {
    test_skipping_struct(void_creator,
                         void_writer,
                         void_destroyer,
                         NULL);
}

/* skip choices! */

static size_t void_choice_creator(tsf_type_t **stt,
                                  uint32_t offset,
                                  void *arg) {
    tsf_type_t *ct;
    
    TEST_ASSERT_TSFERROR(ct=tsf_type_create(TSF_TK_CHOICE));
    
    TEST_ASSERT_TSFERROR(
        tsf_choice_type_append(&ct,"a",
            tsf_type_create(TSF_TK_VOID)));
    TEST_ASSERT_TSFERROR(
        tsf_choice_type_append(&ct,"b",
            tsf_type_create(TSF_TK_VOID)));
    TEST_ASSERT_TSFERROR(
        tsf_choice_type_append(&ct,"c",
            tsf_type_create(TSF_TK_VOID)));

    /* FIXME: also test the not-in-place scenario! */
    TAT(tsf_native_choice_type_map(&ct,tsf_true,0,0,sizeof(tsf_choice_t)));

    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(stt,"choice",ct));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(stt,"choice",offset));
    
    return sizeof(tsf_choice_t);
}

static void void_choice_writer(void *data,
                               void *arg) {
    *((tsf_choice_t*)data) = *((tsf_choice_t*)arg);
}

static void void_choice_destroyer(void *data,
                                  void *arg) {
}

static void test_skip_void_choice(tsf_choice_t choice) {
    test_skipping_struct(void_choice_creator,
                         void_choice_writer,
                         void_choice_destroyer,
                         &choice);
}

typedef struct {
    tsf_choice_t value;
    float data;
} float_choice_t;

typedef struct {
    tsf_choice_t value;
    float *data;
} alt_float_choice_t;

typedef struct {
    tsf_bool_t use_alt;
    float_choice_t stuff;
} float_choice_data_t;

static size_t float_choice_creator(tsf_type_t **stt,
                                   uint32_t offset,
                                   void *_arg) {
    float_choice_data_t *arg=_arg;
    tsf_type_t *ct;
    
    TEST_ASSERT_TSFERROR(ct=tsf_type_create(TSF_TK_CHOICE));
    
    TEST_ASSERT_TSFERROR(
        tsf_choice_type_append(&ct,"a",
            tsf_type_create(TSF_TK_FLOAT)));
    TEST_ASSERT_TSFERROR(
        tsf_choice_type_append(&ct,"b",
            tsf_type_create(TSF_TK_FLOAT)));
    TEST_ASSERT_TSFERROR(
        tsf_choice_type_append(&ct,"c",
            tsf_type_create(TSF_TK_FLOAT)));

    /* FIXME: also test the not-in-place scenario! */
    if (arg->use_alt) {
	TAT(tsf_native_choice_type_map(&ct,
				       tsf_false,
				       tsf_offsetof(alt_float_choice_t,value),
				       tsf_offsetof(alt_float_choice_t,data),
				       sizeof(alt_float_choice_t)));
    } else {
	TAT(tsf_native_choice_type_map(&ct,
				       tsf_true,
				       tsf_offsetof(float_choice_t,value),
				       tsf_offsetof(float_choice_t,data),
				       sizeof(float_choice_t)));
    }

    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(stt,"choice",ct));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(stt,"choice",offset));
    
    if (arg->use_alt) {
	return sizeof(alt_float_choice_t);
    } else {
	return sizeof(float_choice_t);
    }
}

static void float_choice_writer(void *data,
                                void *_arg) {
    float_choice_data_t *arg=_arg;
    if (arg->use_alt) {
	alt_float_choice_t *targ=data;
	targ->value=arg->stuff.value;
	TAE(targ->data=(float*)malloc(sizeof(float)));
	*(targ->data)=arg->stuff.data;
    } else {
	memcpy(data,&(arg->stuff),sizeof(float_choice_t));
    }
}

static void float_choice_destroyer(void *data,
                                   void *_arg) {
    float_choice_data_t *arg=_arg;
    if (arg->use_alt) {
	free(((alt_float_choice_t*)data)->data);
    }
}

static void test_skip_float_choice(tsf_choice_t value,
                                   float data,
				   tsf_bool_t use_alt) {
    float_choice_data_t arg;
    
    arg.use_alt=use_alt;
    arg.stuff.value=value;
    arg.stuff.data=data;
    
    test_skipping_struct(float_choice_creator,
                         float_choice_writer,
                         float_choice_destroyer,
                         &arg);
}

static size_t string_creator(tsf_type_t **stt,
                             uint32_t offset,
                             void *arg) {
    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(stt,"string",
            tsf_type_create(TSF_TK_STRING)));
    TEST_ASSERT_TSFERROR(
        tsf_native_struct_type_map(stt,"string",offset));
    
    return sizeof(char*);
}

static void string_writer(void *data,
                          void *arg) {
    *((char**)data)=strdup("hello, world!");
}

static void string_destroyer(void *data,
                             void *arg) {
    free(*((char**)data));
}

static void test_skip_string() {
    test_skipping_struct(string_creator,
                         string_writer,
                         string_destroyer,
                         NULL);
}

static size_t integer_creator(tsf_type_t **stt,
                              uint32_t offset,
                              void *arg) {
    TAT(tsf_struct_type_append(stt, "integer", tsf_type_create(TSF_TK_INTEGER)));
    TAT(tsf_native_struct_type_map(stt, "integer", offset));
    return sizeof(tsf_integer_t);
}

static void integer_writer(void *data,
                           void *arg) {
    *((tsf_integer_t*)data) = *((tsf_integer_t*)arg);
}

static void integer_destroyer(void *data,
                              void *arg) {
    /* nothing to do */
}

static void test_skip_integer(tsf_integer_t value) {
    test_skipping_struct(integer_creator,
                         integer_writer,
                         integer_destroyer,
                         &value);
}

static size_t long_creator(tsf_type_t **stt,
                           uint32_t offset,
                           void *arg) {
    TAT(tsf_struct_type_append(stt, "long", tsf_type_create(TSF_TK_LONG)));
    TAT(tsf_native_struct_type_map(stt, "long", offset));
    return sizeof(tsf_long_t);
}

static void long_writer(void *data,
                        void *arg) {
    *((tsf_long_t*)data) = *((tsf_long_t*)arg);
}

static void long_destroyer(void *data,
                           void *arg) {
    /* nothing to do */
}

static void test_skip_long(tsf_long_t value) {
    test_skipping_struct(long_creator,
                         long_writer,
                         long_destroyer,
                         &value);
}

/* cur_choice is the choice that is chosen, or NULL to
 * indicate UINT32_MAX.
 * the ... is a NULL-terminated list of the source type's
 * choices followed by a NULL-terminated list of the dest
 * type's choices.  if cur_choice is non-NULL, it must be
 * one of the source choices. */
static void test_choice_void(const char *cur_choice,
                             ...) {
    va_list lst;
    
    tsf_type_t *src_type;
    tsf_type_t *dest_type;
    tsf_choice_t src=UINT32_MAX;
    tsf_choice_t *dest;
    tsf_choice_t expected_dest=UINT32_MAX;
    tsf_choice_t cnt;
    
    TEST_ASSERT_TSFERROR(src_type=tsf_type_create(TSF_TK_CHOICE));
    TEST_ASSERT_TSFERROR(dest_type=tsf_type_create(TSF_TK_CHOICE));
    
    TAT(tsf_native_choice_type_map(&src_type,tsf_true,0,0,sizeof(tsf_choice_t)));
    TAT(tsf_native_choice_type_map(&dest_type,tsf_true,0,0,sizeof(tsf_choice_t)));
    
    va_start(lst,cur_choice);
    
    for (cnt=0;;++cnt) {
        const char *cur=va_arg(lst,const char*);
        if (cur==NULL) {
            break;
        }
        TEST_ASSERT_TSFERROR(
            tsf_choice_type_append(&src_type,cur,
                tsf_type_create(TSF_TK_VOID)));
        if (cur_choice!=NULL && !strcmp(cur_choice,cur)) {
            src=cnt;
        }
    }
    
    TEST_ASSERT(cur_choice==NULL || src!=UINT32_MAX);
    
    for (cnt=0;;++cnt) {
        const char *cur=va_arg(lst,const char*);
        if (cur==NULL) {
            break;
        }
        TEST_ASSERT_TSFERROR(
            tsf_choice_type_append(&dest_type,cur,
                tsf_type_create(TSF_TK_VOID)));
        if (cur_choice!=NULL && !strcmp(cur_choice,cur)) {
            expected_dest=cnt;
        }
    }
    
    va_end(lst);
    
    dest=flip(dest_type,src_type,&src,tsf_false);
    
    TEST_ASSERT(*dest == expected_dest);
    
    tsf_region_free(dest);
    tsf_type_destroy(src_type);
    tsf_type_destroy(dest_type);
}

typedef struct {
    float_choice_t c;
    int32_t suffix;
} float_choice_struct_t;

typedef struct {
    alt_float_choice_t c;
    int32_t suffix;
} alt_float_choice_struct_t;

/* cur_choice is the choice that is chosen, or NULL to
 * indicate UINT32_MAX.  data is the float value to place
 * into the data field.
 * the ... is a NULL-terminated list of the source type's
 * choices followed by a NULL-terminated list of the dest
 * type's choices.  if cur_choice is non-NULL, it must be
 * one of the source choices. */
static void test_choice_float(const char *cur_choice,
                              float data,
                              int32_t suffix,
			      tsf_bool_t src_use_alt,
			      tsf_bool_t dest_use_alt,
                              ...) {
    va_list lst;
    
    tsf_type_t *src_type;
    tsf_type_t *dest_type;
    float_choice_struct_t src;
    float_choice_struct_t *dest;
    alt_float_choice_struct_t alt_src;
    alt_float_choice_struct_t *alt_dest;
    void *result;
    tsf_choice_t expected_dest=UINT32_MAX;
    tsf_choice_t cnt;
    
    src.c.value=UINT32_MAX;
    src.c.data=data;
    src.suffix=suffix;

    alt_src.c.value=UINT32_MAX;
    TAE(alt_src.c.data=malloc(sizeof(float)));
    *alt_src.c.data=data;
    alt_src.suffix=suffix;
    
    TEST_ASSERT_TSFERROR(src_type=tsf_type_create(TSF_TK_CHOICE));
    TEST_ASSERT_TSFERROR(dest_type=tsf_type_create(TSF_TK_CHOICE));
    
    if (src_use_alt) {
	TAT(tsf_native_choice_type_map(&src_type,
				       tsf_false,
				       tsf_offsetof(alt_float_choice_t,value),
				       tsf_offsetof(alt_float_choice_t,data),
				       sizeof(alt_float_choice_t)));
    } else {
	TAT(tsf_native_choice_type_map(&src_type,
				       tsf_true,
				       tsf_offsetof(float_choice_t,value),
				       tsf_offsetof(float_choice_t,data),
				       sizeof(float_choice_t)));
    }

    if (dest_use_alt) {
	TAT(tsf_native_choice_type_map(&dest_type,
				       tsf_false,
				       tsf_offsetof(alt_float_choice_t,value),
				       tsf_offsetof(alt_float_choice_t,data),
				       sizeof(alt_float_choice_t)));
    } else {
	TAT(tsf_native_choice_type_map(&dest_type,
				       tsf_true,
				       tsf_offsetof(float_choice_t,value),
				       tsf_offsetof(float_choice_t,data),
				       sizeof(float_choice_t)));
    }
    
    va_start(lst,dest_use_alt);
    
    for (cnt=0;;++cnt) {
        const char *cur=va_arg(lst,const char*);
        if (cur==NULL) {
            break;
        }
        TEST_ASSERT_TSFERROR(
            tsf_choice_type_append(&src_type,cur,
                tsf_type_create(TSF_TK_FLOAT)));
        if (cur_choice!=NULL && !strcmp(cur_choice,cur)) {
            src.c.value=cnt;
	    alt_src.c.value=cnt;
        }
    }

    TA(cur_choice==NULL || src.c.value!=UINT32_MAX);
    
    for (cnt=0;;++cnt) {
        const char *cur=va_arg(lst,const char*);
        if (cur==NULL) {
            break;
        }
        TEST_ASSERT_TSFERROR(
            tsf_choice_type_append(&dest_type,cur,
                tsf_type_create(TSF_TK_FLOAT)));
        if (cur_choice!=NULL && !strcmp(cur_choice,cur)) {
            expected_dest=cnt;
        }
    }
    
    va_end(lst);
    
    TAT(src_type=tsf_type_create_struct("c",src_type,
                                        "suffix",tsf_type_create(TSF_TK_INT32),
                                        NULL));
    TAT(dest_type=tsf_type_create_struct("c",dest_type,
                                         "suffix",tsf_type_create(TSF_TK_INT32),
                                         NULL));

    if (src_use_alt) {
	TAT(tsf_native_struct_type_map(&src_type,
				       "c",
				       tsf_offsetof(alt_float_choice_struct_t,c)));
	TAT(tsf_native_struct_type_map(&src_type,
				       "suffix",
				       tsf_offsetof(alt_float_choice_struct_t,suffix)));
	TAT(tsf_native_struct_type_set_size(&src_type,
					    sizeof(alt_float_choice_struct_t)));
    } else {
	TAT(tsf_native_struct_type_map(&src_type,
				       "c",
				       tsf_offsetof(float_choice_struct_t,c)));
	TAT(tsf_native_struct_type_map(&src_type,
				       "suffix",
				       tsf_offsetof(float_choice_struct_t,suffix)));
	TAT(tsf_native_struct_type_set_size(&src_type,
					    sizeof(float_choice_struct_t)));
    }
    
    if (dest_use_alt) {
	TAT(tsf_native_struct_type_map(&dest_type,
				       "c",
				       tsf_offsetof(alt_float_choice_struct_t,c)));
	TAT(tsf_native_struct_type_map(&dest_type,
				       "suffix",
				       tsf_offsetof(alt_float_choice_struct_t,suffix)));
	TAT(tsf_native_struct_type_set_size(&dest_type,
					    sizeof(alt_float_choice_struct_t)));
    } else {
	TAT(tsf_native_struct_type_map(&dest_type,
				       "c",
				       tsf_offsetof(float_choice_struct_t,c)));
	TAT(tsf_native_struct_type_map(&dest_type,
				       "suffix",
				       tsf_offsetof(float_choice_struct_t,suffix)));
	TAT(tsf_native_struct_type_set_size(&dest_type,
					    sizeof(float_choice_struct_t)));
    }
    
    if (src_use_alt) {
	result=flip(dest_type,src_type,&alt_src,tsf_false);
    } else {
	result=flip(dest_type,src_type,&src,tsf_false);
    }

    if (dest_use_alt) {
	alt_dest=result;

	TEST_ASSERT(alt_dest->c.value == expected_dest);
	if (alt_dest->c.value!=UINT32_MAX) {
	    TEST_ASSERT(*(alt_dest->c.data) == data);
	}
	TA(alt_dest->suffix == suffix);
    } else {
	dest=result;

	TEST_ASSERT(dest->c.value == expected_dest);
	if (dest->c.value!=UINT32_MAX) {
	    TEST_ASSERT(dest->c.data == data);
	}
	TA(dest->suffix == suffix);
    }
    
    tsf_region_free(result);
    tsf_type_destroy(src_type);
    tsf_type_destroy(dest_type);
    free(alt_src.c.data);
}

/* FIXME: need tests to excersize
   tsf_choice_type_uses_same_indices(), since (as of this writing)
   it looks to be quite broken! */

static void test_string() {
    char *string1="Hello";
    char **string2;
    
    string2=flip(tsf_type_create(TSF_TK_STRING),
                 tsf_type_create(TSF_TK_STRING),
                 &string1,
                 tsf_false);
    
    TEST_ASSERT(!strcmp(string1,*string2));
    
    tsf_region_free(string2);
}

#define DO_ARRAY_TEST(test) do {    \
    DO_TEST(test(0));   \
    DO_TEST(test(1));   \
    DO_TEST(test(2));   \
} while (0)

#define DO_PRIM_TEST_FROM(type) do {    \
    DO_TEST(test_##type##_to_int8_t()); \
    DO_TEST(test_##type##_to_uint8_t());    \
    DO_TEST(test_##type##_to_int16_t());    \
    DO_TEST(test_##type##_to_uint16_t());   \
    DO_TEST(test_##type##_to_int32_t());    \
    DO_TEST(test_##type##_to_uint32_t());   \
    DO_TEST(test_##type##_to_int64_t());    \
    DO_TEST(test_##type##_to_uint64_t());   \
    DO_TEST(test_##type##_to_tsf_integer_t());   \
    DO_TEST(test_##type##_to_tsf_long_t());   \
    DO_TEST(test_##type##_to_float());  \
    DO_TEST(test_##type##_to_double()); \
    DO_TEST(test_##type##_to_tsf_bool_t()); \
    \
    DO_ARRAY_TEST(test_array_of_##type##_to_int8_t);    \
    DO_ARRAY_TEST(test_array_of_##type##_to_uint8_t);   \
    DO_ARRAY_TEST(test_array_of_##type##_to_int16_t);   \
    DO_ARRAY_TEST(test_array_of_##type##_to_uint16_t);  \
    DO_ARRAY_TEST(test_array_of_##type##_to_int32_t);   \
    DO_ARRAY_TEST(test_array_of_##type##_to_uint32_t);  \
    DO_ARRAY_TEST(test_array_of_##type##_to_int64_t);   \
    DO_ARRAY_TEST(test_array_of_##type##_to_uint64_t);  \
    DO_ARRAY_TEST(test_array_of_##type##_to_tsf_integer_t);  \
    DO_ARRAY_TEST(test_array_of_##type##_to_tsf_long_t);  \
    DO_ARRAY_TEST(test_array_of_##type##_to_float); \
    DO_ARRAY_TEST(test_array_of_##type##_to_double);    \
    DO_ARRAY_TEST(test_array_of_##type##_to_tsf_bool_t);    \
} while (0)

#define DO_TEST_CHOICE_FLOAT(alt_src,alt_dest) do {	\
	DO_TEST(test_choice_float(NULL,			\
				  0.0,			\
				  19,			\
				  alt_src,		\
				  alt_dest,		\
				  NULL,			\
				  NULL));		\
	DO_TEST(test_choice_float(NULL,			\
				  1.0,			\
				  19,			\
				  alt_src,		\
				  alt_dest,		\
				  "a",			\
				  NULL,			\
				  "a",			\
				  NULL));		\
	DO_TEST(test_choice_float("a",			\
				  2.0,			\
				  19,			\
				  alt_src,		\
				  alt_dest,		\
				  "a",			\
				  NULL,			\
				  "a",			\
				  NULL));		\
	DO_TEST(test_choice_float("a",			\
				  3.0,			\
				  19,			\
				  alt_src,		\
				  alt_dest,		\
				  "a",			\
				  NULL,			\
				  "b",			\
				  NULL));		\
	DO_TEST(test_choice_float(NULL,			\
				  4.0,			\
				  19,			\
				  alt_src,		\
				  alt_dest,		\
				  "a",			\
				  "b",			\
				  NULL,			\
				  "b",			\
				  "a",			\
				  NULL));		\
	DO_TEST(test_choice_float("a",			\
				  5.0,			\
				  19,			\
				  alt_src,		\
				  alt_dest,		\
				  "a",			\
				  "b",			\
				  NULL,			\
				  "b",			\
				  "a",			\
				  NULL));		\
	DO_TEST(test_choice_float("b",			\
				  6.0,			\
				  19,			\
				  alt_src,		\
				  alt_dest,		\
				  "a",			\
				  "b",			\
				  NULL,			\
				  "b",			\
				  "a",			\
				  NULL));		\
    } while (0)

static void do_all_tests() {
    DO_TEST(test_skip_string());
    DO_TEST(test_string());
    
    DO_TEST(test_choice_void(NULL,NULL,NULL));
    DO_TEST(test_choice_void(NULL,
                             "a",
                             NULL,
                             "a",
                             NULL));
    DO_TEST(test_choice_void("a",
                             "a",
                             NULL,
                             "a",
                             NULL));
    DO_TEST(test_choice_void("a",
                             "a",
                             NULL,
                             "b",
                             NULL));
    DO_TEST(test_choice_void(NULL,
                             "a",
                             "b",
                             NULL,
                             "b",
                             "a",
                             NULL));
    DO_TEST(test_choice_void("a",
                             "a",
                             "b",
                             NULL,
                             "b",
                             "a",
                             NULL));
    DO_TEST(test_choice_void("b",
                             "a",
                             "b",
                             NULL,
                             "b",
                             "a",
                             NULL));
    
    DO_TEST_CHOICE_FLOAT(tsf_false,tsf_false);
    DO_TEST_CHOICE_FLOAT(tsf_true,tsf_false);
    DO_TEST_CHOICE_FLOAT(tsf_false,tsf_true);
    DO_TEST_CHOICE_FLOAT(tsf_true,tsf_true);
    
    DO_TEST(test_skip_void());
    
    DO_TEST(test_skip_void_choice(0));
    DO_TEST(test_skip_void_choice(1));
    DO_TEST(test_skip_void_choice(2));
    DO_TEST(test_skip_void_choice(3));
    DO_TEST(test_skip_void_choice(UINT32_MAX));
    
    DO_TEST(test_skip_float_choice(0,0.0,tsf_false));
    DO_TEST(test_skip_float_choice(1,0.0,tsf_false));
    DO_TEST(test_skip_float_choice(2,0.0,tsf_false));
    DO_TEST(test_skip_float_choice(UINT32_MAX,0.0,tsf_false));
    
    DO_TEST(test_skip_float_choice(0,0.0,tsf_true));
    DO_TEST(test_skip_float_choice(1,0.0,tsf_true));
    DO_TEST(test_skip_float_choice(2,0.0,tsf_true));
    DO_TEST(test_skip_float_choice(UINT32_MAX,0.0,tsf_true));
    
    DO_TEST(test_struct_succeed());
    DO_TEST(test_struct_fail());
    DO_TEST(test_struct_reverse());
    
    DO_TEST(test_skip_primitive(TSF_TK_INT8));
    DO_TEST(test_skip_primitive(TSF_TK_UINT8));
    DO_TEST(test_skip_primitive(TSF_TK_INT16));
    DO_TEST(test_skip_primitive(TSF_TK_UINT16));
    DO_TEST(test_skip_primitive(TSF_TK_INT32));
    DO_TEST(test_skip_primitive(TSF_TK_UINT32));
    DO_TEST(test_skip_primitive(TSF_TK_INT64));
    DO_TEST(test_skip_primitive(TSF_TK_UINT64));
    DO_TEST(test_skip_primitive(TSF_TK_FLOAT));
    DO_TEST(test_skip_primitive(TSF_TK_DOUBLE));
    
    DO_TEST(test_skip_integer(INT32_C(0)));
    DO_TEST(test_skip_integer(INT32_C(62)));
    DO_TEST(test_skip_integer(INT32_C(63)));
    DO_TEST(test_skip_integer(INT32_C(64)));
    DO_TEST(test_skip_integer(INT32_C(16190)));
    DO_TEST(test_skip_integer(INT32_C(16191)));
    DO_TEST(test_skip_integer(INT32_C(16192)));
    DO_TEST(test_skip_integer(INT32_C(8404798)));
    DO_TEST(test_skip_integer(INT32_C(8404799)));
    DO_TEST(test_skip_integer(INT32_C(8404800)));
    DO_TEST(test_skip_integer(INT32_C(8404801)));
    DO_TEST(test_skip_integer(INT32_C(1000000000)));
    DO_TEST(test_skip_integer(INT32_C(-62)));
    DO_TEST(test_skip_integer(INT32_C(-63)));
    DO_TEST(test_skip_integer(INT32_C(-64)));
    DO_TEST(test_skip_integer(INT32_C(-16190)));
    DO_TEST(test_skip_integer(INT32_C(-16191)));
    DO_TEST(test_skip_integer(INT32_C(-16192)));
    DO_TEST(test_skip_integer(INT32_C(-8404798)));
    DO_TEST(test_skip_integer(INT32_C(-8404799)));
    DO_TEST(test_skip_integer(INT32_C(-8404800)));
    DO_TEST(test_skip_integer(INT32_C(-8404801)));
    DO_TEST(test_skip_integer(INT32_C(-1000000000)));

    DO_TEST(test_skip_long(INT64_C(0)));
    DO_TEST(test_skip_long(INT64_C(62)));
    DO_TEST(test_skip_long(INT64_C(63)));
    DO_TEST(test_skip_long(INT64_C(64)));
    DO_TEST(test_skip_long(INT64_C(16190)));
    DO_TEST(test_skip_long(INT64_C(16191)));
    DO_TEST(test_skip_long(INT64_C(16192)));
    DO_TEST(test_skip_long(INT64_C(8404798)));
    DO_TEST(test_skip_long(INT64_C(8404799)));
    DO_TEST(test_skip_long(INT64_C(8404800)));
    DO_TEST(test_skip_long(INT64_C(8404801)));
    DO_TEST(test_skip_long(INT64_C(1000000000)));
    DO_TEST(test_skip_long(INT64_C(1000000000000)));
    DO_TEST(test_skip_long(INT64_C(-62)));
    DO_TEST(test_skip_long(INT64_C(-63)));
    DO_TEST(test_skip_long(INT64_C(-64)));
    DO_TEST(test_skip_long(INT64_C(-16190)));
    DO_TEST(test_skip_long(INT64_C(-16191)));
    DO_TEST(test_skip_long(INT64_C(-16192)));
    DO_TEST(test_skip_long(INT64_C(-8404798)));
    DO_TEST(test_skip_long(INT64_C(-8404799)));
    DO_TEST(test_skip_long(INT64_C(-8404800)));
    DO_TEST(test_skip_long(INT64_C(-8404801)));
    DO_TEST(test_skip_long(INT64_C(-1000000000)));
    DO_TEST(test_skip_long(INT64_C(-1000000000000)));

    DO_TEST(test_skip_struct());
    
    DO_TEST(test_skip_dyn_size_struct(0));
    DO_TEST(test_skip_dyn_size_struct(1));
    DO_TEST(test_skip_dyn_size_struct(2));
    
    DO_TEST(test_skip_bits(0));
    DO_TEST(test_skip_bits(1));
    DO_TEST(test_skip_bits(7));
    DO_TEST(test_skip_bits(8));
    DO_TEST(test_skip_bits(31));
    DO_TEST(test_skip_bits(32));
    DO_TEST(test_skip_bits(63));
    
    DO_TEST(test_default_bits(0));
    DO_TEST(test_default_bits(1));
    DO_TEST(test_default_bits(7));
    DO_TEST(test_default_bits(8));
    DO_TEST(test_default_bits(31));
    DO_TEST(test_default_bits(32));
    DO_TEST(test_default_bits(63));
    
    DO_TEST(test_skip_array(0));
    DO_TEST(test_skip_array(1));
    DO_TEST(test_skip_array(254));
    DO_TEST(test_skip_array(255));
    DO_TEST(test_skip_array(256));
    
    DO_TEST(test_skip_array_of_array(0,0));
    DO_TEST(test_skip_array_of_array(1,0));
    DO_TEST(test_skip_array_of_array(2,0));
    DO_TEST(test_skip_array_of_array(1,1));
    DO_TEST(test_skip_array_of_array(2,1));
    DO_TEST(test_skip_array_of_array(1,2));
    DO_TEST(test_skip_array_of_array(2,2));
    
    DO_PRIM_TEST_FROM(int8_t);
    DO_PRIM_TEST_FROM(uint8_t);
    DO_PRIM_TEST_FROM(int16_t);
    DO_PRIM_TEST_FROM(uint16_t);
    DO_PRIM_TEST_FROM(int32_t);
    DO_PRIM_TEST_FROM(uint32_t);
    DO_PRIM_TEST_FROM(int64_t);
    DO_PRIM_TEST_FROM(uint64_t);
    DO_PRIM_TEST_FROM(tsf_integer_t);
    DO_PRIM_TEST_FROM(tsf_long_t);
    DO_PRIM_TEST_FROM(float);
    DO_PRIM_TEST_FROM(double);
    DO_PRIM_TEST_FROM(tsf_bool_t);
}

int main(int c,char **v) {
    TEST_INIT(c,v);
    
    use_conversion=tsf_false;

    TEST_PUSH("generator_to_parser_conversion");
    do_all_tests();
    TEST_POP();
    
    use_conversion=tsf_true;

    TEST_PUSH("clone_conversion");
    do_all_tests();
    TEST_POP();
    
    TEST_DONE();
    
    return 0;
}


