/*
 * Copyright (C) 2004, 2005, 2015 Filip Pizlo. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY FILIP PIZLO ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL FILIP PIZLO OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "tsf.h"
#include "tsf_test.h"

static void test_space_occupied_by_empty_struct() {
    /* the amount of space occupied by an empty struct should be 0! */
    
    tsf_type_t *empty_struct;
    tsf_genrtr_t *generator;
    tsf_parser_t *parser;
    tsf_buffer_t *buf;
    void *result;
    TAT(empty_struct=tsf_type_create_struct(NULL));
    TAT(tsf_native_type_produce_some_mapping(&empty_struct,tsf_true));
    TAT(generator=tsf_generator_create(empty_struct));
    TAT(parser=tsf_parser_create(empty_struct));
    TAT(buf=tsf_generator_generate(generator,NULL));
    TAT(result=tsf_parser_parse(parser,buf));
    TA(tsf_region_size(result)==0); /* take that, I_ALLOC_ROOT! */
    tsf_region_free(result);
    tsf_buffer_destroy(buf);
    tsf_parser_destroy(parser);
    tsf_generator_destroy(generator);
    tsf_type_destroy(empty_struct);
}

static void test_type_create_struct_empty() {
    tsf_type_t *type_one,
               *type_two;
    
    TEST_ASSERT_TSFERROR(type_one=tsf_type_create(TSF_TK_STRUCT));
    
    TEST_ASSERT_TSFERROR(type_two=tsf_type_create_struct(NULL));
    
    TEST_ASSERT(tsf_type_compare(type_one, type_two));
}

static void test_type_create_struct_one() {
    tsf_type_t *type_one,
               *type_two;
    
    TEST_ASSERT_TSFERROR(type_one=tsf_type_create(TSF_TK_STRUCT));
    
    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(&type_one,
                               "blah",
                               tsf_type_create(TSF_TK_FLOAT)));
    
    TEST_ASSERT_TSFERROR(
        type_two=tsf_type_create_struct("blah",
                                        tsf_type_create(TSF_TK_FLOAT),
                                        NULL));
    
    TEST_ASSERT(tsf_type_compare(type_one, type_two));
}

static void test_type_create_struct_three() {
    tsf_type_t *type_one,
               *type_two;
    
    TEST_ASSERT_TSFERROR(type_one=tsf_type_create(TSF_TK_STRUCT));
    
    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(&type_one,
                               "a",
                               tsf_type_create(TSF_TK_FLOAT)));
    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(&type_one,
                               "b",
                               tsf_type_create(TSF_TK_FLOAT)));
    TEST_ASSERT_TSFERROR(
        tsf_struct_type_append(&type_one,
                               "c",
                               tsf_type_create(TSF_TK_FLOAT)));
    
    TEST_ASSERT_TSFERROR(
        type_two=tsf_type_create_struct("a",
                                        tsf_type_create(TSF_TK_FLOAT),
                                        "b",
                                        tsf_type_create(TSF_TK_FLOAT),
                                        "c",
                                        tsf_type_create(TSF_TK_FLOAT),
                                        NULL));
    
    TEST_ASSERT(tsf_type_compare(type_one, type_two));
}

static void test_type_table_remove_empty(void) {
    tsf_type_table_t *tt;
    TAT(tt = tsf_type_table_create());
    TET(tsf_type_table_remove(tt, "foo"), TSF_E_ELE_NOT_FOUND);
    TA(tsf_type_table_get_num_elements(tt) == 0);
    tsf_type_table_destroy(tt);
}

static void test_type_table_remove_only(void) {
    tsf_type_table_t *tt;
    TAT(tt = tsf_type_table_create());
    TAT(tsf_type_table_append(tt, "foo", tsf_type_create(TSF_TK_INTEGER)));
    TA(tsf_type_table_get_num_elements(tt) == 1);
    TAT(tsf_type_table_remove(tt, "foo"));
    TA(tsf_type_table_get_num_elements(tt) == 0);
    tsf_type_table_destroy(tt);
}

static void test_type_table_remove_first_of_two(void) {
    tsf_type_table_t *tt;
    TAT(tt = tsf_type_table_create());
    TAT(tsf_type_table_append(tt, "foo", tsf_type_create(TSF_TK_INTEGER)));
    TAT(tsf_type_table_append(tt, "bar", tsf_type_create(TSF_TK_LONG)));
    TA(tsf_type_table_get_num_elements(tt) == 2);
    TAT(tsf_type_table_remove(tt, "foo"));
    TA(tsf_type_table_get_num_elements(tt) == 1);
    TA(!strcmp(tsf_named_type_get_name(tsf_type_table_find_by_index(tt, 0)), "bar"));
    TA(tsf_named_type_get_type(tsf_type_table_find_by_index(tt, 0)) == tsf_type_create(TSF_TK_LONG));
    tsf_type_table_destroy(tt);
}

static void test_type_table_remove_second_of_two(void) {
    tsf_type_table_t *tt;
    TAT(tt = tsf_type_table_create());
    TAT(tsf_type_table_append(tt, "foo", tsf_type_create(TSF_TK_INTEGER)));
    TAT(tsf_type_table_append(tt, "bar", tsf_type_create(TSF_TK_LONG)));
    TA(tsf_type_table_get_num_elements(tt) == 2);
    TAT(tsf_type_table_remove(tt, "bar"));
    TA(tsf_type_table_get_num_elements(tt) == 1);
    TA(!strcmp(tsf_named_type_get_name(tsf_type_table_find_by_index(tt, 0)), "foo"));
    TA(tsf_named_type_get_type(tsf_type_table_find_by_index(tt, 0)) == tsf_type_create(TSF_TK_INTEGER));
    tsf_type_table_destroy(tt);
}

static void test_type_table_remove_middle_of_three(void) {
    tsf_type_table_t *tt;
    TAT(tt = tsf_type_table_create());
    TAT(tsf_type_table_append(tt, "foo", tsf_type_create(TSF_TK_INTEGER)));
    TAT(tsf_type_table_append(tt, "bar", tsf_type_create(TSF_TK_LONG)));
    TAT(tsf_type_table_append(tt, "baz", tsf_type_create(TSF_TK_BIT)));
    TA(tsf_type_table_get_num_elements(tt) == 3);
    TAT(tsf_type_table_remove(tt, "bar"));
    TA(tsf_type_table_get_num_elements(tt) == 2);
    TA(!strcmp(tsf_named_type_get_name(tsf_type_table_find_by_index(tt, 0)), "foo"));
    TA(tsf_named_type_get_type(tsf_type_table_find_by_index(tt, 0)) == tsf_type_create(TSF_TK_INTEGER));
    TA(!strcmp(tsf_named_type_get_name(tsf_type_table_find_by_index(tt, 1)), "baz"));
    TA(tsf_named_type_get_type(tsf_type_table_find_by_index(tt, 1)) == tsf_type_create(TSF_TK_BIT));
    tsf_type_table_destroy(tt);
}

typedef struct {
    tsf_integer_t x;
    double y;
} Foo;

static Foo *global_foo;
static int thingy;

static void construct(void *argument) {
    TA(argument == global_foo);
    thingy = 1;
}

static void destruct(void *argument) {
    TA(argument == global_foo);
    thingy = 2;
}

static tsf_type_t *make_foo_type(void) {
    tsf_type_t *type;
    TAT(type = tsf_native_struct_type_create(sizeof(Foo)));
    TAT(tsf_native_struct_type_append(&type, "x", tsf_type_create(TSF_TK_INTEGER), tsf_offsetof(Foo, x)));
    TAT(tsf_native_struct_type_append(&type, "y", tsf_type_create(TSF_TK_DOUBLE), tsf_offsetof(Foo, y)));
    return type;
}

static void test_native_type_creation_and_callbacks(void) {
    tsf_type_t *type;
    tsf_buffer_t *buf;
    tsf_genrtr_t *gen;
    tsf_parser_t *parser;
    Foo foo;
    void *region;
    
    thingy = 0;

    type = make_foo_type();
    TAT(tsf_native_struct_type_set_constructor(&type, construct));
    TAT(tsf_native_struct_type_set_pre_destructor(&type, destruct));
    TA(tsf_native_struct_type_get_constructor(type) == construct);
    TA(tsf_native_struct_type_get_pre_destructor(type) == destruct);
    TA(thingy == 0);
    
    TAT(gen = tsf_generator_create(type));
    TAT(parser = tsf_parser_create(type));
    
    foo.x = 42;
    foo.y = 43.;
    
    TAT(buf = tsf_generator_generate(gen, &foo));
    
    memset(&foo, 0, sizeof(Foo));
    
    TA(thingy == 0);
    global_foo = &foo;
    region = tsf_region_create(0);
    TAT(tsf_parser_parse_into(parser, buf, &foo, region));
    TA(thingy == 1);
    
    tsf_region_free(region);
    TA(thingy == 2);
    
    TA(foo.x == 42);
    TA(foo.y == 43.);
    
    tsf_generator_destroy(gen);
    tsf_parser_destroy(parser);
    tsf_buffer_destroy(buf);
    tsf_type_destroy(type);
}

static void test_resizable_buffer(void) {
    tsf_type_t *type;
    tsf_buffer_t *buf;
    tsf_genrtr_t *gen;
    tsf_parser_t *parser;
    tsf_serial_out_man_t *out_man;
    tsf_serial_in_man_t *in_man;
    tsf_resizable_buffer_t out_buf;
    tsf_light_buffer_t in_buf;
    Foo foo;
    
    type = make_foo_type();

    TAT(gen = tsf_generator_create(type));
    TAT(parser = tsf_parser_create(type));
    
    foo.x = 43;
    foo.y = 44.0;
    
    memset(&out_buf, 0, sizeof(out_buf));
    TAT(out_man = tsf_serial_out_man_create(tsf_resizable_buffer_writer, &out_buf));
    TAT(buf = tsf_generator_generate(gen, &foo));
    TAT(tsf_serial_out_man_write(out_man, buf));
    tsf_buffer_destroy(buf);
    tsf_serial_out_man_destroy(out_man);
    
    memset(&foo, 0, sizeof(Foo));
    
    in_buf.cur = out_buf.base;
    in_buf.end = out_buf.cur;
    
    TAT(in_man = tsf_serial_in_man_create(tsf_buffer_only_reader, &in_buf, NULL));
    TAT(buf = tsf_serial_in_man_read(in_man));
    TAT(tsf_parser_parse_into(parser, buf, &foo, NULL));
    tsf_buffer_destroy(buf);
    tsf_serial_in_man_destroy(in_man);
    free(out_buf.base);
    
    TA(foo.x == 43);
    TA(foo.y == 44.0);
    
    tsf_generator_destroy(gen);
    tsf_parser_destroy(parser);
    tsf_type_destroy(type);
}

typedef struct {
    tsf_choice_t value;
    Foo *data;
} Bar;

static const tsf_choice_t Bar_none = 0;
static const tsf_choice_t Bar_some = 1;

static void test_out_of_place_choice(tsf_bool_t has_some) {
    tsf_type_t *type;
    tsf_buffer_t *buf;
    tsf_genrtr_t *gen;
    tsf_parser_t *parser;
    Bar *bar;
    
    TAT(type = tsf_native_choice_type_create(tsf_false,
                                             tsf_offsetof(Bar, value),
                                             tsf_offsetof(Bar, data),
                                             sizeof(Bar)));
    TAT(tsf_choice_type_append(&type, "none", tsf_type_create(TSF_TK_VOID)));
    TAT(tsf_choice_type_append(&type, "some", make_foo_type()));
    
    TAT(gen = tsf_generator_create(type));
    TAT(parser = tsf_parser_create(type));
    
    TAE(bar = tsf_region_create(sizeof(Bar)));
    if (has_some) {
        bar->value = Bar_some;
        TAE(bar->data = tsf_region_alloc(bar, sizeof(Foo)));
        bar->data->x = 44;
        bar->data->y = 45.;
    } else {
        bar->value = Bar_none;
    }
    
    TAT(buf = tsf_generator_generate(gen, bar));
    tsf_region_free(bar);
    
    TAT(bar = tsf_parser_parse(parser, buf));
    tsf_buffer_destroy(buf);
    
    if (has_some) {
        TA(bar->value == Bar_some);
        TA(bar->data != NULL);
        TA(bar->data->x == 44);
        TA(bar->data->y == 45.);
    } else {
        TA(bar->value == Bar_none);
    }
    
    tsf_region_free(bar);
    tsf_generator_destroy(gen);
    tsf_parser_destroy(parser);
    tsf_type_destroy(type);
}

int main(int c,char **v) {
    TEST_INIT(c,v);
    
    DO_TEST(test_space_occupied_by_empty_struct());
    DO_TEST(test_type_create_struct_empty());
    DO_TEST(test_type_create_struct_one());
    DO_TEST(test_type_create_struct_three());
    DO_TEST(test_type_table_remove_empty());
    DO_TEST(test_type_table_remove_only());
    DO_TEST(test_type_table_remove_first_of_two());
    DO_TEST(test_type_table_remove_second_of_two());
    DO_TEST(test_type_table_remove_middle_of_three());
    DO_TEST(test_native_type_creation_and_callbacks());
    DO_TEST(test_resizable_buffer());
    DO_TEST(test_out_of_place_choice(tsf_false));
    DO_TEST(test_out_of_place_choice(tsf_true));
    
    TEST_DONE();
    
    return 0;
}


