/*
 * Copyright (C) 2004, 2005, 2006 Filip Pizlo. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY FILIP PIZLO ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL FILIP PIZLO OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "tsf.h"
#include "tsf_test.h"

#include <unistd.h>
#include <stdarg.h>

static tsf_zip_wtr_attr_t zip_wtr_attr;
static tsf_zip_rdr_attr_t zip_rdr_attr;
static tsf_stream_file_input_t *input;
static tsf_stream_file_output_t *output;

static void t_reset() {
    unlink("stream_file_test.dat");
}

static void t_open_output(tsf_bool_t append) {
    TEST_ASSERT_TSFERROR(
        output=tsf_stream_file_output_open("stream_file_test.dat",
                                           append,&zip_wtr_attr));
}

static void t_open_input() {
    TEST_ASSERT_TSFERROR(
        input=tsf_stream_file_input_open("stream_file_test.dat",
					 NULL,&zip_rdr_attr));
}

static void t_close_output() {
    TEST_ASSERT_TSFERROR(tsf_stream_file_output_close(output));
}

static void t_end_input() {
    TET(tsf_stream_file_input_read(input), TSF_E_EOF);
}

static void t_close_input() {
    tsf_stream_file_input_close(input);
}

static void setup() {
    t_reset();
    t_open_output(tsf_false);
}

static void flip() {
    t_close_output();
    t_open_input();
}

static void teardown() {
    t_end_input();
    t_close_input();
    t_reset();
}

static void test_n_void(uint32_t num) {
    tsf_buffer_t *buf;
    uint32_t n;
    
    setup();
    
    for (n=num;n-->0;) {
        TEST_ASSERT_TSFERROR(
            buf=tsf_reflect_make_buffer(tsf_reflect_create_void()));
        
        TEST_ASSERT_TSFERROR(
            tsf_stream_file_output_write(output,buf));
        
        tsf_buffer_destroy(buf);
    }
    
    flip();
    
    for (n=num;n-->0;) {
        TEST_ASSERT_TSFERROR(
            buf=tsf_stream_file_input_read(input));
        
        TEST_ASSERT(tsf_type_get_kind_code(tsf_buffer_get_type(buf))
                    == TSF_TK_VOID);
        
        tsf_buffer_destroy(buf);
    }
    
    teardown();
}

static void test_n_different_types(uint32_t num,
                                   uint32_t any_level) {
    tsf_buffer_t *buf;
    uint32_t n,m;
    char name[64];
    tsf_type_t *type,*subtype;
    tsf_reflect_t *r,*r2;
    
    setup();
    
    for (n=num;n-->0;) {
        sprintf(name,fui32,n);
        
        buf=NULL;
        
        for (m=any_level+1;m-->0;) {
            TAT(type=tsf_type_create_struct(
                name,
                buf==NULL?tsf_type_create(TSF_TK_VOID):
                          tsf_type_create(TSF_TK_ANY),
                NULL));
            
            TAT(r=tsf_reflect_create(type));
            
            tsf_type_destroy(type);
            
            TAT(tsf_struct_reflect_set_element(
                r,
                name,
                buf==NULL?tsf_reflect_create_void():
                          tsf_reflect_create_any(buf)));
            
            TAT(buf=tsf_reflect_make_buffer(r));
            
            tsf_reflect_destroy(r);
        }
        
        TAT(tsf_stream_file_output_write(output,buf));
        
        tsf_buffer_destroy(buf);
    }
    
    flip();
    
    for (n=num;n-->0;) {
        sprintf(name,fui32,n);
        
        TAT(buf=tsf_stream_file_input_read(input));
        
        for (m=any_level+1;m-->0;) {
            TAT(r=tsf_reflect_from_buffer(buf));
            
            tsf_buffer_destroy(buf);
            
            type=tsf_reflect_get_type(r);
            
            TA(tsf_type_get_kind_code(type) == TSF_TK_STRUCT);
            TAT(subtype=tsf_struct_type_find(type,name));
            
            if (m==0) {
                TA(tsf_type_get_kind_code(subtype) == TSF_TK_VOID);
            } else {
                TA(tsf_type_get_kind_code(subtype) == TSF_TK_ANY);
                
                TAT(r2=tsf_struct_reflect_get_element(r,name));
                
                TA(tsf_type_get_kind_code(tsf_reflect_get_type(r2))
                   == TSF_TK_ANY);
                
                TAT(buf=tsf_buffer_dup(tsf_any_reflect_get_buffer(r2)));
            }
            
            tsf_reflect_destroy(r);
        }
    }
    
    teardown();
}

static void test_n_void_appends(uint32_t num,const tsf_zip_mode_t *modelist) {
    tsf_buffer_t *buf;
    uint32_t n;
    const tsf_zip_mode_t *curmode;
    
    for (n=num,curmode=modelist;n-->0;curmode++) {
	if (!tsf_zip_wtr_attr_get_zlib_default(&zip_wtr_attr)
	    && *curmode==TSF_ZIP_ZLIB) {
	    return;
	}
	if (!tsf_zip_wtr_attr_get_bzip2_default(&zip_wtr_attr)
	    && *curmode==TSF_ZIP_BZIP2) {
	    return;
	}
    }
    
    t_reset();
    
    for (n=num,curmode=modelist;n-->0;curmode++) {
	tsf_zip_wtr_attr_get_for_mode(&zip_wtr_attr,*curmode);
        t_open_output(tsf_true);

        TEST_ASSERT_TSFERROR(
            buf=tsf_reflect_make_buffer(tsf_reflect_create_void()));
        
        TEST_ASSERT_TSFERROR(
            tsf_stream_file_output_write(output,buf));
        
        tsf_buffer_destroy(buf);

        t_close_output();
    }

    t_open_input();
    
    for (n=num;n-->0;) {
        TEST_ASSERT_TSFERROR(
            buf=tsf_stream_file_input_read(input));
        
        TEST_ASSERT(tsf_type_get_kind_code(tsf_buffer_get_type(buf))
                    == TSF_TK_VOID);
        
        tsf_buffer_destroy(buf);
    }
    
    teardown();
}

static void test_n_void_with_rewind(uint32_t num) {
    tsf_buffer_t *buf;
    uint32_t n;
    tsf_stream_file_mark_t zeromark;
    
    setup();
    
    for (n=num;n-->0;) {
        TEST_ASSERT_TSFERROR(
            buf=tsf_reflect_make_buffer(tsf_reflect_create_void()));
        
        TEST_ASSERT_TSFERROR(
            tsf_stream_file_output_write(output,buf));
        
        tsf_buffer_destroy(buf);
    }
    
    flip();
    
    TAT(tsf_stream_file_input_get_mark(input,&zeromark));
    
    for (n=num;n-->0;) {
        TEST_ASSERT_TSFERROR(
            buf=tsf_stream_file_input_read(input));
        
        TEST_ASSERT(tsf_type_get_kind_code(tsf_buffer_get_type(buf))
                    == TSF_TK_VOID);
        
        tsf_buffer_destroy(buf);
    }
    
    t_end_input();
    
    TAT(tsf_stream_file_input_seek(input,&zeromark));
    
    for (n=num;n-->0;) {
        TEST_ASSERT_TSFERROR(
            buf=tsf_stream_file_input_read(input));
        
        TEST_ASSERT(tsf_type_get_kind_code(tsf_buffer_get_type(buf))
                    == TSF_TK_VOID);
        
        tsf_buffer_destroy(buf);
    }
    
    teardown();
}

static void test_seeks_over_ifd_read(uint32_t i,
                                     uint32_t num_ints,
                                     uint32_t num_floats,
                                     uint32_t num_doubles) {
    tsf_buffer_t *buf;
    tsf_reflect_t *r;
    
    TAT(buf=tsf_stream_file_input_read(input));
    
    TAT(r=tsf_reflect_from_buffer(buf));
    
    tsf_buffer_destroy(buf);
    
    if (i<num_ints) {
        TA(tsf_type_get_kind_code(tsf_reflect_get_type(r))
           == TSF_TK_UINT32);
        
        TA(tsf_primitive_reflect_get(r,uint32_t) == i);
    } else if (i<num_ints+num_floats) {
        TA(tsf_type_get_kind_code(tsf_reflect_get_type(r))
           == TSF_TK_FLOAT);
        
        TA(tsf_primitive_reflect_get(r,float) == (float)i);
    } else {
        TA(tsf_type_get_kind_code(tsf_reflect_get_type(r))
           == TSF_TK_DOUBLE);
        
        TA(tsf_primitive_reflect_get(r,double) == (double)i);
    }
    
    tsf_reflect_destroy(r);
}

static void test_seeks_over_ifd(uint32_t num_ints,
                                uint32_t num_floats,
                                uint32_t num_doubles,
                                uint32_t first_read,
                                ...) {
    tsf_buffer_t *buf;
    tsf_reflect_t *r;
    
    uint32_t i;
    uint32_t arg;
    
    va_list lst;
    
    tsf_stream_file_mark_t *marks=
        malloc(sizeof(tsf_stream_file_mark_t)*
        (num_ints+num_floats+num_doubles+1));
    
    tsf_stream_file_mark_t temp_mark;
    
    t_reset();
    t_open_output(tsf_false);
    
    for (i=0;i<num_ints+num_floats+num_doubles;++i) {
        if (i<num_ints) {
            TAT(r=tsf_reflect_create_uint32_t(i));
        } else if (i<num_ints+num_floats) {
            TAT(r=tsf_reflect_create_float((float)i));
        } else {
            TAT(r=tsf_reflect_create_double((double)i));
        }
        
        TAT(buf=tsf_reflect_make_buffer(r));
        
        tsf_reflect_destroy(r);
        
        TAT(tsf_stream_file_output_write(output,buf));
        
        tsf_buffer_destroy(buf);
    }
    
    t_close_output();
    t_open_input();
    
    for (i=0;i<first_read;++i) {
        TAT(tsf_stream_file_input_get_mark(input,marks+i));
        
        test_seeks_over_ifd_read(i,
                                 num_ints,
                                 num_floats,
                                 num_doubles);
    }
    
    TAT(tsf_stream_file_input_get_mark(input,marks+first_read));
    
    if (first_read == num_ints+num_floats+num_doubles) {
        t_end_input();
    }
    
    va_start(lst,first_read);
    
    i=first_read-1;
    
    while ((arg=va_arg(lst,uint32_t)) != -1) {
        i=arg;
        
        TAT(tsf_stream_file_input_seek(input,marks+i));
        
        if (i == num_ints+num_floats+num_doubles) {
            TET(tsf_stream_file_input_read(input), TSF_E_EOF);
            continue;
        }
        
        test_seeks_over_ifd_read(i,
                                 num_ints,
                                 num_floats,
                                 num_doubles);

        /* some white-box fun! */
        if (i<=first_read) {
            TAT(tsf_stream_file_input_get_mark(input,&temp_mark));
            TA(temp_mark.offset == marks[i+1].offset);
            TA(temp_mark.state == marks[i+1].state);
        }
    }
    
    va_end(lst);
    
    /* ok, we are at position i+1 */
    for (++i;i<num_ints+num_floats+num_doubles;++i) {
        test_seeks_over_ifd_read(i,
                                 num_ints,
                                 num_floats,
                                 num_doubles);
    }
    
    t_end_input();
    t_close_input();
    t_reset();
    
    free(marks);
}

static void do_non_seek_non_append_tests() {
    DO_TEST(test_n_void(0));
    DO_TEST(test_n_void(1));
    DO_TEST(test_n_void(2));
    DO_TEST(test_n_void(10));
    
    DO_TEST(test_n_different_types(0,0));
    DO_TEST(test_n_different_types(1,0));
    DO_TEST(test_n_different_types(2,0));
    DO_TEST(test_n_different_types(10,0));
    DO_TEST(test_n_different_types(65537,0));
    
    DO_TEST(test_n_different_types(1,1));
    DO_TEST(test_n_different_types(2,1));
    DO_TEST(test_n_different_types(10,1));
    
    DO_TEST(test_n_different_types(1,2));
    DO_TEST(test_n_different_types(2,2));
    DO_TEST(test_n_different_types(10,2));
    
    DO_TEST(test_n_different_types(1,10));
    DO_TEST(test_n_different_types(2,10));
    DO_TEST(test_n_different_types(10,10));
}

int main(int c,char **v) {
    int i,j;
    
    TEST_INIT(c,v);
    
    tsf_zip_rdr_attr_get_default(&zip_rdr_attr);
    
    TEST_PUSH("zip=none");
    tsf_zip_wtr_attr_get_default(&zip_wtr_attr);
    do_non_seek_non_append_tests();
    TEST_POP();
    
    TEST_PUSH("zip=zlib");
    if (tsf_zip_wtr_attr_get_zlib_default(&zip_wtr_attr)) {
        do_non_seek_non_append_tests();
    }
    TEST_POP();
    
    TEST_PUSH("zip=bzip2");
    if (tsf_zip_wtr_attr_get_bzip2_default(&zip_wtr_attr)) {
        do_non_seek_non_append_tests();
    }
    TEST_POP();
    
    tsf_zip_wtr_attr_get_default(&zip_wtr_attr);

    tsf_zip_mode_t themodes[3];
    themodes[0]=TSF_ZIP_NONE;
    themodes[1]=TSF_ZIP_ZLIB;
    themodes[2]=TSF_ZIP_BZIP2;

#define nmodes (sizeof(themodes)/sizeof(tsf_zip_mode_t))

    for (i=1;i<4;++i) {
	int *modeidxlist=malloc(sizeof(int)*i);
	tsf_zip_mode_t *modelist=malloc(sizeof(tsf_zip_mode_t)*i);
	char *str=malloc(i+1);
	if (modeidxlist==NULL || modelist==NULL) {
	    fprintf(stderr,
		    "Could not allocate modelists: %s\n",
		    strerror(errno));
	    return 1;
	}
	for (j=0;j<i;++j) {
	    modeidxlist[j]=0;
	}
	for (;;) {
	    for (j=0;j<i;++j) {
		modelist[j]=themodes[modeidxlist[j]];
		str[j]='0'+modeidxlist[j];
	    }
	    str[i]=0;
	    TEST_PUSH("modeidxlist=%s",str);
	    DO_TEST(test_n_void_appends(i,modelist));
	    TEST_POP();
	    for (j=0;j<i;++j) {
		modeidxlist[j]++;
		if (modeidxlist[j]<nmodes) {
		    break;
		}
		modeidxlist[j]=0;
	    }
	    if (j==i) {
		break;
	    }
	}
    }
    
    tsf_zip_wtr_attr_get_default(&zip_wtr_attr);
    tsf_zip_rdr_attr_get_nozip(&zip_rdr_attr);

    DO_TEST(test_n_void_with_rewind(0));
    DO_TEST(test_n_void_with_rewind(1));
    DO_TEST(test_n_void_with_rewind(2));
    DO_TEST(test_n_void_with_rewind(10));
    
    /* tests with one type */
    DO_TEST(test_seeks_over_ifd(0,0,0,0+0+0,-1));
    DO_TEST(test_seeks_over_ifd(0,0,0,0+0+0,0,-1));
    DO_TEST(test_seeks_over_ifd(1,0,0,1+0+0,0,-1));
    DO_TEST(test_seeks_over_ifd(1,0,0,1+0+0,1,-1));
    DO_TEST(test_seeks_over_ifd(1,0,0,1+0+0,0,1,-1));
    DO_TEST(test_seeks_over_ifd(1,0,0,1+0+0,1,0,-1));
    DO_TEST(test_seeks_over_ifd(2,0,0,2+0+0,0,-1));
    DO_TEST(test_seeks_over_ifd(2,0,0,2+0+0,1,-1));
    DO_TEST(test_seeks_over_ifd(2,0,0,2+0+0,2,-1));
    DO_TEST(test_seeks_over_ifd(2,0,0,2+0+0,0,1,2,-1));
    DO_TEST(test_seeks_over_ifd(2,0,0,2+0+0,2,1,0,-1));
    DO_TEST(test_seeks_over_ifd(10,0,0,10+0+0,0,-1));
    DO_TEST(test_seeks_over_ifd(10,0,0,10+0+0,1,-1));
    DO_TEST(test_seeks_over_ifd(10,0,0,10+0+0,9,-1));
    DO_TEST(test_seeks_over_ifd(10,0,0,10+0+0,10,-1));
    DO_TEST(test_seeks_over_ifd(10,0,0,10+0+0,0,1,2,3,4,5,6,7,8,9,10,-1));
    DO_TEST(test_seeks_over_ifd(10,0,0,10+0+0,10,9,8,7,6,5,4,3,2,1,0,-1));
    
    /* tests with the other type */
    DO_TEST(test_seeks_over_ifd(0,1,0,0+1+0,0,-1));
    DO_TEST(test_seeks_over_ifd(0,1,0,0+1+0,1,-1));
    DO_TEST(test_seeks_over_ifd(0,1,0,0+1+0,0,1,-1));
    DO_TEST(test_seeks_over_ifd(0,1,0,0+1+0,1,0,-1));
    DO_TEST(test_seeks_over_ifd(0,2,0,0+2+0,0,-1));
    DO_TEST(test_seeks_over_ifd(0,2,0,0+2+0,1,-1));
    DO_TEST(test_seeks_over_ifd(0,2,0,0+2+0,2,-1));
    DO_TEST(test_seeks_over_ifd(0,2,0,0+2+0,0,1,2,-1));
    DO_TEST(test_seeks_over_ifd(0,2,0,0+2+0,2,1,0,-1));
    DO_TEST(test_seeks_over_ifd(0,10,0,0+10+0,0,-1));
    DO_TEST(test_seeks_over_ifd(0,10,0,0+10+0,1,-1));
    DO_TEST(test_seeks_over_ifd(0,10,0,0+10+0,9,-1));
    DO_TEST(test_seeks_over_ifd(0,10,0,0+10+0,10,-1));
    DO_TEST(test_seeks_over_ifd(0,10,0,0+10+0,0,1,2,3,4,5,6,7,8,9,10,-1));
    DO_TEST(test_seeks_over_ifd(0,10,0,0+10+0,10,9,8,7,6,5,4,3,2,1,0,-1));
    
    /* tests with the third type */
    DO_TEST(test_seeks_over_ifd(0,0,1,0+0+1,0,-1));
    DO_TEST(test_seeks_over_ifd(0,0,1,0+0+1,1,-1));
    DO_TEST(test_seeks_over_ifd(0,0,1,0+0+1,0,1,-1));
    DO_TEST(test_seeks_over_ifd(0,0,1,0+0+1,1,0,-1));
    DO_TEST(test_seeks_over_ifd(0,0,2,0+0+2,0,-1));
    DO_TEST(test_seeks_over_ifd(0,0,2,0+0+2,1,-1));
    DO_TEST(test_seeks_over_ifd(0,0,2,0+0+2,2,-1));
    DO_TEST(test_seeks_over_ifd(0,0,2,0+0+2,0,1,2,-1));
    DO_TEST(test_seeks_over_ifd(0,0,2,0+0+2,2,1,0,-1));
    DO_TEST(test_seeks_over_ifd(0,0,10,0+0+10,0,-1));
    DO_TEST(test_seeks_over_ifd(0,0,10,0+0+10,1,-1));
    DO_TEST(test_seeks_over_ifd(0,0,10,0+0+10,9,-1));
    DO_TEST(test_seeks_over_ifd(0,0,10,0+0+10,10,-1));
    DO_TEST(test_seeks_over_ifd(0,0,10,0+0+10,0,1,2,3,4,5,6,7,8,9,10,-1));
    DO_TEST(test_seeks_over_ifd(0,0,10,0+0+10,10,9,8,7,6,5,4,3,2,1,0,-1));
    
    /* tests with first two types */
    DO_TEST(test_seeks_over_ifd(1,1,0,1+1+0,0,-1));
    DO_TEST(test_seeks_over_ifd(1,1,0,1+1+0,1,-1));
    DO_TEST(test_seeks_over_ifd(1,1,0,1+1+0,2,-1));
    DO_TEST(test_seeks_over_ifd(1,1,0,1+1+0,0,1,2,-1));
    DO_TEST(test_seeks_over_ifd(1,1,0,1+1+0,2,1,0,-1));
    
    DO_TEST(test_seeks_over_ifd(1,1,0,1,0,-1));
    
    DO_TEST(test_seeks_over_ifd(5,5,0,5+5+0,0,-1));
    DO_TEST(test_seeks_over_ifd(5,5,0,5+5+0,1,-1));
    DO_TEST(test_seeks_over_ifd(5,5,0,5+5+0,4,-1));
    DO_TEST(test_seeks_over_ifd(5,5,0,5+5+0,5,-1));
    DO_TEST(test_seeks_over_ifd(5,5,0,5+5+0,9,-1));
    DO_TEST(test_seeks_over_ifd(5,5,0,5+5+0,10,-1));
    DO_TEST(test_seeks_over_ifd(5,5,0,5+5+0,0,1,2,3,4,5,6,7,8,9,10,-1));
    DO_TEST(test_seeks_over_ifd(5,5,0,5+5+0,10,9,8,7,6,5,4,3,2,1,0,-1));
    
    DO_TEST(test_seeks_over_ifd(5,5,0,1,0,-1));
    DO_TEST(test_seeks_over_ifd(5,5,0,5,0,-1));
    
    /* tests with all types */
    DO_TEST(test_seeks_over_ifd(1,1,1,1+1+1,0,-1));
    DO_TEST(test_seeks_over_ifd(1,1,1,1+1+1,1,-1));
    DO_TEST(test_seeks_over_ifd(1,1,1,1+1+1,2,-1));
    DO_TEST(test_seeks_over_ifd(1,1,1,1+1+1,0,1,2,-1));
    DO_TEST(test_seeks_over_ifd(1,1,1,1+1+1,2,1,0,-1));
    
    DO_TEST(test_seeks_over_ifd(1,1,1,1,0,-1));
    DO_TEST(test_seeks_over_ifd(1,1,1,2,0,-1));
    DO_TEST(test_seeks_over_ifd(1,1,1,2,1,-1));
    
    DO_TEST(test_seeks_over_ifd(3,3,3,3+3+3,0,-1));
    DO_TEST(test_seeks_over_ifd(3,3,3,3+3+3,1,-1));
    DO_TEST(test_seeks_over_ifd(3,3,3,3+3+3,4,-1));
    DO_TEST(test_seeks_over_ifd(3,3,3,3+3+3,5,-1));
    DO_TEST(test_seeks_over_ifd(3,3,3,3+3+3,8,-1));
    DO_TEST(test_seeks_over_ifd(3,3,3,3+3+3,9,-1));
    DO_TEST(test_seeks_over_ifd(3,3,3,3+3+3,0,1,2,3,4,5,6,7,8,9,-1));
    DO_TEST(test_seeks_over_ifd(3,3,3,3+3+3,9,8,7,6,5,4,3,2,1,0,-1));
    
    DO_TEST(test_seeks_over_ifd(3,3,3,1,0,-1));
    DO_TEST(test_seeks_over_ifd(3,3,3,3,0,-1));
    DO_TEST(test_seeks_over_ifd(3,3,3,4,0,-1));
    DO_TEST(test_seeks_over_ifd(3,3,3,6,0,-1));
    DO_TEST(test_seeks_over_ifd(3,3,3,4,3,-1));
    DO_TEST(test_seeks_over_ifd(3,3,3,6,3,-1));
    
    TEST_DONE();
    
    return 0;
}



