#include "proto_ir.pb.h"
#include <iostream>
#include <fcntl.h>
#include <fstream>
#include <string>
#include <sys/time.h>
#include <tsf/tsf.h>

using namespace std;

static double milliTime() {
    struct timeval tv;
    gettimeofday(&tv, NULL);
    return tv.tv_sec * 1000. + tv.tv_usec / 1000.;
}

#define TIMEIT(statement) do {                                          \
        double __ti_before = milliTime();                               \
        statement;                                                      \
        double __ti_after = milliTime();                                \
        printf("%s: %.3lf ms\n", #statement, __ti_after - __ti_before); \
    } while (0)

/* If the expression evaluates false then it assumes an error and exits. */
#define CT(exp) do {                                                    \
        if (!(exp)) {                                                   \
            fprintf(stderr, "%s:%d: %s: %s\n",                          \
                    __FILE__, __LINE__, #exp, tsf_get_error());         \
            exit(1);                                                    \
        }                                                               \
    } while (0)
#define CS(exp) do {                                                    \
        if (!(exp)) {                                                   \
            int myErrno = errno;                                        \
            fprintf(stderr, "%s:%d: %s: %s\n",                          \
                    __FILE__, __LINE__, #exp, strerror(myErrno));       \
            exit(1);                                                    \
        }                                                               \
    } while (0)

static const char *filename = "proto_ir_speed_test_file.proto";

static void writeTest(unsigned programSize, unsigned numPrograms) {
    int fd;
    tsf_buf_wtr_t *writer;
    
    CS((fd = open(filename, O_CREAT | O_TRUNC | O_WRONLY, 0644)) >= 0);
    
    CT(writer = tsf_buf_writer_create(fd, TSF_DEF_BUF_SIZE, tsf_true));
    
    char buf[256];
    
    for (unsigned programIndex = 0; programIndex < numPrograms; ++programIndex) {
        const unsigned numDecls = 25;
        const unsigned numDefns = 25;
        
        google::protobuf::Arena arena;
        Program& program = *google::protobuf::Arena::CreateMessage<Program>(&arena);

        for (unsigned elementIndex = 0; elementIndex < numDecls; ++elementIndex) {
            ProgramElement *element = program.add_globals();
            ProcedureDecl *procedure = element->mutable_proceduredecl();
            snprintf(buf, sizeof(buf), "foo%u", elementIndex);
            procedure->set_name(buf);
        }
        
        for (unsigned elementIndex = numDecls;
             elementIndex < numDecls + numDefns;
             ++elementIndex) {
            ProgramElement *element = program.add_globals();
            ProcedureDefn *procedure = element->mutable_proceduredefn();
            
            unsigned numVariables = (programIndex % programSize) + 1;
            unsigned numInstructions = (programIndex % programSize) * 2 + 1;
            unsigned numDebugDatas = (numInstructions + 5) / 6;
            
            snprintf(buf, sizeof(buf), "bar%u", elementIndex);
            procedure->set_name(buf);
            
            for (unsigned variableIndex = 0; variableIndex < numVariables; ++variableIndex) {
                VariableDecl *variable = procedure->add_variables();
                
                switch ((variableIndex + programIndex) % 3) {
                case 0:
                    variable->set_type("foo");
                    break;
                case 1:
                    variable->set_type("bar");
                    break;
                case 2:
                    variable->set_type("baz");
                    break;
                default:
                    abort();
                    break;
                }
                
                snprintf(buf, sizeof(buf), "x%u", variableIndex);
                variable->set_name(buf);
            }
            
            for (unsigned instructionIndex = 0;
                 instructionIndex < numInstructions;
                 ++instructionIndex) {
                Instruction *instruction = procedure->add_code();
                
                switch ((instructionIndex + programIndex) % 8 + 1) {
                case Instruction::kNop: {
                    instruction->mutable_nop();
                    break;
                }
                        
                case Instruction::kMov: {
                    Instruction_Mov *mov = instruction->mutable_mov();
                    mov->set_dest(
                        (instructionIndex + programIndex) % numVariables);
                    mov->set_src(
                        (instructionIndex + programIndex + 1) % numVariables);
                    break;
                }
                        
                case Instruction::kAdd: {
                    Instruction_Add *add = instruction->mutable_add();
                    add->set_dest(
                        (instructionIndex + programIndex + 2) % numVariables);
                    add->set_src(
                        (instructionIndex + programIndex + 3) % numVariables);
                    break;
                }
                        
                case Instruction::kAlloc: {
                    Instruction_Alloc *alloc = instruction->mutable_alloc();

                    snprintf(buf, sizeof(buf), "t%u", instructionIndex + programIndex);
                    alloc->set_type(buf);
                    break;
                }
                        
                case Instruction::kRet: {
                    Instruction_Ret *ret = instruction->mutable_ret();
                    ret->set_src(
                        (instructionIndex + programIndex + 4) % numVariables);
                    break;
                }
                        
                case Instruction::kJump: {
                    Instruction_Jump *jump = instruction->mutable_jump();
                    jump->set_target(
                        (instructionIndex + programIndex) % numInstructions);
                    break;
                }
                        
                case Instruction::kCall: {
                    Instruction_Call *call = instruction->mutable_call();
                    
                    unsigned numArgs = (instructionIndex + programIndex) % 10;
                        
                    call->set_dest(
                        (instructionIndex + programIndex + 7) % numVariables);
                    call->set_callee(
                        ((instructionIndex + programIndex) %
                         (numVariables + numDecls + numDefns)) -
                        numDecls - numDefns);
                    
                    for (unsigned argIndex = 0; argIndex < numArgs; ++argIndex) {
                        call->add_arg(
                            (instructionIndex + programIndex + argIndex) % numVariables);
                    }
                    break;
                }
                        
                case Instruction::kBranchZero: {
                    Instruction_BranchZero *branchZero = instruction->mutable_branchzero();
                    branchZero->set_src(
                        (instructionIndex + programIndex + 5) % numVariables);
                    branchZero->set_target(
                        (instructionIndex + programIndex + 6) % numInstructions);
                    break;
                }
                        
                default:
                    abort();
                    break;
                }
            }
            
            for (unsigned debugDataIndex = 0;
                 debugDataIndex < numDebugDatas;
                 ++debugDataIndex) {
                DebugData *debugData = procedure->add_debug();
                
                debugData->set_startoffset(debugDataIndex);
                debugData->set_spansize(1);
                
                snprintf(buf, sizeof(buf), "debug%u", debugDataIndex);
                debugData->set_data(buf);
            }
        }
        
        string buffer;
        program.SerializeToString(&buffer);
        
        uint32_t length = buffer.length();
        CT(tsf_buf_writer_write(writer, &length, sizeof(length)));
        CT(tsf_buf_writer_write(writer, (void*)buffer.data(), length));
    }
    
    CT(tsf_buf_writer_flush(writer));
    tsf_buf_writer_destroy(writer);
}

static void readTest(unsigned numPrograms) {
    int fd;
    tsf_buf_rdr_t *reader;
    
    CS((fd = open(filename, O_RDONLY)) >= 0);
    CT(reader = tsf_buf_reader_create(fd, TSF_DEF_BUF_SIZE, tsf_true));
    
    unsigned count = 0;
    for (;;) {
        google::protobuf::Arena arena;
        Program& program = *google::protobuf::Arena::CreateMessage<Program>(&arena);

        uint32_t length;
        if (!tsf_buf_reader_read(reader, &length, sizeof(length))) {
            CT(tsf_get_error_code() == TSF_E_UNEXPECTED_EOF);
            break;
        }

        char* buffer;
        CS(buffer = (char*)malloc(length));
        CT(tsf_buf_reader_read(reader, buffer, length));
        
        program.ParseFromArray(buffer, length);
        free(buffer);
        count++;
    }

    if (count != numPrograms)
        abort();
    
    tsf_buf_reader_destroy(reader);
}

int main(int c, char **v) {
    GOOGLE_PROTOBUF_VERIFY_VERSION;
    
    TIMEIT(writeTest(100, 10000));
    TIMEIT(readTest(10000));
    
    return 0;
}

