#include "tsf_tuple.h"
#include <errno.h>
#include <inttypes.h>
#include <stdio.h>
#include <sys/time.h>
#ifdef TSF_BUILD_SYSTEM
#include "tsf.h"
#include "tsf_format.h"
#else
#include <tsf/tsf.h>
#include <tsf/tsf_format.h>
#endif

static void usage(void) {
    fprintf(stderr, "Usage: tsf_tuple_speed [<count>]\n");
    exit(1);
}

static double milliTime(void) {
    struct timeval tv;
    gettimeofday(&tv, NULL);
    return tv.tv_sec * 1000. + tv.tv_usec / 1000.;
}

#define TIMEIT(statement) do {                                          \
        double __ti_before = milliTime();                               \
        statement;                                                      \
        double __ti_after = milliTime();                                \
        printf("%s: %.3lf ms\n", #statement, __ti_after - __ti_before); \
    } while (0)

/* If the expression evaluates false then it assumes an error and exits. */
#define CT(exp) do {                                                    \
        if (!(exp)) {                                                   \
            fprintf(stderr, "%s:%d: %s: %s\n",                          \
                    __FILE__, __LINE__, #exp, tsf_get_error());         \
            exit(1);                                                    \
        }                                                               \
    } while (0)
#define CS(exp) do {                                                    \
        if (!(exp)) {                                                   \
            int myErrno = errno;                                        \
            fprintf(stderr, "%s:%d: %s: %s\n",                          \
                    __FILE__, __LINE__, #exp, strerror(myErrno));       \
            exit(1);                                                    \
        }                                                               \
    } while (0)

static void writeTest(const char *filename, unsigned numTuples,
                      tsf_zip_mode_t zipMode) {
    tsf_zip_wtr_attr_t zipAttributes;
    tsf_stream_file_output_t *out;
    unsigned tupleIndex;
    
    CT(tsf_zip_wtr_attr_get_for_mode(&zipAttributes, zipMode));
    CT(out = tsf_stream_file_output_open(filename, tsf_false, &zipAttributes));
    
    for (tupleIndex = 0; tupleIndex < numTuples; ++tupleIndex) {
        Tuple_t tuple;
        tuple.foo = tupleIndex;
        tuple.bar = tupleIndex + 1;
        tuple.baz = tupleIndex + 2;
        tuple.thingy = tupleIndex + 3;
        
        CT(Tuple__write(out, &tuple));
    }
    
    CT(tsf_stream_file_output_close(out));
}

static void writeClonedTypeTest(const char *filename, unsigned numTuples,
                                tsf_zip_mode_t zipMode) {
    tsf_zip_wtr_attr_t zipAttributes;
    tsf_stream_file_output_t *out;
    Tuple_t tuple;
    unsigned tupleIndex;
    tsf_type_t *type;
    tsf_genrtr_t *generator;
    
    CT(tsf_zip_wtr_attr_get_for_mode(&zipAttributes, zipMode));
    CT(out = tsf_stream_file_output_open(filename, tsf_false, &zipAttributes));
    
    /* For the first tuple, we use the original type. */
    tuple.foo = 0;
    tuple.bar = 1;
    tuple.baz = 2;
    tuple.thingy = 3;
    CT(Tuple__write(out, &tuple));
    
    /* Now create a new type that is logically identical to our original type. */
    CT(type = tsf_type_create(TSF_TK_STRUCT));
    CT(tsf_native_struct_type_set_size(&type, sizeof(Tuple_t)));
    CT(tsf_struct_type_append(&type, "foo", tsf_type_create(TSF_TK_INTEGER)));
    CT(tsf_native_struct_type_map(&type, "foo", tsf_offsetof(Tuple_t, foo)));
    CT(tsf_struct_type_append(&type, "bar", tsf_type_create(TSF_TK_INT32)));
    CT(tsf_native_struct_type_map(&type, "bar", tsf_offsetof(Tuple_t, bar)));
    CT(tsf_struct_type_append(&type, "baz", tsf_type_create(TSF_TK_LONG)));
    CT(tsf_native_struct_type_map(&type, "baz", tsf_offsetof(Tuple_t, baz)));
    CT(tsf_struct_type_append(&type, "thingy", tsf_type_create(TSF_TK_FLOAT)));
    CT(tsf_native_struct_type_map(&type, "thingy", tsf_offsetof(Tuple_t, thingy)));
    
    if (!tsf_type_compare(Tuple__get_type(), type)) {
        fprintf(stderr, "Error: types not equal.\n");
        exit(1);
    }
    
    CT(generator = tsf_generator_create(type));
    
    for (tupleIndex = 1; tupleIndex < numTuples; ++tupleIndex) {
        Tuple_t tuple;
        tsf_buffer_t *buf;
        
        tuple.foo = tupleIndex;
        tuple.bar = tupleIndex + 1;
        tuple.baz = tupleIndex + 2;
        tuple.thingy = tupleIndex + 3;
        
        CT(buf = tsf_generator_generate(generator, &tuple));
        CT(tsf_stream_file_output_write(out, buf));
        tsf_buffer_destroy(buf);
    }
    
    CT(tsf_stream_file_output_close(out));
}

static void readTest(const char *filename, unsigned numTuples) {
    tsf_stream_file_input_t *in;
    unsigned count;
    uint64_t expectedTotal;
    uint64_t total;
    
    CT(in = tsf_stream_file_input_open(filename, NULL, NULL));
    
    count = 0;
    total = 0;
    
    for (;;) {
        Tuple_t tuple;
        if (!Tuple__read_into(in, &tuple)) {
            CT(tsf_get_error_code() == TSF_E_EOF);
            break;
        }
        
        total += (uint64_t)tuple.foo;
        total += (uint64_t)tuple.bar;
        total += (uint64_t)tuple.baz;
        total += (uint64_t)tuple.thingy;
        
        count++;
    }
    
    if (count != numTuples) {
        fprintf(stderr, "Wrong number of tuples read: %u, expected %u.\n", numTuples, count);
        exit(1);
    }
    
    expectedTotal = ((((uint64_t)count - 1) * count) / 2) * 4 + (1 + 2 + 3) * count;
    if (total != expectedTotal) {
        fprintf(stderr, "Wrong total: " fui64 ", expected " fui64 ".\n", total, expectedTotal);
        exit(1);
    }
    
    tsf_stream_file_input_close(in);
}

static void readClonedTypeTest(const char *filename, unsigned numTuples) {
    /* Simply running the same test a second time should cause us to deal with "cloned" types. */
    readTest(filename, numTuples);
}

int main(int c, char **v) {
    static const char *filename = "tsf_tuple_speed_test_file.tsf";
    static const char *zipFilename = "tsf_tuple_speed_test_zip_file.tsf";
    
    unsigned count;
    
    switch (c) {
    case 1:
        /* Use a small problem size suitable for regression testing. */
        count = 10000;
        break;
        
    case 2:
        if (sscanf(v[1], "%u", &count) != 1) {
            usage();
        }
        break;
        
    default:
        usage();
        return 1;
    }
    
    printf("Writing %u tuples.\n", count);
    if (count != 1000000)
        printf("WARNING: If you are benchmarking, please use count = 1000000.\n");
    
    TIMEIT(writeTest(filename, count, TSF_ZIP_NONE));
    TIMEIT(writeClonedTypeTest(filename, count, TSF_ZIP_NONE));
    TIMEIT(readTest(filename, count));
    TIMEIT(readClonedTypeTest(filename, count));
    
    if (tsf_zlib_supported()) {
        TIMEIT(writeTest(zipFilename, count, TSF_ZIP_ZLIB));
        TIMEIT(readTest(zipFilename, count));
    }
    
    /* We don't benchmark bzip2 because it's just too slow to be interesting. */
    
    return 0;
}

