/*
 * Copyright (C) 2011 Filip Pizlo. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY FILIP PIZLO ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL FILIP PIZLO OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "tsf_internal.h"

#if defined(HAVE_PTHREAD) && defined(HAVE_BSDSOCKS) && defined(HAVE_GETOPT)

#include <pthread.h>
#include <getopt.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <arpa/inet.h>
#include <netinet/tcp.h>
#include "tsf_fsdb_protocol.h"
#include "tsf_util.h"
#include "tsf_format.h"

/* functions and types we use */
static void init_log(void);
static void reroute_log(const char *flnm);
static void vlogit(const char *fmt,va_list lst);
static void pre_logit(void);
static void post_logit(void);
static void logit(const char *fmt,...);

typedef struct {
    int fd;
    int64_t id;
    int dbfd;
    char *partflnm;
    size_t num_bytes;
    tsf_buffer_t *key;
    struct sockaddr_in addr;
    uint32_t state;
    tsf_stream_file_input_t *in;
    tsf_stream_file_output_t *out;
    void *region;
    size_t buf_size;
    char *buf;
    tsf_bool_t autoflush;
} handler_data_t;

static void hlogit(handler_data_t *hdata,
                   const char *fmt,...);

static void reset_region(handler_data_t *hdata);

static void reset_state(handler_data_t *hdata);

static void send_response(handler_data_t *data,
                          tsf_fsdb_response_t *rsp);

static void handler_done(handler_data_t *data);
static void handler_fail_tsf(handler_data_t *data,tsf_bool_t send_rsp);
static void handler_fail_sys(handler_data_t *data,tsf_bool_t send_rsp);
static void handler_fail_proto(handler_data_t *data,
                               const char *reason);

static void send_simple_rsp(handler_data_t *data,
                            uint32_t cmd);
static void send_ok(handler_data_t *data);

static void *handler_thread_main(void *arg);

/* need some logging functionality. */
static int log_level;
static char *log_flnm;
static FILE *log_file;
static pthread_mutex_t log_lock;

static void init_log(void) {
    log_flnm=NULL;
    
    pthread_mutex_init(&log_lock,NULL);
    
    log_file=stderr;
    setvbuf(log_file,NULL,_IONBF,0);
    logit("TSF FSDB daemon starting!");
}

static void reroute_log(const char *flnm) {
    log_flnm=strdup(flnm);
    
    CHECKSYSZ(log_file=fopen(log_flnm,"a"));
#if defined(HAVE_SETVBUF)
    setvbuf(log_file,NULL,_IONBF,0);
#endif
    logit("TSF FSDB daemon starting!");
}

static void pre_logit(void) {
    time_t t;
    unsigned millis;
#ifdef HAVE_GETTIMEOFDAY
    struct timeval tv;
#endif
    struct tm *st;
#ifdef HAVE_LOCALTIME_R
    struct tm stbuf;
#endif

#ifdef HAVE_GETTIMEOFDAY
    gettimeofday(&tv,NULL);
    t=tv.tv_sec;
    millis=tv.tv_usec/1000;
#else
    time(&t);
    millis=0;
#endif

#ifdef HAVE_LOCALTIME_R
    st=localtime_r(&t,&stbuf);
#else
    st=localtime(&t);
#endif

    pthread_mutex_lock(&log_lock);
    fprintf(log_file,"[%02d/%02d/%04d %02d:%02d:%02d.%03d] ",
            st->tm_mday,st->tm_mon+1,st->tm_year+1900,st->tm_hour,st->tm_min,
            st->tm_sec,millis);
}

static void post_logit(void) {
    fprintf(log_file,"\n");
    pthread_mutex_unlock(&log_lock);
}

static void vlogit(const char *fmt,va_list lst) {
    pre_logit();
    vfprintf(log_file,fmt,lst);
    post_logit();
}

static void logit(const char *fmt,...) {
    va_list lst;
    va_start(lst,fmt);
    vlogit(fmt,lst);
    va_end(lst);
}

static void hlogit(handler_data_t *hdata,
                   const char *fmt,...) {
    va_list lst;
#ifdef HAVE_INET_NTOP
    char dstbuf[64];
#endif
    const char *hname;
    
    pre_logit();
#ifdef HAVE_INET_NTOP
    hname=inet_ntop(AF_INET,(const void*)&hdata->addr.sin_addr,
                    dstbuf,sizeof(dstbuf));
#else
    hname=inet_ntoa(&hdata->addr.sin_addr);
#endif
    fprintf(log_file,"client%05" PRIx64 "@%s: ",hdata->id,hname);
    va_start(lst,fmt);
    vfprintf(log_file,fmt,lst);
    va_end(lst);
    post_logit();
}

#undef CHECKTSF
#undef CHECKSYS
#undef CHECKSYSZ

#define CHECKTSF(exp) do {                                      \
        if (!(exp)) {                                           \
            logit("TSF Error at %s:%d: %s\n",                   \
                  __FILE__,__LINE__,tsf_get_error());           \
            exit(1);                                            \
        }                                                       \
    } while (0)

#define CHECKSYS(exp) do {                                  \
        if ((exp)<0) {                                      \
            logit("System Error at %s:%d: %s\n",            \
                  __FILE__,__LINE__,strerror(errno));       \
            exit(1);                                        \
        }                                                   \
    } while (0)

#define CHECKSYSZ(exp) do {                                 \
        if (!(exp)) {                                       \
            logit("System Error at %s:%d: %s\n",            \
                  __FILE__,__LINE__,strerror(errno));       \
            exit(1);                                        \
        }                                                   \
    } while (0)

static void reset_region(handler_data_t *hdata) {
    tsf_region_free(hdata->region);
    CHECKSYSZ(hdata->region=tsf_region_create(0));
}

/* scoreboard */
static pthread_mutex_t score_lock;
static unsigned n_connections;

/* constants for connections and such. */
static tsf_limits_t *limits;
static tsf_zip_wtr_attr_t wtr_attr;
static tsf_zip_rdr_attr_t rdr_attr;

/* the actual database. */
static tsf_fsdb_t *fsdb;

/* utilities for doing handler stuff */
static void send_response(handler_data_t *data,
                          tsf_fsdb_response_t *rsp) {
    if (log_level>=1) {
        hlogit(data,"sending %s in state %s",
               tsf_fsdb_response__payload__to_str(rsp->payload.value),
               tsf_fsdb_state__to_str(rsp->state.value));
    }
    
    if (!tsf_fsdb_response__write(data->out,rsp)) {
        handler_fail_tsf(data,tsf_false);
    }
    
    if (data->autoflush) {
        if (!tsf_stream_file_output_flush(data->out)) {
            handler_fail_tsf(data,tsf_false);
        }
    }
}

static void reset_state(handler_data_t *data) {
    if (data->key!=NULL) {
        tsf_buffer_destroy(data->key);
        data->key=NULL;
    }
    if (data->dbfd>=0) {
        CHECKSYS(close(data->dbfd));
        data->dbfd=-1;
    }
    if (data->partflnm!=NULL) {
        hlogit(data,
               "aborting transaction for %s after writing " fsz " bytes",
               data->partflnm,
               data->num_bytes);
        tsf_fsdb_guts_abort(fsdb,data->partflnm);
        free(data->partflnm);
        data->partflnm=NULL;
    }
    data->state=tsf_fsdb_state__clear;
    data->num_bytes=0;
}

static void handler_done(handler_data_t *data) {
    pthread_mutex_lock(&score_lock);
    n_connections--;
    pthread_mutex_unlock(&score_lock);
    
    hlogit(data,
           "disconnecting; stats: gpc load = %d, num connections = %u.",
           tsf_parser_num_codes(tsf_fsdb_command__get_parser()),
           n_connections);
    tsf_stream_file_output_close(data->out);
    tsf_stream_file_input_close(data->in);
    CHECKSYS(close(data->fd));
    tsf_region_free(data->region);
    reset_state(data);
    if (data->buf!=NULL) {
        free(data->buf);
    }
    free(data);
    pthread_exit(NULL);
    
    tsf_abort("should not get here");
}

static void handler_fail_tsf(handler_data_t *data,
                             tsf_bool_t send_rsp) {
    char *err=tsf_region_strdup(data->region,tsf_get_error());
    CHECKSYSZ(err);
    
    hlogit(data,"got TSF error: %s",err);
    
    if (send_rsp) {
        tsf_fsdb_response_t rsp;
        rsp.state.value=data->state;
        rsp.payload.value=tsf_fsdb_response__payload__sys_error;
        rsp.payload.u.sys_error.msg=err;
        send_response(data,&rsp);
    }
    
    handler_done(data);
    
    tsf_abort("should not get here");
}

static void handler_fail_sys(handler_data_t *data,
                             tsf_bool_t send_rsp) {
    char *err=tsf_region_strdup(data->region,strerror(errno));
    CHECKSYSZ(err);
    
    hlogit(data,"got system error: %s",err);
    
    if (send_rsp) {
        tsf_fsdb_response_t rsp;
        rsp.state.value=data->state;
        rsp.payload.value=tsf_fsdb_response__payload__sys_error;
        rsp.payload.u.sys_error.msg=err;
        send_response(data,&rsp);
    }
    
    handler_done(data);
    
    tsf_abort("should not get here");
}

static void handler_fail_proto(handler_data_t *data,
                               const char *reason) {
    tsf_fsdb_response_t rsp;
    
    hlogit(data,"got protocol error: %s",reason);
    
    rsp.state.value=data->state;
    rsp.payload.value=tsf_fsdb_response__payload__proto_error;
    rsp.payload.u.proto_error.msg=(char*)reason;
    send_response(data,&rsp);
    
    handler_done(data);
    
    tsf_abort("should not get here");
}

static void send_simple_rsp(handler_data_t *data,
                            uint32_t cmd) {
    tsf_fsdb_response_t rsp;
    
    rsp.state.value=data->state;
    rsp.payload.value=cmd;
    
    send_response(data,&rsp);
}

static void send_ok(handler_data_t *data) {
    send_simple_rsp(data,tsf_fsdb_response__payload__ok);
}

/* the thread that does the handling */
static void *handler_thread_main(void *arg) {
    handler_data_t *data;
    
    tsf_fsdb_command_t *cmd;
    tsf_fsdb_response_t rsp;
    tsf_buffer_t *buf;
    
    int flag;
    
    data=(handler_data_t*)arg;
    
    hlogit(data,"connection accepted; num connections = %u.",n_connections);
    
    flag=1;
    CHECKSYS(setsockopt(data->fd, IPPROTO_TCP, TCP_NODELAY, (char*)&flag, sizeof(flag)));
    
    data->dbfd=-1;
    data->partflnm=NULL;
    data->key=NULL;
    data->buf=NULL;
    data->buf_size=0;
    data->autoflush=tsf_true;
    data->num_bytes=0;
    
    data->state=tsf_fsdb_state__clear;
    CHECKSYSZ(data->region=tsf_region_create(0));
    CHECKTSF(data->out=tsf_stream_file_output_fd_open(
                 data->fd,&wtr_attr,tsf_false,TSF_DEF_NET_BUF_SIZE));
    CHECKTSF(data->in=tsf_stream_file_input_fd_open(
                 data->fd,limits,&rdr_attr,tsf_false,TSF_DEF_NET_BUF_SIZE));
    
    rsp.state.value=data->state;
    rsp.payload.value=tsf_fsdb_response__payload__welcome;
    rsp.payload.u.welcome.name="tsf_fsdb_daemon";
    rsp.payload.u.welcome.version.major=tsf_version_major();
    rsp.payload.u.welcome.version.minor=tsf_version_minor();
    rsp.payload.u.welcome.version.patch=tsf_version_patch();
    send_response(data,&rsp);
    
    for (;;) {
        reset_region(data);
        
        buf=tsf_stream_file_input_read(data->in);
        if (buf==NULL) {
            if (tsf_get_error_code()==TSF_E_EOF) {
                handler_done(data);
            } else {
                handler_fail_tsf(data,tsf_false);
            }
        }
        
        cmd=tsf_fsdb_command__parse(buf);
        if (cmd==NULL) {
            /* this could be a protocol error */
            char *err=tsf_region_strdup(data->region,tsf_get_error());
            CHECKSYSZ(err);
            
            tsf_buffer_destroy(buf);
            
            handler_fail_proto(data,err);
        }
        
        tsf_buffer_destroy(buf);
        
        tsf_region_add_cback(data->region,
                             (tsf_region_cback_t)tsf_region_free,
                             cmd);
        
        if (log_level>=1) {
            hlogit(data,"received %s",tsf_fsdb_command__to_str(cmd->value));
        }
        
        switch (cmd->value) {
        case tsf_fsdb_command__ping:
            send_ok(data);
            break;
        case tsf_fsdb_command__put:
            if (data->state!=tsf_fsdb_state__clear) {
                handler_fail_proto(data,
                                   "cannot use put command outside of clear state");
            }
            
            data->key=tsf_buffer_dup(cmd->u.put.key);
            if (data->key==NULL) {
                handler_fail_tsf(data,tsf_true);
            }
            
            if (!tsf_fsdb_guts_put(fsdb,
                                   cmd->u.put.key,
                                   &data->dbfd,
                                   &data->partflnm)) {
                data->dbfd=-1;
                data->partflnm=NULL;
                handler_fail_tsf(data,tsf_true);
            }
            
            if (log_level>=1) {
                hlogit(data,"beginning transaction for %s",data->partflnm);
            }

            data->state=tsf_fsdb_state__put;
            send_ok(data);
            break;
        case tsf_fsdb_command__out_write:
            if (data->state!=tsf_fsdb_state__put) {
                handler_fail_proto(data,
                                   "cannot use out_write command outside of put state");
            }
            
            if (log_level>=1) {
                hlogit(data,
                       "writing " fui32 " bytes to %s",
                       cmd->u.out_write.data.len,
                       data->partflnm);
            }
            
            if (!tsf_fd_writer(&data->dbfd,
                               cmd->u.out_write.data.data,
                               cmd->u.out_write.data.len)) {
                handler_fail_tsf(data,tsf_true);
            }
            
            data->num_bytes+=cmd->u.out_write.data.len;
            
            send_ok(data);
            break;
        case tsf_fsdb_command__out_commit: {
            char *dataflnm;
            
            if (data->state!=tsf_fsdb_state__put) {
                handler_fail_proto(data,
                                   "cannot use out_commit command outside of put state");
            }
            
            if (log_level>=1) {
                hlogit(data,
                       "committing transaction for %s after writing " fsz " bytes",
                       data->partflnm,
                       data->num_bytes);
            }
            
            close(data->dbfd);
            data->dbfd=-1;
            
            if (!tsf_fsdb_guts_commit(fsdb,data->partflnm,data->key,&dataflnm)) {
                handler_fail_tsf(data,tsf_true);
            }
            
            hlogit(data,
                   "committed transaction to %s with " fsz " bytes",
                   dataflnm,
                   data->num_bytes);
            free(dataflnm);
            
            tsf_buffer_destroy(data->key);
            data->key=NULL;
            free(data->partflnm);
            data->partflnm=NULL;
            data->num_bytes=0;
            
            data->state=tsf_fsdb_state__clear;
            
            send_ok(data);
            break;
        }
        case tsf_fsdb_command__get: {
            char *dataflnm;
            
            if (data->state!=tsf_fsdb_state__clear) {
                handler_fail_proto(data,
                                   "cannot use get command outside of clear state");
            }
            
            if (!tsf_fsdb_guts_get(fsdb,cmd->u.get.key,&data->dbfd,&dataflnm)) {
                data->dbfd=-1;
                if (tsf_get_error_code()==TSF_E_ELE_NOT_FOUND) {
                    send_simple_rsp(data,tsf_fsdb_response__payload__not_found);
                    break;
                } else {
                    handler_fail_tsf(data,tsf_true);
                }
            }
            
            if (log_level>=1) {
                hlogit(data,"retrieving data from %s",dataflnm);
            }
            free(dataflnm);
            
            data->state=tsf_fsdb_state__get;
            
            data->autoflush=tsf_false;
            
            send_ok(data);
            
            /* how to do this?  ideally we would do this in a way that
             * allows for asynchronous reset.  or we could just ignore that
             * feature for now.  if you really want a reset then you can just
             * disconnect.... */
            
            /* ok, just blitz the data.  we can afford a big buffer so lets
             * just do that. */
            if (data->buf==NULL) {
                data->buf_size=TSF_DEF_NET_BUF_SIZE;
                data->buf=malloc(data->buf_size);
                CHECKSYSZ(data->buf);
            }
            for (;;) {
                int res=read(data->dbfd,data->buf,data->buf_size);
                if (res<0) {
                    handler_fail_sys(data,tsf_true);
                } else if (res==0) {
                    /* done! */
                    break;
                } else {
                    tsf_fsdb_response_t rsp;
                    rsp.state.value=data->state;
                    rsp.payload.value=tsf_fsdb_response__payload__in_data;
                    rsp.payload.u.in_data.data.data=(uint8_t*)data->buf;
                    rsp.payload.u.in_data.data.len=res;
                    send_response(data,&rsp);
                }
            }
            
            close(data->dbfd);
            data->dbfd=-1;
            data->state=tsf_fsdb_state__clear;
            
            data->autoflush=tsf_true;
            
            send_simple_rsp(data,tsf_fsdb_response__payload__in_eof);
            break;
        }
        case tsf_fsdb_command__rm: {
            char *dataflnm;
            
            if (data->state!=tsf_fsdb_state__clear) {
                handler_fail_proto(data,
                                   "cannot use rm command outside of clear state");
            }
            
            if (!tsf_fsdb_guts_rm(fsdb,cmd->u.get.key,&dataflnm)) {
                if (tsf_get_error_code()==TSF_E_ELE_NOT_FOUND) {
                    send_simple_rsp(data,tsf_fsdb_response__payload__not_found);
                    break;
                } else {
                    handler_fail_tsf(data,tsf_true);
                }
            }
            
            hlogit(data,"executed rm command on %s",dataflnm);
            free(dataflnm);
            
            send_ok(data);
            break;
        }
        case tsf_fsdb_command__reset:
            reset_state(data);
            send_ok(data);
        default:
            handler_fail_proto(data,"received unrecognized command");
            break;
        }
    }
    
    handler_done(data);
    tsf_abort("should not get here");
}

/* entrypoint */

static void usage(void) {
    fprintf(stderr,"tsf_fsdb_daemon -p <port> -f <database path>\n");
    fprintf(stderr,"                [-i <pid file>] [-l <log file>] [-v] [-b]\n");
    fprintf(stderr,"-p <port>   the port on which the daemon listens\n");
    fprintf(stderr,"-f <path>   the path to the database being served\n");
    fprintf(stderr,"-i <path>   what to call the PID file\n");
    fprintf(stderr,"-l <path>   what to call the log file\n");
    fprintf(stderr,"-v          do more logging\n");
    fprintf(stderr,"-b          fork and run in background\n");
    fprintf(stderr,"-h          display this message\n");
    exit(1);
}

int main(int c,char **v) {
    int opt;
    int port=-1;
    char *dbflnm=NULL;
    char *pidflnm=NULL;
    char *logflnm=NULL;
    tsf_bool_t bg=tsf_false;
    struct sockaddr_in addr;
    socklen_t slen=sizeof(addr);
    int res;
    int servfd;
    int on=1;
    int64_t cnt=0;
    
    while ((opt=getopt(c,v,"p:f:i:l:bhv"))!=-1) {
        switch (opt) {
        case 'p':
            if (sscanf(optarg,"%d",&port)!=1) {
                usage();
            }
            break;
        case 'f':
            dbflnm=strdup(optarg);
            break;
        case 'i':
            pidflnm=strdup(optarg);
            break;
        case 'l':
            logflnm=strdup(optarg);
            break;
        case 'b':
            bg=tsf_true;
            break;
        case 'h':
            usage();
            break;
        case 'v':
            log_level++;
            break;
        default: usage();
        }
    }
    
    if (port<0 || dbflnm==NULL) {
        usage();
    }
    
    signal(SIGPIPE,SIG_IGN);

    limits=NULL;
    tsf_zip_wtr_attr_get_default(&wtr_attr);
    tsf_zip_rdr_attr_get_default(&rdr_attr);
    
    init_log();
    
    logit("opening database '%s'",dbflnm);
    
    CHECKTSF(fsdb=tsf_fsdb_open_local(dbflnm,
                                      NULL,NULL,NULL,
                                      tsf_true,
                                      tsf_false));
    
    logit("opening server socket on port %d",port);
    
    bzero(&addr,slen);
    addr.sin_family=AF_INET;
    addr.sin_port=htons(port);
    addr.sin_addr.s_addr=htonl(INADDR_ANY);
    CHECKSYS(servfd=socket(AF_INET,SOCK_STREAM,0));
    CHECKSYS(setsockopt(servfd,SOL_SOCKET,SO_REUSEADDR,&on,sizeof(on)));
    CHECKSYS(bind(servfd,(struct sockaddr*)&addr,slen));
    CHECKSYS(listen(servfd,5));
    
    if (bg) {
        if (logflnm!=NULL) {
            logit("going to background and switching to log file '%s'",logflnm);
        } else {
            logit("going to background");
        }
        CHECKSYS(res=fork());
        if (!res) {
            if (logflnm!=NULL) {
                reroute_log(logflnm);
            }
            logit("I'm at pid = %d",getpid());
            CHECKSYS(setsid());
        } else {
            exit(0);
        }
        signal(SIGINT,SIG_IGN);
    } else if (logflnm!=NULL) {
        logit("switching to log file '%s'",logflnm);
        reroute_log(logflnm);
    }
    
    if (pidflnm!=NULL) {
        FILE *fout=fopen(pidflnm,"w");
        CHECKSYSZ(fout);
        fprintf(fout,"%d\n",getpid());
        fclose(fout);
    }
    
    logit("operating on database '%s' and listening on port %d",
          dbflnm,port);
    
    pthread_mutex_init(&score_lock,NULL);
    n_connections=0;
    
    logit("waiting for connections...");

    for (;;) {
        slen=sizeof(addr);
        int s=accept(servfd,(struct sockaddr*)&addr,&slen);
        if (s<0) {
            logit("accept(): %s",strerror(errno));
            switch (errno) {
            case EAGAIN:
            case EPERM:
            case ENOBUFS:
            case ENOMEM:
            case ENETDOWN:
            case ENOPROTOOPT:
            case EHOSTDOWN:
            case EHOSTUNREACH:
            case EOPNOTSUPP:
            case ENETUNREACH:
            case EINTR:
                /* ignore */
                break;
            default:
                logit("fatal error in accept()");
                exit(1);
                break;
            }
        } else {
            handler_data_t *handler;
            pthread_t t;
            
            pthread_mutex_lock(&score_lock);
            n_connections++;
            pthread_mutex_unlock(&score_lock);
            
            handler=malloc(sizeof(handler_data_t));
            CHECKSYSZ(handler);
            handler->fd=s;
            handler->id=++cnt;
            handler->addr=addr;
            
            CHECKSYS(pthread_create(&t,NULL,handler_thread_main,handler));
            CHECKSYS(pthread_detach(t));
        }
    }

    return 1;
}

#else

int main(int c,char **v) {
    fprintf(stderr,"ERROR: pthreads, BSD sockets, or getopt not supported; aborting.\n");
    return 1;
}

#endif
