/*
 * Copyright (C) 2005 Filip Pizlo. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY FILIP PIZLO ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL FILIP PIZLO OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "tsf.h"
#include "tsf_indent.h"
#include "tsf_format.h"
#include "tsf_util.h"

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

static void usage() {
    fprintf(stderr,"Usage: tsf_to_math [<choice symbol prefix>] < <filename>\n");
    fprintf(stderr,"\n");
    fprintf(stderr,"tsf_to_math produces a Mathematica code file that defines a list corresponding to\n");
    fprintf(stderr,"the TSF input.  You can optionally specify a symbol prefix that will be\n");
    fprintf(stderr,"prepended to each choice symbol.  Other symbols are left unchanged.\n");
    fprintf(stderr,"\n");
    fprintf(stderr,"tsf_to_math reads the same kind of input as tsf_dump.  That is, it can read stream\n");
    fprintf(stderr,"files in the following formats:\n");
    fprintf(stderr,"\n");
    fprintf(stderr,"-> Raw TSF files, which use no compression.  In this case only a\n");
    fprintf(stderr,"   reset types command is required at the beginning of the stream.\n");
    fprintf(stderr,"   If it is not present, you can easily inject it.  It is simply\n");
    fprintf(stderr,"   \\0 followed by the three characters 'TSF'.\n");
    fprintf(stderr,"\n");
    fprintf(stderr,"-> TSF files compressed with ZLib.  In this case only a 'TSFz' magic\n");
    fprintf(stderr,"   header is required; if it is not present, you can easily inject it.\n");
    fprintf(stderr,"\n");
    fprintf(stderr,"-> TSF files compressed with libbzip2.  In this case only a 'TSFb'\n");
    fprintf(stderr,"   magic header is required; if it is not present, you can easily\n");
    fprintf(stderr,"   inject it.\n");
    exit(1);
}

static const char *choice_prefix="";

static void print(uint32_t indent,tsf_reflect_t *r) {
    switch (tsf_reflect_get_kind_code(r)) {
        case TSF_TK_INT8:
            printf(fi8,tsf_primitive_reflect_get(r,int8_t));
            break;
        case TSF_TK_UINT8:
            printf(fui8,tsf_primitive_reflect_get(r,uint8_t));
            break;
        case TSF_TK_INT16:
            printf(fi16,tsf_primitive_reflect_get(r,int16_t));
            break;
        case TSF_TK_UINT16:
            printf(fui16,tsf_primitive_reflect_get(r,uint16_t));
            break;
        case TSF_TK_INT32:
            printf(fi32,tsf_primitive_reflect_get(r,int32_t));
            break;
        case TSF_TK_UINT32:
            printf(fui32,tsf_primitive_reflect_get(r,uint32_t));
            break;
        case TSF_TK_INT64:
            printf(fi64,tsf_primitive_reflect_get(r,int64_t));
            break;
        case TSF_TK_UINT64:
            printf(fui64,tsf_primitive_reflect_get(r,uint64_t));
            break;
        case TSF_TK_FLOAT:
            printf("%f",tsf_primitive_reflect_get(r,float));
            break;
        case TSF_TK_DOUBLE:
            printf("%lf",tsf_primitive_reflect_get(r,double));
            break;
        case TSF_TK_BIT:
            printf("%s",tsf_bit_reflect_get(r)?"True":"False");
            break;
        case TSF_TK_VOID:
            printf("Null");
            break;
        case TSF_TK_ARRAY:
            if (tsf_array_reflect_get_num_elements(r)==0) {
                printf("{}");
                break;
            }
            printf("{\n");
            ++indent;
            {
                uint32_t i;
                for (i=0;i<tsf_array_reflect_get_num_elements(r);++i) {
                    printf("%s",tsf_indent_str(indent));
                    print(indent,tsf_array_reflect_get_element(r,i));
                    if (i!=tsf_array_reflect_get_num_elements(r)-1) {
                        printf(",");
                    }
                    printf("\n");
                }
            }
            --indent;
            printf("%s}",tsf_indent_str(indent));
            break;
        case TSF_TK_STRUCT:
            printf("{\n");
            ++indent;
            {
                tsf_type_t *type=tsf_reflect_get_type(r);
                uint32_t i;
                for (i=0;i<tsf_struct_type_get_num_elements(type);++i) {
                    tsf_named_type_t *n=tsf_struct_type_get_element(type,i);
                    printf("%s%s -> ",
                           tsf_indent_str(indent),
                           tsf_named_type_get_name(n));
                    print(indent,
			  tsf_struct_reflect_get_element(r,
							 tsf_named_type_get_name(n)));
		    if (i!=tsf_struct_type_get_num_elements(type)-1) {
			printf(",");
		    }
		    printf("\n");
                }
            }
            --indent;
            printf("%s}",tsf_indent_str(indent));
            break;
        case TSF_TK_CHOICE:
            if (tsf_choice_reflect_is_known(r)) {
                printf("%s%s[",choice_prefix,tsf_choice_reflect_get_choice(r));
                print(indent,
		      tsf_choice_reflect_get_data(r));
		printf("]");
            } else {
                printf("Null");
            }
            break;
        case TSF_TK_STRING:
            printf("\"");
            {
                const char *str;
                for (str=tsf_string_reflect_get(r);
                     *str;
                     ++str) {
                    if (*str=='\r') {
                        printf("\\r");
                    } else if (*str=='\n') {
                        printf("\\n");
                    } else if (*str=='\\'||
                               *str=='\"') {
                        printf("\\%c",*str);
                    } else {
			printf("%c",(unsigned char)*str);
                    }
                }
            }
            printf("\"");
            break;
        case TSF_TK_ANY:
            {
                tsf_reflect_t *sub_r;
                CHECKTSF(sub_r=tsf_reflect_from_buffer(
                    tsf_any_reflect_get_buffer(r)));
                print(indent,sub_r);
                tsf_reflect_destroy(sub_r);
            }
            break;
        default:
            printf("$Failed");
            break;
    }
}

int main(int c,char **v) {
    tsf_buffer_t *buf;
    tsf_stream_file_input_t *in;
    tsf_reflect_t *r;
    
    if (c>2) {
        usage();
    }
    
    if (c==2 && v[1][0]!='-') {
	choice_prefix=strdup(v[1]);
    }
    
    CHECKTSF(in=tsf_stream_file_input_fd_open(0,NULL,NULL,tsf_false,0));
    
    printf("{\n");

    for (;;) {
        buf=tsf_stream_file_input_read(in);
        if (buf==NULL && tsf_get_error_code()==TSF_E_EOF) {
            break;
        }
        
        CHECKTSF(buf);
        
        CHECKTSF(r=tsf_reflect_from_buffer(buf));
        
	printf("%s",tsf_indent_str(1));
        print(1,r);
        printf(",\n");
        
        tsf_reflect_destroy(r);
        tsf_buffer_destroy(buf);
    }
    
    printf("}\n");

    tsf_stream_file_input_close(in);
    
    return 0;
}

